Package: gh
Title: 'GitHub' 'API'
Version: 1.6.0
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("cre", "ctb")),
    person("Jennifer", "Bryan", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Minimal client to access the 'GitHub' 'API'.
License: MIT + file LICENSE
URL: https://gh.r-lib.org/, https://github.com/r-lib/gh
BugReports: https://github.com/r-lib/gh/issues
Depends: R (>= 4.1)
Imports: cli (>= 3.0.1), gitcreds, glue, httr2 (>= 1.0.6), ini,
        jsonlite, lifecycle, rlang (>= 1.1.0)
Suggests: connectcreds, covr, knitr, rmarkdown, rprojroot, spelling,
        testthat (>= 3.0.0), vctrs, webfakes (>= 1.5.0), withr
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-29
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-29 10:28:58 UTC; gaborcsardi
Author: Gábor Csárdi [cre, ctb],
  Jennifer Bryan [aut],
  Hadley Wickham [aut],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-29 13:20:07 UTC
Built: R 4.6.0; ; 2026-05-29 16:24:00 UTC; unix

Package: gh
Title: 'GitHub' 'API'
Version: 1.6.1
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("cre", "ctb")),
    person("Jennifer", "Bryan", role = "aut"),
    person("Hadley", "Wickham", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Minimal client to access the 'GitHub' 'API'.
License: MIT + file LICENSE
URL: https://gh.r-lib.org/, https://github.com/r-lib/gh
BugReports: https://github.com/r-lib/gh/issues
Depends: R (>= 4.1)
Imports: cli (>= 3.0.1), gitcreds, glue, httr2 (>= 1.0.6), ini,
        jsonlite, lifecycle, rlang (>= 1.1.0)
Suggests: connectcreds, covr, knitr, rmarkdown, rprojroot, spelling,
        testthat (>= 3.0.0), vctrs, webfakes (>= 1.5.0), withr
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-29
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-20 09:37:27 UTC; gaborcsardi
Author: Gábor Csárdi [cre, ctb],
  Jennifer Bryan [aut],
  Hadley Wickham [aut],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-20 10:20:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(format,gh_pat)
S3method(print,gh_pat)
S3method(print,gh_response)
S3method(str,gh_pat)
S3method(vctrs::vec_cast,list.gh_response)
S3method(vctrs::vec_ptype2,gh_response.gh_response)
export(fake_github_app)
export(gh)
export(gh_first)
export(gh_gql)
export(gh_last)
export(gh_next)
export(gh_prev)
export(gh_rate_limit)
export(gh_rate_limits)
export(gh_token)
export(gh_token_exists)
export(gh_tree_remote)
export(gh_whoami)
import(rlang)
importFrom(cli,cli_status)
importFrom(cli,cli_status_update)
importFrom(glue,glue)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,prettify)
importFrom(jsonlite,toJSON)
importFrom(lifecycle,deprecated)
importFrom(utils,URLencode)
importFrom(utils,capture.output)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# gh

<!-- badges: start -->

[![R-CMD-check](https://github.com/r-lib/gh/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/gh/actions)
[![](https://www.r-pkg.org/badges/version/gh)](https://www.r-pkg.org/pkg/gh)
[![CRAN Posit mirror
downloads](https://cranlogs.r-pkg.org/badges/gh)](https://www.r-pkg.org/pkg/gh)
[![R-CMD-check](https://github.com/r-lib/gh/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/gh/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/gh/graph/badge.svg)](https://app.codecov.io/gh/r-lib/gh)
<!-- badges: end -->

Minimalistic client to access GitHub’s
[REST](https://docs.github.com/rest) and
[GraphQL](https://docs.github.com/graphql) APIs.

## Installation and setup

Install the package from CRAN as usual:

``` r
install.packages("gh")
```

Install the development version from GitHub:

``` r
pak::pak("r-lib/gh")
```

### Authentication

The value returned by `gh::gh_token()` is used as Personal Access Token
(PAT). A token is needed for some requests, and to help with rate
limiting. gh can use your regular git credentials in the git credential
store, via the gitcreds package. Use `gitcreds::gitcreds_set()` to put a
PAT into the git credential store. If you cannot use the credential
store, set the `GITHUB_PAT` environment variable to your PAT. See the
details in the `?gh::gh_token` manual page and the manual of the
gitcreds package.

### API URL

- The `GITHUB_API_URL` environment variable, if set, is used for the
  default github api url.

## Usage

``` r
library(gh)
```

Use the `gh()` function to access all API endpoints. The endpoints are
listed in the [documentation](https://docs.github.com/rest).

The first argument of `gh()` is the endpoint. You can just copy and
paste the API endpoints from the documentation. Note that the leading
slash must be included as well.

From
<https://docs.github.com/rest/reference/repos#list-repositories-for-a-user>
you can copy and paste `GET /users/{username}/repos` into your `gh()`
call. E.g.

``` r
my_repos <- gh("GET /users/{username}/repos", username = "gaborcsardi")
vapply(my_repos, "[[", "", "name")
#>  [1] "after"                "air"                  "alda"                
#>  [4] "alexr"                "all.primer.tutorials" "altlist"             
#>  [7] "anticlust"            "argufy"               "ask"                 
#> [10] "async"                "autobrew-bundler"     "available-work"      
#> [13] "baguette"             "BCEA"                 "BH"                  
#> [16] "bigrquerystorage"     "brew-big-sur"         "brokenPackage"       
#> [19] "brulee"               "build-r-app"          "butcher"             
#> [22] "censored"             "cf-tunnel"            "checkinstall"        
#> [25] "cli"                  "clock"                "comments"            
#> [28] "covr"                 "covrlabs"             "cran-metadata"
```

The JSON result sent by the API is converted to an R object.

Parameters can be passed as extra arguments. E.g.

``` r
my_repos <- gh(
  "/users/{username}/repos",
  username = "gaborcsardi",
  sort = "created")
vapply(my_repos, "[[", "", "name")
#>  [1] "data-dict.yaml"     "vroom"              "rds2rust"          
#>  [4] "tree-sitter-toml"   "tstoml"             "tree-sitter-json"  
#>  [7] "uncovr"             "tsjsonc"            "opentelemetry-cpp" 
#> [10] "ellmer"             "air"                "s2"                
#> [13] "parzer"             "shinycoreci"        "secret-service-cli"
#> [16] "phantomjs"          "FSA"                "greta"             
#> [19] "webdriver"          "clock"              "testthat"          
#> [22] "jsonlite"           "duckdb"             "duckdb-r"          
#> [25] "httpuv"             "unwind"             "httr2"             
#> [28] "pins-r"             "install-figlet"     "weird-package"
```

### POST, PATCH, PUT and DELETE requests

POST, PATCH, PUT, and DELETE requests can be sent by including the HTTP
verb before the endpoint, in the first argument. E.g. to create a
repository:

``` r
new_repo <- gh("POST /user/repos", name = "my-new-repo-for-gh-testing")
```

and then delete it:

``` r
gh("DELETE /repos/{owner}/{repo}", owner = "gaborcsardi",
   repo = "my-new-repo-for-gh-testing")
```

### Tokens

By default the `GITHUB_PAT` environment variable is used. Alternatively,
one can set the `.token` argument of `gh()`.

### Pagination

Supply the `page` parameter to get subsequent pages:

``` r
my_repos2 <- gh("GET /orgs/{org}/repos", org = "r-lib", page = 2)
vapply(my_repos2, "[[", "", "name")
#>  [1] "gh"          "desc"        "profvis"     "sodium"      "gargle"     
#>  [6] "remotes"     "jose"        "backports"   "rcmdcheck"   "vdiffr"     
#> [11] "callr"       "mockery"     "here"        "revdepcheck" "processx"   
#> [16] "vctrs"       "debugme"     "usethis"     "rlang"       "pkgload"    
#> [21] "httrmock"    "pkgbuild"    "prettycode"  "roxygen2md"  "pkgapi"     
#> [26] "zeallot"     "liteq"       "keyring"     "sloop"       "styler"
```

## Environment Variables and Options

- The `GITHUB_API_URL` environment variable is used for the default
  github api url.
- The `GITHUB_PAT` and `GITHUB_TOKEN` environment variables are used, if
  set, in this order, as default token. Consider using the git
  credential store instead, see `?gh::gh_token`.
- The `GH_VALIDATE_TOKENS` environment variable controls what happens
  when gh retrieves a PAT in an unrecognized format. Set it to `"off"`
  to skip validation, `"warn"` (the default) to issue a warning and use
  the PAT anyway, or `"error"` to abort. The `gh_validate_tokens` R
  option takes precedence over the environment variable.

## Code of Conduct

Please note that the gh project is released with a [Contributor Code of
Conduct](https://gh.r-lib.org/CODE_OF_CONDUCT.html). By contributing to
this project, you agree to abide by its terms.

## License

MIT © Gábor Csárdi, Jennifer Bryan, Hadley Wickham

# GitHub CLI

`gh` is GitHub on the command line. It brings pull requests, issues, and other GitHub concepts to the terminal next to where you are already working with `git` and your code.

![screenshot of gh pr status](https://user-images.githubusercontent.com/98482/84171218-327e7a80-aa40-11ea-8cd1-5177fc2d0e72.png)

GitHub CLI is supported for users on GitHub.com and GitHub Enterprise Server 2.20+ with support for macOS, Windows, and Linux.

## Documentation

For [installation options see below](#installation), for usage instructions [see the manual][manual].

## Contributing

If anything feels off, or if you feel that some functionality is missing, please check out the [contributing page][contributing]. There you will find instructions for sharing your feedback, building the tool locally, and submitting pull requests to the project.

If you are a hubber and are interested in shipping new commands for the CLI, check out our [doc on internal contributions][intake-doc].

<!-- this anchor is linked to from elsewhere, so avoid renaming it -->
## Installation

### macOS

`gh` is available via [Homebrew][], [MacPorts][], [Conda][], [Spack][], [Webi][], and as a downloadable binary from the [releases page][].

#### Homebrew

| Install:          | Upgrade:          |
| ----------------- | ----------------- |
| `brew install gh` | `brew upgrade gh` |

#### MacPorts

| Install:               | Upgrade:                                       |
| ---------------------- | ---------------------------------------------- |
| `sudo port install gh` | `sudo port selfupdate && sudo port upgrade gh` |

#### Conda

| Install:                                 | Upgrade:                                |
|------------------------------------------|-----------------------------------------|
| `conda install gh --channel conda-forge` | `conda update gh --channel conda-forge` |

Additional Conda installation options available on the [gh-feedstock page](https://github.com/conda-forge/gh-feedstock#installing-gh).

#### Spack

| Install:           | Upgrade:                                 |
| ------------------ | ---------------------------------------- |
| `spack install gh` | `spack uninstall gh && spack install gh` |

#### Webi

| Install:                            | Upgrade:         |
| ----------------------------------- | ---------------- |
| `curl -sS https://webi.sh/gh \| sh` | `webi gh@stable` |

For more information about the Webi installer see [its homepage](https://webinstall.dev/).

### Linux & BSD

`gh` is available via:
- [our Debian and RPM repositories](./docs/install_linux.md);
- community-maintained repositories in various Linux distros;
- OS-agnostic package managers such as [Homebrew](#homebrew), [Conda](#conda), [Spack](#spack), [Webi](#webi); and
- our [releases page][] as precompiled binaries.

For more information, see [Linux & BSD installation](./docs/install_linux.md).

### Windows

`gh` is available via [WinGet][], [scoop][], [Chocolatey][], [Conda](#conda), [Webi](#webi), and as downloadable MSI.

#### WinGet

| Install:            | Upgrade:            |
| ------------------- | --------------------|
| `winget install --id GitHub.cli` | `winget upgrade --id GitHub.cli` |

> **Note**  
> The Windows installer modifies your PATH. When using Windows Terminal, you will need to **open a new window** for the changes to take effect. (Simply opening a new tab will _not_ be sufficient.)

#### scoop

| Install:           | Upgrade:           |
| ------------------ | ------------------ |
| `scoop install gh` | `scoop update gh`  |

#### Chocolatey

| Install:           | Upgrade:           |
| ------------------ | ------------------ |
| `choco install gh` | `choco upgrade gh` |

#### Signed MSI

MSI installers are available for download on the [releases page][].

### Codespaces

To add GitHub CLI to your codespace, add the following to your [devcontainer file](https://docs.github.com/en/codespaces/setting-up-your-project-for-codespaces/adding-features-to-a-devcontainer-file):

```json
"features": {
  "ghcr.io/devcontainers/features/github-cli:1": {}
}
```

### GitHub Actions

GitHub CLI comes pre-installed in all [GitHub-Hosted Runners](https://docs.github.com/en/actions/using-github-hosted-runners/about-github-hosted-runners).

### Other platforms

Download packaged binaries from the [releases page][].

### Build from source

See here on how to [build GitHub CLI from source][build from source].

## Comparison with hub

For many years, [hub][] was the unofficial GitHub CLI tool. `gh` is a new project that helps us explore
what an official GitHub CLI tool can look like with a fundamentally different design. While both
tools bring GitHub to the terminal, `hub` behaves as a proxy to `git`, and `gh` is a standalone
tool. Check out our [more detailed explanation][gh-vs-hub] to learn more.

[manual]: https://cli.github.com/manual/
[Homebrew]: https://brew.sh
[MacPorts]: https://www.macports.org
[winget]: https://github.com/microsoft/winget-cli
[scoop]: https://scoop.sh
[Chocolatey]: https://chocolatey.org
[Conda]: https://docs.conda.io/en/latest/
[Spack]: https://spack.io
[Webi]: https://webinstall.dev
[releases page]: https://github.com/cli/cli/releases/latest
[hub]: https://github.com/github/hub
[contributing]: ./.github/CONTRIBUTING.md
[gh-vs-hub]: ./docs/gh-vs-hub.md
[build from source]: ./docs/source.md
[intake-doc]: ./docs/working-with-us.md

To install in the BDGP/LBL, we want most of the stuff to end up in the
local tree, but the utility scripts (overlap, etc...) to land in 
/usr/local/bdgp/bin.

   make
   make test
   make INSTALLSCRIPT=/usr/local/bdgp/bin install

skip_if_no_github <- function(has_scope = NULL) {
  skip_if_offline("github.com")
  skip_on_cran()

  if (gh_token() == "") {
    skip("No GitHub token")
  }

  if (!is.null(has_scope) && !has_scope %in% test_scopes()) {
    skip(cli::format_inline("Current token lacks '{has_scope}' scope"))
  }
}

test_scopes <- function() {
  # whoami fails on GHA
  whoami <- env_cache(cache, "whoami", tryCatch(
    gh_whoami(),
    error = function(err) list(scopes = "")
  ))
  strsplit(whoami$scopes, ", ")[[1]]
}

cache <- new_environment()

#' Find the GitHub remote associated with a path
#'
#' This is handy helper if you want to make gh requests related to the
#' current project.
#'
#' @param path Path that is contained within a git repo.
#' @return If the repo has a github remote, a list containing `username`
#'    and `repo`. Otherwise, an error.
#' @export
#' @examplesIf interactive()
#' gh_tree_remote()
gh_tree_remote <- function(path = ".") {
  github_remote(git_remotes(path), path)
}

github_remote <- function(x, path) {
  remotes <- lapply(x, github_remote_parse)
  remotes <- remotes[!vapply(remotes, is.null, logical(1))]

  if (length(remotes) == 0) {
    cli::cli_abort("No GitHub remotes found at {.path {path}}")
  }

  if (length(remotes) > 1) {
    if (any(names(remotes) == "origin")) {
      warning("Multiple github remotes found. Using origin.", call. = FALSE)
      remotes <- remotes[["origin"]]
    } else {
      warning("Multiple github remotes found. Using first.", call. = FALSE)
      remotes <- remotes[[1]]
    }
  } else {
    remotes[[1]]
  }
}

github_remote_parse <- function(x) {
  if (length(x) == 0) {
    return(NULL)
  }
  if (!grepl("github", x)) {
    return(NULL)
  }

  # https://github.com/hadley/devtools.git
  # https://github.com/hadley/devtools
  # git@github.com:hadley/devtools.git
  re <- "github[^/:]*[/:]([^/]+)/(.*?)(?:\\.git)?$"
  m <- regexec(re, x)
  match <- regmatches(x, m)[[1]]

  if (length(match) == 0) {
    return(NULL)
  }

  list(
    username = match[2],
    repo = match[3]
  )
}

git_remotes <- function(path = ".") {
  conf <- git_config(path)
  remotes <- conf[grepl("^remote", names(conf))]

  remotes <- discard(remotes, function(x) is.null(x$url))
  urls <- vapply(remotes, "[[", "url", FUN.VALUE = character(1))

  names(urls) <- gsub('^remote "(.*?)"$', "\\1", names(remotes))
  urls
}



git_config <- function(path = ".") {
  config_path <- file.path(repo_root(path), ".git", "config")
  if (!file.exists(config_path)) {
    cli::cli_abort("git config does not exist at {.path {path}}")
  }
  ini::read.ini(config_path, "UTF-8")
}

repo_root <- function(path = ".") {
  if (!file.exists(path)) {
    cli::cli_abort("Can't find repo at {.path {path}}")
  }

  # Walk up to root directory
  while (!has_git(path)) {
    if (is_root(path)) {
      cli::cli_abort("Could not find git root from {.path {path}}.")
    }

    path <- dirname(path)
  }

  path
}

has_git <- function(path) {
  file.exists(file.path(path, ".git"))
}

is_root <- function(path) {
  identical(path, dirname(path))
}

test_that("gh_tree_remote returns username and repo for a github remote", {
  repo <- make_git_repo(list(origin = "https://github.com/x/y.git"))
  expect_equal(
    gh_tree_remote(repo),
    list(username = "x", repo = "y")
  )
})

test_that("gh_tree_remote errors when no github remotes are configured", {
  repo <- make_git_repo(list(origin = "https://gitlab.com/x/y.git"))
  expect_snapshot(
    error = TRUE,
    gh_tree_remote(repo),
    transform = transform_tempdir
  )
})

test_that("gh_tree_remote walks up from a subdirectory to find the repo root", {
  repo <- make_git_repo(list(origin = "https://github.com/x/y.git"))
  sub <- file.path(repo, "pkg", "R")
  dir.create(sub, recursive = TRUE)
  expect_equal(
    gh_tree_remote(sub),
    list(username = "x", repo = "y")
  )
})

test_that("git_remotes skips remotes without a url", {
  dir <- withr::local_tempdir()
  dir.create(file.path(dir, ".git"))
  writeLines(
    c(
      '[remote "origin"]',
      "  url = https://github.com/x/y.git",
      '[remote "broken"]',
      "  fetch = +refs/heads/*:refs/remotes/broken/*"
    ),
    file.path(dir, ".git", "config")
  )
  expect_equal(
    git_remotes(dir),
    c(origin = "https://github.com/x/y.git")
  )
})

test_that("git_config errors when .git/config does not exist", {
  dir <- withr::local_tempdir()
  dir.create(file.path(dir, ".git"))
  expect_snapshot(error = TRUE, git_config(dir), transform = transform_tempdir)
})

test_that("repo_root errors on a path that does not exist", {
  expect_snapshot(
    error = TRUE,
    repo_root(file.path(tempdir(), "does-not-exist-xyz")),
    transform = transform_tempdir
  )
})

test_that("repo_root errors when no git root is found", {
  dir <- withr::local_tempdir()
  expect_snapshot(error = TRUE, repo_root(dir), transform = transform_tempdir)
})

test_that("picks origin if available", {
  remotes <- list(
    upstream = "https://github.com/x/1",
    origin = "https://github.com/x/2"
  )

  expect_warning(gr <- github_remote(remotes, "."), "Using origin")
  expect_equal(gr$repo, "2")
})

test_that("otherwise picks first", {
  remotes <- list(
    a = "https://github.com/x/1",
    b = "https://github.com/x/2"
  )

  expect_warning(gr <- github_remote(remotes, "."), "Using first")
  expect_equal(gr$repo, "1")
})


# Parsing -----------------------------------------------------------------

test_that("parses common url forms", {
  expected <- list(username = "x", repo = "y")

  expect_equal(github_remote_parse("https://github.com/x/y.git"), expected)
  expect_equal(github_remote_parse("https://github.com/x/y"), expected)
  expect_equal(github_remote_parse("git@github.com:x/y.git"), expected)
})

test_that("returns NULL if can't parse", {
  expect_equal(github_remote_parse("blah"), NULL)
})

test_that("github_remote_parse returns NULL for zero-length input", {
  expect_null(github_remote_parse(character(0)))
})

test_that("github_remote_parse returns NULL when regex doesn't match", {
  expect_null(github_remote_parse("github"))
})

test_that("github_remote returns the single remote unchanged", {
  remotes <- list(origin = "https://github.com/x/y.git")
  expect_silent(gr <- github_remote(remotes, "."))
  expect_equal(gr, list(username = "x", repo = "y"))
})

#' Print the result of a GitHub API call
#'
#' @param x The result object.
#' @param ... Ignored.
#' @return The JSON result.
#'
#' @importFrom jsonlite prettify toJSON
#' @export
#' @method print gh_response

print.gh_response <- function(x, ...) {
  if (inherits(x, c("raw", "path"))) {
    attributes(x) <- list(class = class(x))
    print.default(x)
  } else {
    print(toJSON(unclass(x), pretty = TRUE, auto_unbox = TRUE, force = TRUE))
  }
}

gh_process_response <- function(resp, gh_req) {
  if (!inherits(resp, "httr2_response")) {
    stop_input_type(resp, "an <httr2_response> object")
  }

  status <- httr2::resp_status(resp)
  content_type <- httr2::resp_content_type(resp)
  gh_media_type <- httr2::resp_header(resp, "x-github-media-type")

  is_raw <- identical(content_type, "application/octet-stream") ||
    isTRUE(grepl("param=raw$", gh_media_type, ignore.case = TRUE))
  is_ondisk <- inherits(resp$body, "httr2_path") && !is.null(gh_req$dest)
  # An empty body short-circuits before the JSON / raw branches: this
  # covers a plain `304 Not Modified` (where Content-Type is typically
  # absent) and any other content-less response. If the body is present
  # — e.g. when httr2's HTTP cache attaches a stored body to a 304
  # revalidation reply — we fall through and parse it.
  is_empty <- length(resp$body) == 0

  if (is_ondisk) {
    res <- as.character(resp$body)
    file.rename(res, gh_req$dest)
    res <- gh_req$dest
  } else if (is_empty) {
    res <- list()
  } else if (grepl("^application/json", content_type, ignore.case = TRUE)) {
    res <- httr2::resp_body_json(resp)
  } else if (is_raw) {
    res <- httr2::resp_body_raw(resp)
  } else {
    if (grepl("^text/html", content_type, ignore.case = TRUE)) {
      warning("Response came back as html :(", call. = FALSE)
    }
    res <- list(message = httr2::resp_body_string(resp))
  }

  attr(res, "response") <- httr2::resp_headers(resp)
  attr(res, "httr2_response") <- resp
  attr(res, "request") <- remove_headers(gh_req)

  if (is_ondisk) {
    class(res) <- c("gh_response", "path")
  } else if (is_raw) {
    class(res) <- c("gh_response", "raw")
  } else {
    class(res) <- c("gh_response", "list")
  }
  res
}

remove_headers <- function(x) {
  x[names(x) != "headers"]
}

# Add vctrs methods that strip attributes from gh_response when combining,
# enabling rectangling via unnesting etc
# See <https://github.com/r-lib/gh/issues/161> for more details
#' @exportS3Method vctrs::vec_ptype2
vec_ptype2.gh_response.gh_response <- function(x, y, ...) {
  list()
}

#' @exportS3Method vctrs::vec_cast
vec_cast.list.gh_response <- function(x, to, ...) {
  attributes(x) <- NULL
  x
}

test_that("gh_process_response errors on non-httr2_response input", {
  expect_snapshot(error = TRUE, gh_process_response(list(), list()))
})

test_that("works with empty bodies", {
  local_fake_github()
  out <- gh("GET /orgs/{org}/repos", org = "gh-org-testing-no-repos")
  expect_equal(out, list(), ignore_attr = TRUE)

  out <- gh("POST /markdown", text = "")
  expect_equal(out, list(), ignore_attr = TRUE)
})

test_that("handles 304 Not Modified (#219)", {
  local_fake_github()
  withr::local_options(gh_cache = FALSE)

  res <- gh("/users/{user}/repos", user = "hadley", .limit = 2)
  etag <- attr(res, "response")$etag
  expect_false(is.null(etag))

  res2 <- gh(
    "/users/{user}/repos",
    user = "hadley",
    .limit = 2,
    .send_headers = c("If-None-Match" = etag)
  )
  expect_equal(res2, list(), ignore_attr = TRUE)
  expect_s3_class(res2, "gh_response")
  expect_equal(attr(res2, "response")$etag, etag)
})

test_that("works with empty bodies from DELETE", {
  local_fake_github()
  out <- gh(
    "POST /gists",
    files = list(x = list(content = "y")),
    public = FALSE
  )
  out <- gh("DELETE /gists/{gist_id}", gist_id = out$id)
  expect_equal(out, list(), ignore_attr = TRUE)
})

test_that("can get raw response", {
  local_fake_github()
  res <- gh(
    "GET /repos/{owner}/{repo}/contents/{path}",
    owner = "r-lib",
    repo = "gh",
    path = "DESCRIPTION",
    .send_headers = c(Accept = "application/vnd.github.v3.raw")
  )

  expect_equal(
    attr(res, "response")[["x-github-media-type"]],
    "github.v3; param=raw"
  )
  expect_equal(class(res), c("gh_response", "raw"))
})

test_that("can download files", {
  local_fake_github()
  tmp <- withr::local_tempfile()
  res_file <- gh(
    "/orgs/{org}/repos",
    org = "r-lib",
    type = "sources",
    .destfile = tmp
  )
  expect_equal(class(res_file), c("gh_response", "path"))
  expect_equal(res_file, tmp, ignore_attr = TRUE)
})

test_that("warns if output is HTML", {
  local_fake_github()
  expect_snapshot(res <- gh("POST /markdown", text = "foo"))

  expect_equal(res, list(message = "<p>foo</p>\n"), ignore_attr = TRUE)
  expect_equal(class(res), c("gh_response", "list"))
})

test_that("captures details to recreate request", {
  local_fake_github()
  res <- gh("/orgs/{org}/repos", org = "r-lib", .per_page = 1)

  req <- attr(res, "request")
  expect_type(req, "list")
  expect_match(req$url, "/orgs/r-lib/repos$")
  expect_equal(req$query, list(per_page = 1))
})

test_that("output file is not overwritten on error", {
  local_fake_github()
  tmp <- withr::local_tempfile()
  writeLines("foo", tmp)

  err <- tryCatch(
    gh("/missing", .destfile = tmp),
    error = function(e) e
  )

  expect_true(file.exists(tmp))
  expect_equal(readLines(tmp), "foo")
  expect_true(!is.null((err$response_content)))
})


test_that("gh_response objects can be combined via vctrs #161", {
  local_fake_github()
  skip_if_not_installed("vctrs")
  user_1 <- gh("/users", .limit = 1)
  user_2 <- gh("/users", .limit = 1, )
  user_vec <- vctrs::vec_c(user_1, user_2)
  user_df <- vctrs::vec_rbind(user_1[[1]], user_2[[1]])
  expect_equal(length(user_vec), 2)
  expect_equal(nrow(user_df), 2)
})

trim_ws <- function(x) {
  sub("\\s*$", "", sub("^\\s*", "", x))
}

## from devtools, among other places
compact <- function(x) {
  is_empty <- vapply(x, function(x) length(x) == 0, logical(1))
  x[!is_empty]
}

## from purrr, among other places
`%||%` <- function(x, y) {
  if (is.null(x)) {
    y
  } else {
    x
  }
}

## as seen in purrr, with the name `has_names()`
has_name <- function(x) {
  nms <- names(x)
  if (is.null(nms)) {
    rep_len(FALSE, length(x))
  } else {
    !(is.na(nms) | nms == "")
  }
}

has_no_names <- function(x) all(!has_name(x))

## if all names are "", strip completely
cleanse_names <- function(x) {
  if (has_no_names(x)) {
    names(x) <- NULL
  }
  x
}

## to process HTTP headers, i.e. combine defaults w/ user-specified headers
## in the spirit of modifyList(), except
## x and y are vectors (not lists)
## name comparison is case insensitive
## http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
## x will be default headers, y will be user-specified
modify_vector <- function(x, y = NULL) {
  if (length(y) == 0L) {
    return(x)
  }
  lnames <- function(x) tolower(names(x))
  c(x[!(lnames(x) %in% lnames(y))], y)
}


discard <- function(.x, .p, ...) {
  sel <- probe(.x, .p, ...)
  .x[is.na(sel) | !sel]
}
probe <- function(.x, .p, ...) {
  if (is.logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    vapply(.x, .p, logical(1), ...)
  }
}

drop_named_nulls <- function(x) {
  if (has_no_names(x)) {
    return(x)
  }
  named <- has_name(x)
  null <- vapply(x, is.null, logical(1))
  cleanse_names(x[!named | !null])
}

check_named_nas <- function(x) {
  if (has_no_names(x)) {
    return(x)
  }
  named <- has_name(x)
  na <- vapply(x, FUN.VALUE = logical(1), function(v) {
    is.atomic(v) && anyNA(v)
  })
  bad <- which(named & na)
  if (length(bad)) {
    str <- paste0("`", names(x)[bad], "`", collapse = ", ")
    stop("Named NA parameters are not allowed: ", str)
  }
}

can_load <- function(pkg) {
  isTRUE(requireNamespace(pkg, quietly = TRUE))
}

is_interactive <- function() {
  opt <- getOption("rlib_interactive")
  if (isTRUE(opt)) {
    TRUE
  } else if (identical(opt, FALSE)) {
    FALSE
  } else if (tolower(getOption("knitr.in.progress", "false")) == "true") {
    FALSE
  } else if (identical(Sys.getenv("TESTTHAT"), "true")) {
    FALSE
  } else {
    interactive()
  }
}

is_testing <- function() {
  identical(Sys.getenv("TESTTHAT"), "true")
}

test_that("can detect presence vs absence names", {
  expect_identical(has_name(list("foo", "bar")), c(FALSE, FALSE))
  expect_identical(has_name(list(a = "foo", "bar")), c(TRUE, FALSE))

  expect_identical(
    has_name({
      x <- list("foo", "bar")
      names(x)[1] <- "a"
      x
    }),
    c(TRUE, FALSE)
  )
  expect_identical(
    has_name({
      x <- list("foo", "bar")
      names(x)[1] <- "a"
      names(x)[2] <- ""
      x
    }),
    c(TRUE, FALSE)
  )

  expect_identical(
    has_name({
      x <- list("foo", "bar")
      names(x)[1] <- ""
      x
    }),
    c(FALSE, FALSE)
  )
  expect_identical(
    has_name({
      x <- list("foo", "bar")
      names(x)[1] <- ""
      names(x)[2] <- ""
      x
    }),
    c(FALSE, FALSE)
  )
})

test_that("named NULL is dropped", {
  tcs <- list(
    list(list(), list()),
    list(list(a = 1), list(a = 1)),
    list(list(NULL), list(NULL)),
    list(list(a = NULL), list()),
    list(list(NULL, a = NULL, 1), list(NULL, 1)),
    list(list(a = NULL, b = 1, 5), list(b = 1, 5))
  )

  for (tc in tcs) {
    expect_identical(
      drop_named_nulls(tc[[1]]),
      tc[[2]],
      info = tc
    )
  }
})

test_that("named NA is error", {
  goodtcs <- list(
    list(),
    list(NA),
    list(NA, NA_integer_, a = 1)
  )

  badtcs <- list(
    list(b = NULL, a = NA),
    list(a = NA_integer_),
    list(NA, c = NA_real_)
  )

  for (tc in goodtcs) {
    expect_silent(check_named_nas(tc))
  }

  for (tc in badtcs) {
    expect_snapshot(error = TRUE, check_named_nas(tc))
  }
})


test_that(".parse_params combines list .params with ... params", {
  params <- list(
    .parse_params(org = "ORG", repo = "REPO", number = "1"),
    .parse_params(org = "ORG", repo = "REPO", .params = list(number = "1")),
    .parse_params(.params = list(org = "ORG", repo = "REPO", number = "1"))
  )

  expect_identical(params[[1]], params[[2]])
  expect_identical(params[[2]], params[[3]])
})

test_that("trim_ws strips leading and trailing whitespace", {
  expect_identical(trim_ws("foo"), "foo")
  expect_identical(trim_ws("  foo"), "foo")
  expect_identical(trim_ws("foo  "), "foo")
  expect_identical(trim_ws("  foo  "), "foo")
  expect_identical(trim_ws("\t foo \n"), "foo")
  expect_identical(trim_ws("foo bar"), "foo bar")
  expect_identical(trim_ws(""), "")
  expect_identical(trim_ws(c("  a", "b  ")), c("a", "b"))
})

test_that("has_no_names detects fully-unnamed input", {
  expect_true(has_no_names(list("a", "b")))
  expect_true(has_no_names(list()))
  expect_false(has_no_names(list(a = 1, "b")))
  expect_false(has_no_names(list(a = 1, b = 2)))

  x <- list("a", "b")
  names(x) <- c("", "")
  expect_true(has_no_names(x))
})

test_that("cleanse_names drops all-empty names but keeps real ones", {
  x <- list("a", "b")
  names(x) <- c("", "")
  expect_null(names(cleanse_names(x)))

  y <- list(a = 1, b = 2)
  expect_identical(cleanse_names(y), y)

  z <- list(a = 1, "b")
  expect_identical(cleanse_names(z), z)

  expect_identical(cleanse_names(list()), list())
})

test_that("modify_vector overrides x with y case-insensitively", {
  expect_identical(modify_vector(c(a = "1", b = "2")), c(a = "1", b = "2"))
  expect_identical(
    modify_vector(c(a = "1", b = "2"), NULL),
    c(a = "1", b = "2")
  )

  expect_identical(
    modify_vector(c(a = "1", b = "2"), c(b = "9")),
    c(a = "1", b = "9")
  )

  expect_identical(
    modify_vector(c(Accept = "json"), c(accept = "xml")),
    c(accept = "xml")
  )

  expect_identical(
    modify_vector(c(a = "1"), c(b = "2")),
    c(a = "1", b = "2")
  )
})

test_that("discard drops elements matching predicate", {
  expect_identical(
    discard(list(1, NULL, 2, NULL), is.null),
    list(1, 2)
  )
  expect_identical(
    discard(list(a = 1, b = NULL, c = 2), is.null),
    list(a = 1, c = 2)
  )
  expect_identical(discard(list(), is.null), list())
})

test_that("discard accepts a logical selector", {
  expect_identical(
    discard(list("a", "b", "c"), c(TRUE, FALSE, TRUE)),
    list("b")
  )
  expect_identical(
    discard(list("a", "b", "c"), c(TRUE, NA, FALSE)),
    list("b", "c")
  )
})

test_that("discard errors on logical selector of wrong length", {
  expect_snapshot(
    error = TRUE,
    discard(list("a", "b"), c(TRUE, FALSE, TRUE))
  )
})

#' Query the GitHub API
#'
#' This is an extremely minimal client. You need to know the API
#' to be able to use this client. All this function does is:
#' * Try to substitute each listed parameter into `endpoint`, using the
#'   `{parameter}` notation.
#' * If a GET request (the default), then add all other listed parameters
#'   as query parameters.
#' * If not a GET request, then send the other parameters in the request
#'   body, as JSON.
#' * Convert the response to an R list using [jsonlite::fromJSON()].
#'
#' @param endpoint GitHub API endpoint. Must be one of the following forms:
#'    * `METHOD path`, e.g. `GET /rate_limit`,
#'    * `path`, e.g. `/rate_limit`,
#'    * `METHOD url`, e.g. `GET https://api.github.com/rate_limit`,
#'    * `url`, e.g. `https://api.github.com/rate_limit`.
#'
#'    If the method is not supplied, will use `.method`, which defaults
#'    to `"GET"`.
#' @param ... Name-value pairs giving API parameters. Will be matched into
#'   `endpoint` placeholders, sent as query parameters in GET requests, and as a
#'   JSON body of POST requests. If there is only one unnamed parameter, and it
#'   is a raw vector, then it will not be JSON encoded, but sent as raw data, as
#'   is. This can be used for example to add assets to releases. Named `NULL`
#'   values are silently dropped. For GET requests, named `NA` values trigger an
#'   error. For other methods, named `NA` values are included in the body of the
#'   request, as JSON `null`.
#' @param per_page,.per_page Number of items to return per page. If omitted,
#'   will be substituted by `max(.limit, 100)` if `.limit` is set,
#'   otherwise determined by the API (never greater than 100).
#' @param .destfile Path to write response to disk. If `NULL` (default),
#'   response will be processed and returned as an object. If path is given,
#'   response will be written to disk in the form sent. gh writes the
#'   response to a temporary file, and renames that file to `.destfile`
#'   after the request was successful. The name of the temporary file is
#'   created by adding a `-<random>.gh-tmp` suffix to it, where `<random>`
#'   is an ASCII string with random characters. gh removes the temporary
#'   file on error.
#' @param .overwrite If `.destfile` is provided, whether to overwrite an
#'   existing file.  Defaults to `FALSE`. If an error happens the original
#'   file is kept.
#' @param .token Authentication token. Defaults to [gh_token()].
#' @param .api_url Github API url (default: <https://api.github.com>). Used
#'   if `endpoint` just contains a path. Defaults to `GITHUB_API_URL`
#'   environment variable if set.
#' @param .method HTTP method to use if not explicitly supplied in the
#'    `endpoint`.
#' @param .limit Number of records to return. This can be used
#'   instead of manual pagination. By default it is `NULL`,
#'   which means that the defaults of the GitHub API are used.
#'   You can set it to a number to request more (or less)
#'   records, and also to `Inf` to request all records.
#'   Note, that if you request many records, then multiple GitHub
#'   API calls are used to get them, and this can take a potentially
#'   long time.
#' @param .accept The value of the `Accept` HTTP header. Defaults to
#'   `"application/vnd.github.v3+json"` . If `Accept` is given in
#'   `.send_headers`, then that will be used. This parameter can be used to
#'   provide a custom media type, in order to access a preview feature of
#'   the API.
#' @param .send_headers Named character vector of header field values
#'   (except `Authorization`, which is handled via `.token`). This can be
#'   used to override or augment the default `User-Agent` header:
#'   `"https://github.com/r-lib/gh"`.
#' @param .progress Whether to show a progress indicator for calls that
#'   need more than one HTTP request.
#' @param .params Additional list of parameters to append to `...`.
#'   It is easier to use this than `...` if you have your parameters in
#'   a list already.
#' @param .max_wait Maximum number of seconds to wait if rate limited.
#'   Defaults to 10 minutes.
#' @param .max_rate Maximum request rate in requests per second. Set
#'   this to automatically throttle requests.
#' @return Answer from the API as a `gh_response` object, which is also a
#'   `list`. Failed requests will generate an R error. Requests that
#'   generate a raw response will return a raw vector.
#'
#' @section Caching:
#' `gh()` uses httr2's HTTP cache by default. It is stored in
#' `tools::R_user_dir("gh", "cache")`, capped at 100 MB, and shared
#' across R sessions. When the cache holds a previous response for a
#' request, `gh()` lets httr2 attach `If-None-Match` / `If-Modified-Since`
#' headers automatically and replays the cached body on a `304 Not
#' Modified` reply, so the call doesn't count against your GitHub rate
#' limit. Set `options(gh_cache = FALSE)` to disable, or delete the
#' directory above to clear it.
#'
#' If you pass `If-None-Match` (or `If-Modified-Since`) yourself via
#' `.send_headers`, httr2's cache does not intervene: a `304` response
#' is returned as an empty `gh_response` with the response headers
#' (including `ETag`) still attached on `attr(res, "response")`. You are
#' responsible for holding on to the previous body and matching it
#' against the ETag.
#'
#' @export
#' @seealso [gh_gql()] if you want to use the GitHub GraphQL API,
#' [gh_whoami()] for details on GitHub API token management.
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' ## Repositories of a user, these are equivalent
#' gh("/users/hadley/repos", .limit = 2)
#' gh("/users/{username}/repos", username = "hadley", .limit = 2)
#'
#' ## Starred repositories of a user
#' gh("/users/hadley/starred", .limit = 2)
#' gh("/users/{username}/starred", username = "hadley", .limit = 2)
#' @examplesIf FALSE
#' ## Create a repository, needs a token (see gh_token())
#' gh("POST /user/repos", name = "foobar")
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' ## Issues of a repository
#' gh("/repos/hadley/dplyr/issues")
#' gh("/repos/{owner}/{repo}/issues", owner = "hadley", repo = "dplyr")
#'
#' ## Automatic pagination
#' users <- gh("/users", .limit = 50)
#' length(users)
#' @examplesIf FALSE
#' ## Access developer preview of Licenses API (in preview as of 2015-09-24)
#' gh("/licenses") # used to error code 415
#' gh("/licenses", .accept = "application/vnd.github.drax-preview+json")
#' @examplesIf FALSE
#' ## Access Github Enterprise API
#' ## Use GITHUB_API_URL environment variable to change the default.
#' gh("/user/repos", type = "public", .api_url = "https://github.foobar.edu/api/v3")
#' @examplesIf FALSE
#' ## Use I() to force body part to be sent as an array, even if length 1
#' ## This works whether assignees has length 1 or > 1
#' assignees <- "gh_user"
#' assignees <- c("gh_user1", "gh_user2")
#' gh("PATCH /repos/OWNER/REPO/issues/1", assignees = I(assignees))
#' @examplesIf FALSE
#' ## There are two ways to send JSON data. One is that you supply one or
#' ## more objects that will be converted to JSON automatically via
#' ## jsonlite::toJSON(). In this case sometimes you need to use
#' ## jsonlite::unbox() because fromJSON() creates lists from scalar vectors
#' ## by default. The Content-Type header is automatically added in this
#' ## case. For example this request turns on GitHub Pages, using this
#' ## API: https://docs.github.com/v3/repos/pages/#enable-a-pages-site
#'
#' gh::gh(
#'   "POST /repos/{owner}/{repo}/pages",
#'   owner = "r-lib",
#'   repo = "gh",
#'   source = list(
#'     branch = jsonlite::unbox("gh-pages"),
#'     path = jsonlite::unbox("/")
#'   ),
#'   .send_headers = c(Accept = "application/vnd.github.switcheroo-preview+json")
#' )
#'
#' ## The second way is to handle the JSON encoding manually, and supply it
#' ## as a raw vector in an unnamed argument, and also a Content-Type header:
#'
#' body <- '{ "source": { "branch": "gh-pages", "path": "/" } }'
#' gh::gh(
#'   "POST /repos/{owner}/{repo}/pages",
#'   owner = "r-lib",
#'   repo = "gh",
#'   charToRaw(body),
#'   .send_headers = c(
#'     Accept = "application/vnd.github.switcheroo-preview+json",
#'     "Content-Type" = "application/json"
#'   )
#' )
#' @examplesIf FALSE
#' ## Pass along a query to the search/code endpoint via the ... argument
#' x <- gh::gh(
#'             "/search/code",
#'             q = "installation repo:r-lib/gh",
#'             .send_headers = c("X-GitHub-Api-Version" = "2022-11-28")
#'             )
#'  str(x, list.len = 3, give.attr = FALSE)
#'
#'
gh <- function(
  endpoint,
  ...,
  per_page = NULL,
  .per_page = NULL,
  .token = NULL,
  .destfile = NULL,
  .overwrite = FALSE,
  .api_url = NULL,
  .method = "GET",
  .limit = NULL,
  .accept = "application/vnd.github.v3+json",
  .send_headers = NULL,
  .progress = TRUE,
  .params = list(),
  .max_wait = 600,
  .max_rate = NULL
) {
  params <- .parse_params(..., .params = .params)

  check_exclusive(per_page, .per_page, .require = FALSE)
  per_page <- per_page %||% .per_page
  if (is.null(per_page) && !is.null(.limit)) {
    per_page <- max(min(.limit, 100), 1)
  }
  if (!is.null(per_page)) {
    params <- c(params, list(per_page = per_page))
  }

  req <- gh_build_request(
    endpoint = endpoint,
    params = params,
    token = .token,
    destfile = .destfile,
    overwrite = .overwrite,
    accept = .accept,
    send_headers = .send_headers,
    max_wait = .max_wait,
    max_rate = .max_rate,
    api_url = .api_url,
    method = .method
  )

  if (req$method == "GET") {
    check_named_nas(params)
  }

  if (is.null(.limit)) {
    raw <- gh_make_request(req)
    return(gh_process_response(raw, req))
  }

  res <- gh_paginate(req, .limit, per_page, .progress)
  len <- attr(res, "gh_pagination_length")
  attr(res, "gh_pagination_length") <- NULL

  # We only subset for a non-named response.
  if (
    len > .limit &&
      !"total_count" %in% names(res) &&
      length(res) == len
  ) {
    res_attr <- attributes(res)
    res <- res[seq_len(.limit)]
    attributes(res) <- res_attr
  }

  res
}

gh_paginate <- function(
  gh_req,
  .limit,
  per_page,
  .progress,
  error_call = caller_env()
) {
  httr2_req <- gh_build_httr2_request(gh_req)
  limit <- .limit
  max_reqs <- if (is.infinite(limit)) Inf else ceiling(limit / per_page)
  next_url <- httr2::iterate_with_link_url(rel = "next")

  page <- 0L
  n_items <- 0L
  # Totals are known up front when `.limit` is finite; otherwise NA until
  # we (possibly) learn them from the "last" Link header on page 1.
  display_total <- if (is.finite(limit)) limit else NA_integer_
  display_pages <- if (is.finite(max_reqs)) max_reqs else NA_integer_

  # Progress bar lifecycle: always open with an initial spinner. After the
  # first response we tear it down and re-open with one of two formats:
  #   - "determinate": filled bar + % + counts + ETA  (totals known)
  #   - "indeterminate": spinner + counts + elapsed   (totals unknown,
  #                      e.g. cursor-paginated endpoints with no rel=last)
  fmt_init <- "{cli::pb_spin} Fetching... | {cli::pb_elapsed}"
  fmt_determinate <- paste(
    "{cli::pb_bar} {cli::pb_percent} |",
    "{n_items}/{display_total} items, page {page}/{display_pages} |",
    "ETA {cli::pb_eta}"
  )
  fmt_indeterminate <-
    "{cli::pb_spin} {n_items} items, page {page} | {cli::pb_elapsed}"

  if (isTRUE(.progress)) {
    cli::cli_progress_bar(
      format = fmt_init,
      clear = TRUE,
      .envir = environment()
    )
  }

  res <- NULL
  cur_req <- httr2_req
  tryCatch(
    repeat {
      page <- page + 1L
      resp <- httr2::req_perform(cur_req)
      if (httr2::resp_status(resp) >= 400) {
        gh_error(resp, gh_req = gh_req, error_call = error_call)
      }
      res2 <- gh_process_response(resp, gh_req)
      n_items <- n_items + gh_response_length(res2)
      res <- if (is.null(res)) res2 else gh_merge_pages(res, res2)

      # After page 1: discover total from the "last" link if available, then
      # swap the initial spinner for the appropriate final progress bar.
      if (page == 1L) {
        last_url <- httr2::resp_link_url(resp, "last")
        if (!is.null(last_url)) {
          last_page <- as.integer(httr2::url_parse(last_url)$query$page)
          if (!is.na(last_page)) {
            capped <- cap_display_totals(
              display_total,
              display_pages,
              last_page,
              per_page
            )
            display_pages <- capped$pages
            display_total <- capped$total
          }
        }
        if (isTRUE(.progress)) {
          cli::cli_progress_done()
          cli::cli_progress_bar(
            total = if (is.na(display_total)) NA else display_total,
            format = if (is.na(display_total)) {
              fmt_indeterminate
            } else {
              fmt_determinate
            },
            clear = TRUE,
            .envir = environment()
          )
        }
      }

      if (isTRUE(.progress)) {
        cli::cli_progress_update(
          set = if (is.na(display_total)) NULL else min(n_items, display_total),
          force = TRUE
        )
      }

      if (page >= max_reqs) {
        break
      }
      nxt <- next_url(resp, cur_req)
      if (is.null(nxt)) {
        break
      }
      cur_req <- nxt
    },
    interrupt = function(e) {
      if (isTRUE(.progress)) {
        cli::cli_progress_done() # nocov
      }
      cond <- structure(
        class = c("gh_interrupt", "interrupt", "condition"),
        list(
          message = cli::format_inline(
            "{.fn gh} interrupted after fetching {n_items} record{?s}."
          ),
          call = error_call,
          gh_result = res
        )
      )
      withRestarts(
        {
          signalCondition(cond)
          # No exiting handler claimed it. Stash for `rlang::last_error()`
          # recovery, then propagate as a real interrupt.
          asNamespace("rlang")$poke_last_error(cond) # nocov
          # nocov start
          cli::cli_inform(
            c(
              "!" = cond$message,
              "i" = paste(
                "Partial results are available in",
                "{.code rlang::last_error()$gh_result}."
              )
            )
          )
          # nocov end
          rlang::interrupt() # nocov
        },
        muffle_gh_interrupt = function() invisible(NULL)
      )
    }
  )

  attr(res, "gh_pagination_length") <- n_items
  res
}

cap_display_totals <- function(
  display_total,
  display_pages,
  last_page,
  per_page
) {
  list(
    pages = min(display_pages, last_page, na.rm = TRUE),
    total = min(display_total, last_page * per_page, na.rm = TRUE)
  )
}

gh_merge_pages <- function(res, res2) {
  if (!is.null(names(res2)) && identical(names(res), names(res2))) {
    out <- mapply(
      # Handle named array case (e.g. GET /search/repositories)
      function(x, y, n) {
        z <- c(x, y)
        atm <- is.atomic(z)
        if (atm && n %in% c("total_count", "incomplete_results")) {
          y
        } else if (atm) {
          unique(z)
        } else {
          z
        }
      },
      res,
      res2,
      names(res),
      SIMPLIFY = FALSE
    )
  } else {
    # Handle unnamed array case (e.g. GET /orgs/:org/invitations)
    out <- c(res, res2)
  }
  attributes(out) <- attributes(res2)
  out
}

gh_response_length <- function(res) {
  if (
    !is.null(names(res)) && length(res) > 1 && names(res)[1] == "total_count"
  ) {
    # Ignore total_count, incomplete_results, repository_selection
    # and take the first list element to get the length
    lst <- vapply(res, is.list, logical(1))
    nm <- setdiff(
      names(res),
      c("total_count", "incomplete_results", "repository_selection")
    )
    tgt <- which(lst[nm])[1]
    if (is.na(tgt)) length(res) else length(res[[nm[tgt]]])
  } else {
    length(res)
  }
}

gh_build_httr2_request <- function(x) {
  if (!x$method %in% c("GET", "POST", "PATCH", "PUT", "DELETE")) {
    cli::cli_abort("Unknown HTTP verb: {.val {x$method}}")
  }

  req <- httr2::request(x$url)
  req <- httr2::req_method(req, x$method)
  req <- httr2::req_url_query(req, !!!x$query)

  if (!is.null((x$body))) {
    if (is.raw(x$body)) {
      req <- httr2::req_body_raw(req, x$body)
    } else {
      req <- httr2::req_body_json(req, x$body, null = "list", digits = 4)
    }
  }
  req <- httr2::req_headers(req, !!!x$headers)

  # Reduce connection timeout from curl's 10s default to 5s
  req <- httr2::req_options(req, connecttimeout_ms = 5000)
  if (Sys.getenv("GH_FORCE_HTTP_1_1") == "true") {
    req <- httr2::req_options(req, http_version = 2)
  }

  if (!isFALSE(getOption("gh_cache"))) {
    req <- httr2::req_cache(
      req,
      max_size = 100 * 1024 * 1024, # 100 MB
      path = tools::R_user_dir("gh", "cache")
    )
  }

  req <- httr2::req_retry(
    req,
    max_tries = 3,
    is_transient = function(resp) github_is_transient(resp, x$max_wait),
    after = github_after
  )

  if (!is.null(x$max_rate)) {
    req <- httr2::req_throttle(req, x$max_rate)
  }

  # allow custom handling with gh_error
  req <- httr2::req_error(req, is_error = function(resp) FALSE)

  req
}

gh_make_request <- function(x, error_call = caller_env()) {
  req <- gh_build_httr2_request(x)
  resp <- httr2::req_perform(req, path = x$desttmp)
  if (httr2::resp_status(resp) >= 400) {
    gh_error(resp, gh_req = x, error_call = error_call)
  }

  resp
}

# https://docs.github.com/v3/#client-errors
gh_error <- function(response, gh_req, error_call = caller_env()) {
  heads <- httr2::resp_headers(response)
  res <- httr2::resp_body_json(response)
  status <- httr2::resp_status(response)
  if (!is.null(gh_req$desttmp)) {
    unlink(gh_req$desttmp)
  }

  msg <- "GitHub API error ({status}): {heads$status %||% ''} {res$message}"

  if (status == 404) {
    msg <- c(msg, x = c("URL not found: {.url {response$url}}"))
  }

  doc_url <- res$documentation_url
  if (!is.null(doc_url)) {
    msg <- c(msg, c("i" = "Read more at {.url {doc_url}}"))
  }

  errors <- res$errors
  if (!is.null(errors)) {
    # GitHub usually returns `errors` as an array of objects, but some 422s
    # (e.g. exceeding the max statuses for a (sha, context) pair) return it
    # as a plain string. Wrap any non-list entries in a `message` field.
    if (!is.list(errors)) {
      errors <- as.list(errors)
    }
    errors <- lapply(errors, function(e) {
      if (is.list(e)) e else list(message = as.character(e))
    })
    errors <- as.data.frame(do.call(rbind, errors))
    nms <- c("resource", "field", "code", "message")
    nms <- nms[nms %in% names(errors)]
    msg <- c(
      msg,
      capture.output(print(errors[nms], row.names = FALSE))
    )
  }

  cli::cli_abort(
    msg,
    class = c("github_error", paste0("http_error_", status)),
    call = error_call,
    response_headers = heads,
    response_content = res
  )
}


# use retry-after info when possible
# https://docs.github.com/en/rest/overview/resources-in-the-rest-api#exceeding-the-rate-limit
github_is_transient <- function(resp, max_wait) {
  if (httr2::resp_status(resp) != 403) {
    return(FALSE)
  }
  if (!identical(httr2::resp_header(resp, "x-ratelimit-remaining"), "0")) {
    return(FALSE)
  }

  time <- httr2::resp_header(resp, "x-ratelimit-reset")
  if (is.null(time)) {
    return(FALSE)
  }

  time <- as.numeric(time)
  minutes_to_wait <- (time - unclass(Sys.time()))
  minutes_to_wait <= max_wait
}
github_after <- function(resp) {
  time <- as.numeric(httr2::resp_header(resp, "x-ratelimit-reset"))
  time - unclass(Sys.time())
}

test_that("generates a useful message", {
  local_fake_github()
  expect_snapshot(
    gh("/missing"),
    error = TRUE,
    transform = redact_fake_host
  )
})

test_that("errors return a github_error object", {
  local_fake_github()
  e <- tryCatch(gh("/missing"), error = identity)

  expect_s3_class(e, "github_error")
  expect_s3_class(e, "http_error_404")
})

# https://github.com/r-lib/gh/issues/229
test_that("handles 422 responses with `errors` as a plain string", {
  local_fake_github()
  expect_snapshot(
    gh(
      "POST /repos/{owner}/{repo}/statuses/{sha}",
      owner = "r-lib",
      repo = "gh",
      sha = "deadbeef",
      state = "success"
    ),
    error = TRUE,
    transform = redact_fake_host
  )
})

test_that("can catch a given status directly", {
  local_fake_github()
  e <- tryCatch(gh("/missing"), "http_error_404" = identity)

  expect_s3_class(e, "github_error")
  expect_s3_class(e, "http_error_404")
})

test_that("can ignore trailing commas", {
  local_fake_github()
  expect_no_error(gh("/orgs/tidyverse/repos", ))
})

test_that("can use per_page or .per_page but not both", {
  local_fake_github()
  resp <- gh("/orgs/tidyverse/repos", per_page = 2)
  expect_equal(attr(resp, "request")$query$per_page, 2)

  resp <- gh("/orgs/tidyverse/repos", .per_page = 2)
  expect_equal(attr(resp, "request")$query$per_page, 2)

  expect_snapshot(
    error = TRUE,
    gh("/orgs/tidyverse/repos", per_page = 1, .per_page = 2)
  )
})

test_that("can paginate", {
  local_fake_github()
  pages <- gh(
    "/orgs/tidyverse/repos",
    per_page = 1,
    .limit = 5,
    .progress = FALSE
  )
  expect_length(pages, 5)
})

test_that("trim output when .limit isn't a multiple of .per_page", {
  local_fake_github()
  pages <- gh(
    "/orgs/tidyverse/repos",
    per_page = 2,
    .limit = 3,
    .progress = FALSE
  )
  expect_length(pages, 3)
})

test_that("can paginate repository search", {
  local_fake_github()
  pages <- gh(
    "/search/repositories",
    q = "tidyverse",
    per_page = 10,
    .limit = 35
  )
  expect_named(pages, c("total_count", "incomplete_results", "items"))
  # Items aren't trimmed to .limit in this case
  expect_length(pages$items, 40)
})

test_that("paginated request surfaces a 4xx error", {
  local_fake_github()
  expect_snapshot(
    error = TRUE,
    gh("/orgs/{org}/repos", org = "gh-org-testing-404", .limit = 100),
    transform = redact_fake_host
  )
})

test_that("paginate exhausts when .limit is Inf and learns total from last link", {
  local_fake_github()
  pages <- gh(
    "/orgs/r-lib/repos",
    per_page = 5,
    .limit = Inf,
    .progress = FALSE
  )
  expect_length(pages, 16)
})

test_that("POST accepts a raw body", {
  local_fake_github()
  payload <- charToRaw("hello, raw world")
  res <- gh("POST /echo-raw", payload)
  expect_equal(as.character(class(res))[1], "gh_response")
})

test_that("GH_FORCE_HTTP_1_1 envvar is honored", {
  local_fake_github()
  withr::with_envvar(c(GH_FORCE_HTTP_1_1 = "true"), {
    res <- gh("/orgs/{org}/repos", org = "r-lib", per_page = 1)
  })
  expect_s3_class(res, "gh_response")
})

test_that("requests use the on-disk cache when gh_cache is not FALSE", {
  local_fake_github()
  withr::local_options(gh_cache = NULL)
  res <- gh("/orgs/{org}/repos", org = "r-lib", per_page = 1)
  expect_s3_class(res, "gh_response")
})

test_that("cursor-style pagination uses the indeterminate progress format", {
  local_fake_github()
  res <- gh("/cursor-list", per_page = 10, .limit = Inf, .progress = TRUE)
  expect_length(res, 25)
})

test_that("github_is_transient identifies rate-limit 403s", {
  fake_resp <- function(status, headers = list()) {
    structure(
      list(status_code = status, headers = headers, url = "x", method = "GET"),
      class = "httr2_response"
    )
  }
  reset <- as.character(as.integer(unclass(Sys.time())) + 5L)
  far_reset <- as.character(as.integer(unclass(Sys.time())) + 86400L)

  # Not a 403
  expect_false(github_is_transient(fake_resp(200), max_wait = 60))
  # 403 but plenty of remaining quota
  expect_false(github_is_transient(
    fake_resp(403, list("x-ratelimit-remaining" = "100")),
    max_wait = 60
  ))
  # 403, no remaining quota, missing reset header
  expect_false(github_is_transient(
    fake_resp(403, list("x-ratelimit-remaining" = "0")),
    max_wait = 60
  ))
  # 403, no remaining, reset is too far away
  expect_false(github_is_transient(
    fake_resp(
      403,
      list("x-ratelimit-remaining" = "0", "x-ratelimit-reset" = far_reset)
    ),
    max_wait = 60
  ))
  # 403, no remaining, reset within max_wait
  expect_true(github_is_transient(
    fake_resp(
      403,
      list("x-ratelimit-remaining" = "0", "x-ratelimit-reset" = reset)
    ),
    max_wait = 60
  ))
})

test_that("github_after returns the time until the rate limit resets", {
  reset <- as.integer(unclass(Sys.time())) + 30L
  fake_resp <- structure(
    list(
      status_code = 403,
      headers = list("x-ratelimit-reset" = as.character(reset)),
      url = "x",
      method = "GET"
    ),
    class = "httr2_response"
  )
  expect_lte(github_after(fake_resp), 30)
  expect_gte(github_after(fake_resp), 25)
})

test_that("gh_merge_pages dedupes named atomic fields that are not search metadata", {
  res1 <- list(repository_selection = "all", items = list("a"))
  res2 <- list(repository_selection = "all", items = list("b"))
  merged <- gh_merge_pages(res1, res2)
  expect_equal(merged$repository_selection, "all")
  expect_equal(merged$items, list("a", "b"))
})

## Main API URL
default_api_url <- function() {
  Sys.getenv("GITHUB_API_URL", unset = "https://api.github.com")
}

## Headers to send with each API request
default_send_headers <- c("User-Agent" = "https://github.com/r-lib/gh")

gh_build_request <- function(
  endpoint = "/user",
  params = list(),
  token = NULL,
  destfile = NULL,
  overwrite = NULL,
  accept = NULL,
  send_headers = NULL,
  max_wait = 10,
  max_rate = NULL,
  api_url = NULL,
  method = "GET"
) {
  working <- list(
    method = method,
    url = character(),
    headers = NULL,
    query = NULL,
    body = NULL,
    endpoint = endpoint,
    params = params,
    token = token,
    accept = c(Accept = accept),
    send_headers = send_headers,
    api_url = api_url,
    dest = destfile,
    overwrite = overwrite,
    max_wait = max_wait,
    max_rate = max_rate
  )

  working <- gh_set_verb(working)
  working <- gh_set_endpoint(working)
  working <- gh_set_query(working)
  working <- gh_set_body(working)
  working <- gh_set_url(working)
  working <- gh_set_headers(working)
  working <- gh_set_temp_destfile(working)
  working[c(
    "method",
    "url",
    "headers",
    "query",
    "body",
    "dest",
    "desttmp",
    "max_wait",
    "max_rate"
  )]
}


## gh_set_*(x)
## x = a list in which we build up an httr2 request
## x goes in, x comes out, possibly modified

gh_set_verb <- function(x) {
  if (!nzchar(x$endpoint)) {
    return(x)
  }

  # No method defined, so use default
  if (grepl("^/", x$endpoint) || grepl("^http", x$endpoint)) {
    return(x)
  }

  # Method can be lower-case (e.g. copy-pasting from API docs in Firefox)
  method <- gsub("^([^/ ]+)\\s+.*$", "\\1", x$endpoint)
  x$endpoint <- gsub(sprintf("^%s+ ", method), "", x$endpoint)
  # Now switch method to upper-case
  x$method <- toupper(method)
  x
}

gh_set_endpoint <- function(x) {
  params <- x$params
  if (
    !is_template(x$endpoint) || length(params) == 0L || has_no_names(params)
  ) {
    return(x)
  }

  named_params <- which(has_name(params))
  done <- rep_len(FALSE, length(params))
  endpoint <- endpoint2 <- x$endpoint

  for (i in named_params) {
    endpoint2 <- expand_variable(
      varname = names(params)[i],
      value = params[[i]][1],
      template = endpoint
    )
    if (is.na(endpoint2)) {
      cli::cli_abort(
        "Named NA parameters are not allowed: {names(params)[i]}"
      )
    }
    if (endpoint2 != endpoint) {
      endpoint <- endpoint2
      done[i] <- TRUE
    }
    if (!is_template(endpoint)) {
      break
    }
  }

  x$endpoint <- endpoint
  x$params <- x$params[!done]
  x$params <- cleanse_names(x$params)
  x
}

gh_set_query <- function(x) {
  params <- x$params
  if (x$method != "GET" || length(params) == 0L) {
    return(x)
  }
  if (!all(has_name(params))) {
    cli::cli_abort(
      "All elements of {.arg params} must be named for {.code GET} requests."
    )
  }
  x$query <- params
  x$params <- NULL
  x
}

gh_set_body <- function(x) {
  if (length(x$params) == 0L) {
    return(x)
  }
  if (x$method == "GET") {
    warning("This is a 'GET' request and unnamed parameters are being ignored.")
    return(x)
  }
  if (length(x$params) == 1 && is.raw(x$params[[1]])) {
    x$body <- x$params[[1]]
  } else {
    x$body <- x$params
  }
  x
}

gh_set_url <- function(x) {
  if (grepl("^https?://", x$endpoint)) {
    x$url <- URLencode(x$endpoint)
    x$api_url <- get_baseurl(x$url)
  } else {
    x$api_url <- get_apiurl(x$api_url %||% default_api_url())
    x$url <- URLencode(paste0(x$api_url, x$endpoint))
  }

  x
}

gh_set_temp_destfile <- function(working) {
  working$desttmp <- if (is.null(working$dest)) {
    NULL
  } else {
    paste0(working$dest, "-", basename(tempfile("")), ".gh-tmp")
  }
  working
}

get_baseurl <- function(url) {
  # https://github.uni.edu/api/v3/
  if (!any(grepl("^https?://", url))) {
    cli::cli_abort("Only works with HTTP(S) protocols")
  }
  prot <- sub("^(https?://).*$", "\\1", url) # https://
  rest <- sub("^https?://(.*)$", "\\1", url) #         github.uni.edu/api/v3/
  host <- sub("/.*$", "", rest) #         github.uni.edu
  paste0(prot, host) # https://github.uni.edu
}

# https://api.github.com --> https://github.com
# api.github.com --> github.com
normalize_host <- function(x) {
  sub("api[.]github[.]com", "github.com", x)
}

get_hosturl <- function(url) {
  url <- get_baseurl(url)
  normalize_host(url)
}

# (almost) the inverse of get_hosturl()
# https://github.com     --> https://api.github.com
# https://github.uni.edu --> https://github.uni.edu/api/v3
get_apiurl <- function(url) {
  host_url <- get_hosturl(url)
  prot_host <- strsplit(host_url, "://", fixed = TRUE)[[1]]
  if (is_github_dot_com(host_url)) {
    paste0(prot_host[[1]], "://api.github.com")
  } else {
    paste0(host_url, "/api/v3")
  }
}

is_github_dot_com <- function(url) {
  url <- get_baseurl(url)
  url <- normalize_host(url)
  grepl("^https?://github.com", url)
}

gh_set_headers <- function(x) {
  # x$api_url must be set properly at this point
  auth <- gh_auth(x$token %||% gh_token(x$api_url))
  send_headers <- gh_send_headers(x$accept, x$send_headers)
  x$headers <- c(send_headers, auth)
  x
}

gh_send_headers <- function(accept_header = NULL, headers = NULL) {
  modify_vector(
    modify_vector(default_send_headers, accept_header),
    headers
  )
}

# helpers ----
# https://tools.ietf.org/html/rfc6570
# we support what the RFC calls "Level 1 templates", which only require
# simple string expansion of a placeholder consisting of [A-Za-z0-9_]
is_template <- function(x) {
  is_colon_template(x) || is_uri_template(x)
}

is_colon_template <- function(x) grepl(":", x)

is_uri_template <- function(x) grepl("[{]\\w+?[}]", x)

template_type <- function(x) {
  if (is_uri_template(x)) {
    return("uri")
  }
  if (is_colon_template(x)) {
    return("colon")
  }
}

expand_variable <- function(varname, value, template) {
  type <- template_type(template)
  if (is.null(type)) {
    return(template)
  }
  pattern <- switch(
    type,
    uri = paste0("[{]", varname, "[}]"),
    colon = paste0(":", varname, "\\b"),
    cli::cli_abort("Internal error: unrecognized template type")
  )
  gsub(pattern, value, template)
}

test_that("all forms of specifying endpoint are equivalent", {
  r1 <- gh_build_request("GET /rate_limit")
  expect_equal(r1$method, "GET")
  expect_equal(r1$url, "https://api.github.com/rate_limit")

  expect_equal(gh_build_request("/rate_limit"), r1)
  expect_equal(gh_build_request("GET https://api.github.com/rate_limit"), r1)
  expect_equal(gh_build_request("https://api.github.com/rate_limit"), r1)
})

test_that("method arg sets default method", {
  r <- gh_build_request("/rate_limit", method = "POST")
  expect_equal(r$method, "POST")
})

test_that("parameter substitution is equivalent to direct specification (:)", {
  subst <-
    gh_build_request("POST /repos/:org/:repo/issues/:number/labels",
      params = list(
        org = "ORG", repo = "REPO", number = "1",
        "body"
      )
    )
  spec <-
    gh_build_request("POST /repos/ORG/REPO/issues/1/labels",
      params = list("body")
    )
  expect_identical(subst, spec)
})

test_that("parameter substitution is equivalent to direct specification", {
  subst <-
    gh_build_request("POST /repos/{org}/{repo}/issues/{number}/labels",
      params = list(
        org = "ORG", repo = "REPO", number = "1",
        "body"
      )
    )
  spec <-
    gh_build_request("POST /repos/ORG/REPO/issues/1/labels",
      params = list("body")
    )
  expect_identical(subst, spec)
})

test_that("URI templates that need expansion are detected", {
  expect_true(is_uri_template("/orgs/{org}/repos"))
  expect_true(is_uri_template("/repos/{owner}/{repo}"))
  expect_false(is_uri_template("/user/repos"))
})

test_that("older 'colon templates' are detected", {
  expect_true(is_colon_template("/orgs/:org/repos"))
  expect_true(is_colon_template("/repos/:owner/:repo"))
  expect_false(is_colon_template("/user/repos"))
})

test_that("gh_set_endpoint() works", {
  # no expansion, no extra params
  input <- list(endpoint = "/user/repos")
  expect_equal(input, gh_set_endpoint(input))

  # no expansion, with extra params
  input <- list(endpoint = "/user/repos", params = list(page = 2))
  expect_equal(input, gh_set_endpoint(input))

  # expansion, no extra params
  input <- list(
    endpoint = "/repos/{owner}/{repo}",
    params = list(owner = "OWNER", repo = "REPO")
  )
  out <- gh_set_endpoint(input)
  expect_equal(
    out,
    list(endpoint = "/repos/OWNER/REPO", params = list())
  )

  # expansion, with extra params
  input <- list(
    endpoint = "/repos/{owner}/{repo}/issues",
    params = list(state = "open", owner = "OWNER", repo = "REPO", page = 2)
  )
  out <- gh_set_endpoint(input)
  expect_equal(out$endpoint, "/repos/OWNER/REPO/issues")
  expect_equal(out$params, list(state = "open", page = 2))
})

test_that("gh_set_endpoint() refuses to substitute an NA", {
  input <- list(
    endpoint = "POST /orgs/{org}/repos",
    params = list(org = NA)
  )
  expect_error(gh_set_endpoint(input), "Named NA")
})

test_that("gh_set_endpoint() allows a named NA in body for non-GET", {
  input <- list(
    endpoint = "PUT /repos/{owner}/{repo}/pages",
    params = list(owner = "OWNER", repo = "REPO", cname = NA)
  )
  out <- gh_set_endpoint(input)
  expect_equal(out$endpoint, "PUT /repos/OWNER/REPO/pages")
  expect_equal(out$params, list(cname = NA))
})

test_that("gh_set_url() ensures URL is in 'API form'", {
  input <- list(
    endpoint = "/user/repos",
    api_url = "https://github.com"
  )
  out <- gh_set_url(input)
  expect_equal(out$api_url, "https://api.github.com")

  input$api_url <- "https://github.acme.com"
  out <- gh_set_url(input)
  expect_equal(out$api_url, "https://github.acme.com/api/v3")
})

test_that("gh_make_request() errors if unknown verb", {
  expect_snapshot_error(gh("geeet /users/hadley/repos", .limit = 2))
})

#' A simple interface for the GitHub GraphQL API v4.
#'
#' See more about the GraphQL API here:
#' <https://docs.github.com/graphql>
#'
#' Note: pagination and the `.limit` argument does not work currently,
#' as pagination in the GraphQL API is different from the v3 API.
#' If you need pagination with GraphQL, you'll need to do that manually.
#'
#' @inheritParams gh
#' @param query The GraphQL query, as a string.
#' @export
#' @seealso [gh()] for the GitHub v3 API.
#' @examplesIf FALSE
#' gh_gql("query { viewer { login }}")
#'
#' # Get rate limit
#' ratelimit_query <- "query {
#'   viewer {
#'     login
#'   }
#'   rateLimit {
#'     limit
#'     cost
#'     remaining
#'     resetAt
#'   }
#' }"
#'
#' gh_gql(ratelimit_query)
gh_gql <- function(query, ...) {
  if (".limit" %in% names(list(...))) {
    stop("`.limit` does not work with the GraphQL API")
  }

  gh(endpoint = "POST /graphql", query = query, ...)
}

test_that("gh_gql posts to /graphql and returns parsed response", {
  local_fake_github()
  res <- gh_gql("query { viewer { login } }")
  expect_equal(res$data$viewer$login, "fakeuser")
  expect_equal(res$data$echo, "query { viewer { login } }")
})

test_that("gh_gql rejects .limit", {
  expect_snapshot(
    error = TRUE,
    gh_gql("query { viewer { login } }", .limit = 5)
  )
})

#' @keywords internal
#' @aliases gh-package
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @import rlang
#' @importFrom cli cli_status cli_status_update
#' @importFrom glue glue
#' @importFrom jsonlite fromJSON toJSON
#' @importFrom lifecycle deprecated
#' @importFrom utils URLencode capture.output
## usethis namespace: end
NULL

# # Standalone file for better error handling ----------------------------
#
# If can allow package dependencies, then you are probably better off
# using rlang's functions for errors.
#
# The canonical location of this file is in the processx package:
# https://github.com/r-lib/processx/master/R/errors.R
#
# ## Features
#
# - Throw conditions and errors with the same API.
# - Automatically captures the right calls and adds them to the conditions.
# - Sets `.Last.error`, so you can easily inspect the errors, even if they
#   were not caught.
# - It only sets `.Last.error` for the errors that are not caught.
# - Hierarchical errors, to allow higher level error messages, that are
#   more meaningful for the users, while also keeping the lower level
#   details in the error object. (So in `.Last.error` as well.)
# - `.Last.error` always includes a stack trace. (The stack trace is
#   common for the whole error hierarchy.) The trace is accessible within
#   the error, e.g. `.Last.error$trace`. The trace of the last error is
#   also at `.Last.error.trace`.
# - Can merge errors and traces across multiple processes.
# - Pretty-print errors and traces, if the crayon package is loaded.
# - Automatically hides uninformative parts of the stack trace when
#   printing.
#
# ## API
#
# ```
# new_cond(..., call. = TRUE, domain = NULL)
# new_error(..., call. = TRUE, domain = NULL)
# throw(cond, parent = NULL)
# catch_rethrow(expr, ...)
# rethrow(expr, cond)
# rethrow_call(.NAME, ...)
# add_trace_back(cond)
# ```
#
# ## Roadmap:
# - better printing of anonymous function in the trace
#
# ## NEWS:
#
# ### 1.0.0 -- 2019-06-18
#
# * First release.
#
# ### 1.0.1 -- 2019-06-20
#
# * Add `rlib_error_always_trace` option to always add a trace
#
# ### 1.0.2 -- 2019-06-27
#
# * Internal change: change topenv of the functions to baseenv()

err <- local({

  # -- condition constructors -------------------------------------------

  #' Create a new condition
  #'
  #' @noRd
  #' @param ... Parts of the error message, they will be converted to
  #'   character and then concatenated, like in [stop()].
  #' @param call. A call object to include in the condition, or `TRUE`
  #'   or `NULL`, meaning that [throw()] should add a call object
  #'   automatically.
  #' @param domain Translation domain, see [stop()].
  #' @return Condition object. Currently a list, but you should not rely
  #'   on that.

  new_cond <- function(..., call. = TRUE, domain = NULL) {
    message <- .makeMessage(..., domain = domain)
    structure(
      list(message = message, call = call.),
      class = c("condition"))
  }

  #' Create a new error condition
  #'
  #' It also adds the `rlib_error` class.
  #'
  #' @noRd
  #' @param ... Passed to [new_cond()].
  #' @param call. Passed to [new_cond()].
  #' @param domain Passed to [new_cond()].
  #' @return Error condition object with classes `rlib_error`, `error`
  #'   and `condition`.

  new_error <- function(..., call. = TRUE, domain = NULL) {
    cond <- new_cond(..., call. = call., domain = domain)
    class(cond) <- c("rlib_error", "error", "condition")
    cond
  }

  # -- throwing conditions ----------------------------------------------

  #' Throw a condition
  #'
  #' If the condition is an error, it will also call [stop()], after
  #' signalling the condition first. This means that if the condition is
  #' caught by an exiting handler, then [stop()] is not called.
  #'
  #' @noRd
  #' @param cond Condition object to throw. If it is an error condition,
  #'   then it calls [stop()].
  #' @param parent Parent condition. Use this within [rethrow()] and
  #'   [catch_rethrow()].

  throw <- function(cond, parent = NULL) {
    if (!inherits(cond, "condition")) {
      throw(new_error("You can only throw conditions"))
    }
    if (!is.null(parent) && !inherits(parent, "condition")) {
      throw(new_error("Parent condition must be a condition object"))
    }

    if (is.null(cond$call) || isTRUE(cond$call)) {
      cond$call <- sys.call(-1) %||% sys.call()
    }

    # Eventually the nframe numbers will help us print a better trace
    # When a child condition is created, the child will use the parent
    # error object to make note of its own nframe. Here we copy that back
    # to the parent.
    if (is.null(cond$`_nframe`)) cond$`_nframe` <- sys.nframe()
    if (!is.null(parent)) {
      cond$parent <- parent
      cond$call <- cond$parent$`_childcall`
      cond$`_nframe` <- cond$parent$`_childframe`
      cond$`_ignore` <- cond$parent$`_childignore`
    }

    # We can set an option to always add the trace to the thrown
    # conditions. This is useful for example in context that always catch
    # errors, e.g. in testthat tests or knitr. This options is usually not
    # set and we signal the condition here
    always_trace <- isTRUE(getOption("rlib_error_always_trace"))
    if (!always_trace) signalCondition(cond)

    # If this is not an error, then we'll just return here. This allows
    # throwing interrupt conditions for example, with the same UI.
    if (! inherits(cond, "error")) return(invisible())

    if (is.null(cond$`_pid`)) cond$`_pid` <- Sys.getpid()
    if (is.null(cond$`_timestamp`)) cond$`_timestamp` <- Sys.time()

    # If we get here that means that the condition was not caught by
    # an exiting handler. That means that we need to create a trace.
    cond <- add_trace_back(cond)

    # Set up environment to store .Last.error, it will be just before
    # baseenv(), so it is almost as if it was in baseenv() itself, like
    # .Last.value. We save the print methos here as well, and then they
    # will be found automatically.
    if (! "org:r-lib" %in% search()) {
      do.call("attach", list(new.env(), pos = length(search()),
                             name = "org:r-lib"))
    }
    env <- as.environment("org:r-lib")
    env$print.rlib_error <- print_rlib_error
    env$print.rlib_trace <- print_rlib_trace
    env$.Last.error <- cond
    env$.Last.error.trace <- cond$trace

    # If we always wanted a trace, then we signal the condition here
    if (always_trace) signalCondition(cond)

    # Top-level handler, this is intended for testing only for now,
    # and its design might change.
    if (!is.null(th <- getOption("rlib_error_handler")) &&
        is.function(th)) {
      th(cond)

    } else {
      # Dropping the classes and adding "duplicate_condition" is a workaround
      # for the case when we have non-exiting handlers on throw()-n
      # conditions. These would get the condition twice, because stop()
      # will also signal it. If we drop the classes, then only handlers
      # on "condition" objects (i.e. all conditions) get duplicate signals.
      # This is probably quite rare, but for this rare case they can also
      # recognize the duplicates from the "duplicate_condition" extra class.
      class(cond) <- c("duplicate_condition", "condition")
      stop(cond)
    }
  }

  # -- rethrowing conditions --------------------------------------------

  #' Catch and re-throw conditions
  #'
  #' See [rethrow()] for a simpler interface that handles `error`
  #' conditions automatically.
  #'
  #' @noRd
  #' @param expr Expression to evaluate.
  #' @param ... Condition handler specification, the same way as in
  #'   [withCallingHandlers()]. You are supposed to call [throw()] from
  #'   the error handler, with a new error object, setting the original
  #'   error object as parent. See examples below.
  #' @examples
  #' f <- function() {
  #'   ...
  #'   err$catch_rethrow(
  #'     ... code that potentially errors ...,
  #'     error = function(e) {
  #'       throw(new_error("This will be the child error"), parent = e)
  #'     }
  #'   )
  #' }

  catch_rethrow <- function(expr, ...) {
    realcall <- sys.call(-1) %||% sys.call()
    realframe <- sys.nframe()
    parent <- parent.frame()

    cl <- match.call()
    cl[[1]] <- quote(withCallingHandlers)
    handlers <- list(...)
    for (h in names(handlers)) {
      cl[[h]] <- function(e) {
        # This will be NULL if the error is not throw()-n
        if (is.null(e$`_nframe`)) e$`_nframe` <- sys.parent()
        e$`_childcall` <- realcall
        e$`_childframe` <- realframe
        # We drop after realframe, until the first withCallingHandlers
        wch <- find_call(sys.calls(), quote(withCallingHandlers))
        if (!is.na(wch)) e$`_childignore` <- list(c(realframe + 1L, wch))
        handlers[[h]](e)
      }
    }
    eval(cl, envir = parent)
  }

  find_call <- function(calls, call) {
    which(vapply(
      calls, function(x) length(x) >= 1 && identical(x[[1]], call),
      logical(1)))[1]
  }

  #' Catch and re-throw conditions
  #'
  #' `rethrow()` is similar to [catch_rethrow()], but it has a simpler
  #' interface. It catches conditions with class `error`, and re-throws
  #' `cond` instead, using the original condition as the parent.
  #'
  #' @noRd
  #' @param expr Expression to evaluate.
  #' @param ... Condition handler specification, the same way as in
  #'   [withCallingHandlers()].

  rethrow <- function(expr, cond) {
    realcall <- sys.call(-1) %||% sys.call()
    realframe <- sys.nframe()
    withCallingHandlers(
      expr,
      error = function(e) {
        # This will be NULL if the error is not throw()-n
        if (is.null(e$`_nframe`)) e$`_nframe` <- sys.parent()
        e$`_childcall` <- realcall
        e$`_childframe` <- realframe
        # We just ignore the withCallingHandlers call, and the tail
        e$`_childignore` <- list(
          c(realframe + 1L, realframe + 1L),
          c(e$`_nframe` + 1L, sys.nframe() + 1L))
        throw(cond, parent = e)
      }
    )
  }

  #' Version of .Call that throw()s errors
  #'
  #' It re-throws error from interpreted code. If the error had class
  #' `simpleError`, like all errors, thrown via `error()` in C do, it also
  #' adds the `c_error` class.
  #'
  #' @noRd
  #' @param .NAME Compiled function to call, see [.Call()].
  #' @param ... Function arguments, see [.Call()].
  #' @return Result of the call.

  rethrow_call <- function(.NAME, ...) {
    call <- sys.call()
    nframe <- sys.nframe()
    withCallingHandlers(
      # do.call to work around an R CMD check issue
      do.call(".Call", list(.NAME, ...)),
      error = function(e) {
        e$`_nframe` <- nframe
        e$call <- call
        if (inherits(e, "simpleError")) {
          class(e) <- c("c_error", "rlib_error", "error", "condition")
        }
        e$`_ignore` <- list(c(nframe + 1L, sys.nframe() + 1L))
        throw(e)
      }
    )
  }

  # -- create traceback -------------------------------------------------

  #' Create a traceback
  #'
  #' [throw()] calls this function automatically if an error is not caught,
  #' so there is currently not much use to call it directly.
  #'
  #' @param cond Condition to add the trace to
  #'
  #' @return A condition object, with the trace added.

  add_trace_back <- function(cond) {
    idx <- seq_len(sys.parent(1L))
    frames <- sys.frames()[idx]

    parents <- sys.parents()[idx]
    calls <- as.list(sys.calls()[idx])
    envs <- lapply(frames, env_label)
    topenvs <- lapply(
      seq_along(frames),
      function(i) env_label(topenv(environment(sys.function(i)))))
    nframes <- if (!is.null(cond$`_nframe`)) cond$`_nframe` else sys.parent()
    messages <- list(conditionMessage(cond))
    ignore <- cond$`_ignore`
    classes <- class(cond)
    pids <- rep(cond$`_pid` %||% Sys.getpid(), length(calls))

    if (is.null(cond$parent)) {
      # Nothing to do, no parent

    } else if (is.null(cond$parent$trace)) {
      # If the parent does not have a trace, that means that it is using
      # the same trace as us.
      parent <- cond
      while (!is.null(parent <- parent$parent)) {
        nframes <- c(nframes, parent$`_nframe`)
        messages <- c(messages, list(conditionMessage(parent)))
        ignore <- c(ignore, parent$`_ignore`)
      }

    } else {
      # If it has a trace, that means that it is coming from another
      # process or top level evaluation. In this case we'll merge the two
      # traces.
      pt <- cond$parent$trace
      parents <- c(parents, pt$parents + length(calls))
      nframes <- c(nframes, pt$nframes + length(calls))
      ignore <- c(ignore, lapply(pt$ignore, function(x) x + length(calls)))
      envs <- c(envs, pt$envs)
      topenvs <- c(topenvs, pt$topenvs)
      calls <- c(calls, pt$calls)
      messages <- c(messages, pt$messages)
      pids <- c(pids, pt$pids)
    }

    cond$trace <- new_trace(
      calls, parents, envs, topenvs, nframes, messages, ignore, classes,
      pids)

    cond
  }

  new_trace <- function (calls, parents, envs, topenvs, nframes, messages,
                         ignore, classes, pids) {
    indices <- seq_along(calls)
    structure(
      list(calls = calls, parents = parents, envs = envs, topenvs = topenvs,
           indices = indices, nframes = nframes, messages = messages,
           ignore = ignore, classes = classes, pids = pids),
      class = "rlib_trace")
  }

  env_label <- function(env) {
    nm <- env_name(env)
    if (nzchar(nm)) {
      nm
    } else {
      env_address(env)
    }
  }

  env_address <- function(env) {
    class(env) <- "environment"
    sub("^.*(0x[0-9a-f]+)>$", "\\1", format(env), perl = TRUE)
  }

  env_name <- function(env) {
    if (identical(env, globalenv())) {
      return("global")
    }
    if (identical(env, baseenv())) {
      return("namespace:base")
    }
    if (identical(env, emptyenv())) {
      return("empty")
    }
    nm <- environmentName(env)
    if (isNamespace(env)) {
      return(paste0("namespace:", nm))
    }
    nm
  }

  # -- printing ---------------------------------------------------------

  print_rlib_error <- function(x, ...) {

    msg <- conditionMessage(x)
    call <- conditionCall(x)
    cl <- class(x)[1L]
    if (!is.null(call)) {
      cat("<", cl, " in ", format_call(call), ":\n ", msg, ">\n", sep = "")
    } else {
      cat("<", cl, ": ", msg, ">\n", sep = "")
    }

    print_srcref(x$call)

    if (!identical(x$`_pid`, Sys.getpid())) {
      cat(" in process", x$`_pid`, "\n")
    }

    if (!is.null(x$parent)) {
      cat("-->\n")
      print(x$parent)
    }

    invisible(x)
  }

  print_rlib_trace <- function(x, ...) {
    cl <- setdiff(x$classes, c("error", "condition"))
    cl <- paste0(" ERROR TRACE for ", paste(cl, collapse = ", "), "")
    cat(sep = "", "\n", style_trace_title(cl), "\n\n")
    calls <- map2(x$calls, x$topenv, namespace_calls)
    callstr <- vapply(calls, format_call_src, character(1))
    callstr[x$nframes] <-
      paste0(callstr[x$nframes], "\n", style_error(x$messages), "\n")
    callstr <- enumerate(callstr)

    # Ignore what we were told to ignore
    ign <- integer()
    for (iv in x$ignore) {
      if (iv[2] == Inf) iv[2] <- length(callstr)
      ign <- c(ign, iv[1]:iv[2])
    }

    # Plus always ignore the tail. This is not always good for
    # catch_rethrow(), but should be good otherwise
    last_err_frame <- x$nframes[length(x$nframes)]
    if (!is.na(last_err_frame) && last_err_frame < length(callstr)) {
      ign <- c(ign, (last_err_frame+1):length(callstr))
    }

    ign <- unique(ign)
    if (length(ign)) callstr <- callstr[-ign]

    # Add markers for subprocesses
    if (length(unique(x$pids)) >= 2) {
      pids <- x$pids[-ign]
      pid_add <- which(!duplicated(pids))
      pid_str <- style_process(paste0("Process ", pids[pid_add], ":"))
      callstr[pid_add] <- paste0(" ", pid_str, "\n", callstr[pid_add])
    }

    cat(callstr, sep = "\n")
    invisible(x)
  }

  namespace_calls <- function(call, env) {
    if (length(call) < 1) return(call)
    if (typeof(call[[1]]) != "symbol") return(call)
    pkg <- strsplit(env, "^namespace:")[[1]][2]
    if (is.na(pkg)) return(call)
    call[[1]] <- substitute(p:::f, list(p = as.symbol(pkg), f = call[[1]]))
    call
  }

  print_srcref <- function(call) {
    src <- format_srcref(call)
    if (length(src)) cat(sep = "", " ", src, "\n")
  }

  `%||%` <- function(l, r) if (is.null(l)) r else l

  format_srcref <- function(call) {
    if (is.null(call)) return(NULL)
    file <- utils::getSrcFilename(call)
    if (!length(file)) return(NULL)
    dir <- utils::getSrcDirectory(call)
    if (length(dir) && nzchar(dir) && nzchar(file)) {
      srcfile <- attr(utils::getSrcref(call), "srcfile")
      if (isTRUE(srcfile$isFile)) {
        file <- file.path(dir, file)
      } else {
        file <- file.path("R", file)
      }
    } else {
      file <- "??"
    }
    line <- utils::getSrcLocation(call) %||% "??"
    col <- utils::getSrcLocation(call, which = "column") %||% "??"
    style_srcref(paste0(file, ":", line, ":", col))
  }

  format_call <- function(call) {
    width <- getOption("width")
    str <- format(call)
    callstr <- if (length(str) > 1 || nchar(str[1]) > width) {
      paste0(substr(str[1], 1, width - 5), " ...")
    } else {
      str[1]
    }
    style_call(callstr)
  }

  format_call_src <- function(call) {
    callstr <- format_call(call)
    src <- format_srcref(call)
    if (length(src)) callstr <- paste0(callstr, "\n    ", src)
    callstr
  }

  enumerate <- function(x) {
    paste0(style_numbers(paste0(" ", seq_along(x), ". ")), x)
  }

  map2 <- function (.x, .y, .f, ...) {
    mapply(.f, .x, .y, MoreArgs = list(...), SIMPLIFY = FALSE,
           USE.NAMES = FALSE)
  }

  # -- printing, styles -------------------------------------------------

  has_crayon <- function() "crayon" %in% loadedNamespaces()

  style_numbers <- function(x) {
    if (has_crayon()) crayon::silver(x) else x
  }

  style_srcref <- function(x) {
    if (has_crayon()) crayon::italic(crayon::cyan(x))
  }

  style_error <- function(x) {
    sx <- paste0("\n x ", x, " ")
    if (has_crayon()) crayon::bold(crayon::red(sx)) else sx
  }

  style_trace_title <- function(x) {
    if (has_crayon()) crayon::bold(x) else x
  }

  style_process <- function(x) {
    if (has_crayon()) crayon::bold(x) else x
  }

  style_call <- function(x) {
    if (!has_crayon()) return(x)
    call <- sub("^([^(]+)[(].*$", "\\1", x)
    rest <- sub("^[^(]+([(].*)$", "\\1", x)
    if (call == x || rest == x) return(x)
    paste0(crayon::yellow(call), rest)
  }

  env <- environment()
  parent.env(env) <- baseenv()

  structure(
    list(
      .internal      = env,
      new_cond       = new_cond,
      new_error      = new_error,
      throw          = throw,
      rethrow        = rethrow,
      catch_rethrow  = catch_rethrow,
      rethrow_call   = rethrow_call,
      add_trace_back = add_trace_back
    ),
    class = c("standalone_errors", "standalone"))
})

# These are optional, and feel free to remove them if you prefer to
# call them through the `err` object.

new_cond  <- err$new_cond
new_error <- err$new_error
throw     <- err$throw
rethrow   <- err$rethrow
rethrow_call <- err$rethrow_call

fake_token <- function() paste0("ghp_", strrep("A", 36))

# A single fake-server subprocess is started on demand and reused across
# test files. Its lifetime is bound to testthat::teardown_env(), which lives
# until the test run ends.
.fake_server <- new.env(parent = emptyenv())

fake_server <- function() {
  skip_if_not_installed("webfakes")
  if (is.null(.fake_server$proc)) {
    .fake_server$proc <- webfakes::local_app_process(
      gh::fake_github_app(),
      .local_envir = testthat::teardown_env()
    )
  }
  .fake_server$proc
}

local_fake_github <- function(
  .local_envir = parent.frame(),
  token = fake_token()
) {
  proc <- fake_server()
  url <- proc$url()
  envvar <- gitcreds::gitcreds_cache_envvar(url)

  envs <- c(
    GITHUB_API_URL = url,
    GITHUB_PAT = NA_character_,
    GITHUB_TOKEN = NA_character_,
    setNames(token, envvar)
  )
  withr::local_envvar(envs, .local_envir = .local_envir)

  proc
}

# Replace the dynamic fake-server host with a stable placeholder so snapshots
# stay deterministic across runs.
redact_fake_host <- function(lines) {
  lines <- gsub(
    "https?://127\\.0\\.0\\.1:[0-9]+/api/v3",
    "https://api.github.com",
    lines
  )
  gsub("https?://127\\.0\\.0\\.1:[0-9]+", "https://api.github.com", lines)
}

if (!exists("TMPL", environment(), inherits = FALSE)) {
  TMPL <- function(x) x
}

test_that("repos, some basics", {
  local_fake_github()
  res <- gh(
    TMPL("/users/{username}/repos"),
    username = "gaborcsardi"
  )
  expect_true(all(c("id", "name", "full_name") %in% names(res[[1]])))

  res <- gh(
    TMPL("/orgs/{org}/repos"),
    org = "r-lib",
    type = "sources",
    sort = "full_name"
  )
  expect_true("actions" %in% vapply(res, "[[", "name", FUN.VALUE = ""))

  res <- gh("/repositories")
  expect_true(all(c("id", "name", "full_name") %in% names(res[[1]])))
})

test_that("can POST, PATCH, and DELETE", {
  local_fake_github()
  res <- gh(
    "POST /gists",
    files = list(test.R = list(content = "test")),
    description = "A test gist for gh",
    public = FALSE
  )
  expect_equal(res$description, "A test gist for gh")
  expect_false(res$public)

  res <- gh(
    TMPL("PATCH /gists/{gist_id}"),
    gist_id = res$id,
    description = "Still a test repo"
  )
  expect_equal(res$description, "Still a test repo")

  res <- gh(
    TMPL("DELETE /gists/{gist_id}"),
    gist_id = res$id
  )
  expect_s3_class(res, c("gh_response", "list"))
})

test_that("can print all types of object", {
  local_fake_github()
  local_options(gh_cache = FALSE)

  get_license <- function(...) {
    gh(
      "GET /repos/{owner}/{repo}/contents/{path}",
      owner = "r-lib",
      repo = "gh",
      path = "LICENSE",
      ref = "v1.2.0",
      ...
    )
  }

  json <- get_license()
  raw <- get_license(
    .send_headers = c(Accept = "application/vnd.github.v3.raw")
  )

  path <- withr::local_file(test_path("LICENSE"))
  file <- get_license(
    .destfile = path,
    .send_headers = c(Accept = "application/vnd.github.v3.raw")
  )

  expect_snapshot({
    json
    file
    raw
  })
})

test_that("whoami works in presence of PAT", {
  local_fake_github()
  res <- gh_whoami()
  expect_s3_class(res, "gh_response")
  expect_match(res[["scopes"]], "\\buser\\b")
})

test_that("whoami errors with bad/absent PAT", {
  local_fake_github()
  expect_snapshot(error = TRUE, {
    gh_whoami(.token = "")
    gh_whoami(.token = NA)
    gh_whoami(.token = "blah")
  })
})

#' Info on current GitHub user and token
#'
#' Reports wallet name, GitHub login, and GitHub URL for the current
#' authenticated user, the first bit of the token, and the associated scopes.
#'
#' Get a personal access token for the GitHub API from
#' <https://github.com/settings/tokens> and select the scopes necessary for your
#' planned tasks. The `repo` scope, for example, is one many are likely to need.
#'
#' On macOS and Windows it is best to store the token in the git credential
#' store, where most GitHub clients, including gh, can access it. You can
#' use the gitcreds package to add your token to the credential store:
#'
#' ```r
#' gitcreds::gitcreds_set()
#' ```
#'
#' See <https://gh.r-lib.org/articles/managing-personal-access-tokens.html>
#' and <https://usethis.r-lib.org/articles/articles/git-credentials.html>
#' for more about managing GitHub (and generic git) credentials.
#'
#' On other systems, including Linux, the git credential store is
#' typically not as convenient, and you might want to store your token in
#' the `GITHUB_PAT` environment variable, which you can set in your
#' `.Renviron` file.
#'
#' @inheritParams gh
#'
#' @return A `gh_response` object, which is also a `list`.
#' @export
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' gh_whoami()
#' @examplesIf FALSE
#' ## explicit token + use with GitHub Enterprise
#' gh_whoami(
#'   .token = "8c70fd8419398999c9ac5bacf3192882193cadf2",
#'   .api_url = "https://github.foobar.edu/api/v3"
#' )
gh_whoami <- function(.token = NULL, .api_url = NULL, .send_headers = NULL) {
  .token <- .token %||% gh_token(.api_url)
  if (isTRUE(.token == "")) {
    message(
      "No personal access token (PAT) available.\n",
      "Obtain a PAT from here:\n",
      "https://github.com/settings/tokens\n",
      "For more on what to do with the PAT, see ?gh_whoami."
    )
    return(invisible(NULL))
  }
  res <- gh(
    endpoint = "/user",
    .token = .token,
    .api_url = .api_url,
    .send_headers = .send_headers
  )
  scopes <- attr(res, "response")[["x-oauth-scopes"]]
  res <- res[c("name", "login", "html_url")]
  res$scopes <- scopes
  res$token <- format(gh_pat(.token))
  ## 'gh_response' class has to be restored
  class(res) <- c("gh_response", "list")
  res
}

#' A fake GitHub web app
#'
#' A [webfakes::new_app()] application that implements the subset of the
#' GitHub REST API needed by the gh test suite. It is exported so that
#' downstream packages that depend on gh can use it to test their own
#' GitHub-backed code without hitting the real API.
#'
#' The app accepts any token whose shape is a valid GitHub PAT (40 hex
#' characters or a `ghp_` / `github_pat_` / `ghs_` / `ghr_` / `gho_` /
#' `ghu_` prefix). Any other non-empty token is rejected with a 401
#' "Bad credentials". A missing `Authorization` header yields a 401
#' "Requires authentication" for endpoints that need a user.
#'
#' @return A `webfakes_app` object.
#' @export
#' @examplesIf rlang::is_installed("webfakes")
#' app <- fake_github_app()
#' proc <- webfakes::new_app_process(app)
#' proc$url()
#' proc$stop()
fake_github_app <- function() {
  check_installed("webfakes")

  app <- webfakes::new_app()
  app$use(webfakes::mw_json())
  app$use(webfakes::mw_raw())

  # gh treats non-github.com hosts as GitHub Enterprise and prefixes
  # request paths with /api/v3. Strip that here so the fake can mount
  # routes at the GitHub.com API paths.
  app$use(function(req, res) {
    if (startsWith(req$path, "/api/v3/")) {
      req$path <- substring(req$path, nchar("/api/v3") + 1L)
    }
    "next"
  })

  app$use(function(req, res) {
    auth <- req$get_header("Authorization") %||% ""
    res$locals$auth_present <- nzchar(auth)
    token <- if (grepl("^token\\s+", auth, ignore.case = TRUE)) {
      sub("^token\\s+", "", auth, ignore.case = TRUE)
    } else if (grepl("^bearer\\s+", auth, ignore.case = TRUE)) {
      sub("^bearer\\s+", "", auth, ignore.case = TRUE)
    } else {
      ""
    }
    res$locals$token <- token
    res$locals$token_valid <- nzchar(token) && fake_gh_token_valid(token)
    "next"
  })

  app$get("/missing", function(req, res) {
    fake_gh_error(res, 404L, "Not Found")
  })

  # Mirrors the real 422 returned when a (sha, context) pair has reached the
  # max number of commit statuses. GitHub returns `errors` as a plain string
  # here rather than the usual array of objects (see #229).
  app$post(
    "/repos/:owner/:repo/statuses/:sha",
    function(req, res) {
      res$set_status(422L)$send_json(
        list(
          message = "Validation Failed",
          errors = "Validation failed: This SHA and context has reached the maximum number of statuses.",
          documentation_url = "https://docs.github.com/rest/commits/statuses#create-a-commit-status",
          status = "422"
        ),
        auto_unbox = TRUE
      )
    }
  )

  app$get("/user", function(req, res) {
    if (!isTRUE(res$locals$auth_present)) {
      return(fake_gh_error(res, 401L, "Requires authentication"))
    }
    if (!isTRUE(res$locals$token_valid)) {
      return(fake_gh_error(res, 401L, "Bad credentials"))
    }
    res$set_header("x-oauth-scopes", "gist, repo, user")$send_json(
      list(
        login = "fakeuser",
        name = "Fake User",
        html_url = "https://github.com/fakeuser"
      ),
      auto_unbox = TRUE
    )
  })

  app$get("/rate_limit", function(req, res) {
    if (isTRUE(res$locals$auth_present) && !isTRUE(res$locals$token_valid)) {
      return(fake_gh_error(res, 401L, "Bad credentials"))
    }
    reset <- as.integer(unclass(Sys.time())) + 3600L
    one <- list(limit = 5000L, used = 0L, remaining = 5000L, reset = reset)
    res$set_header("x-ratelimit-limit", "5000")$set_header(
      "x-ratelimit-remaining",
      "5000"
    )$set_header("x-ratelimit-reset", as.character(reset))$send_json(
      list(
        resources = list(
          core = one,
          search = one,
          graphql = one,
          integration_manifest = one,
          code_scanning_upload = one
        ),
        rate = one
      ),
      auto_unbox = TRUE
    )
  })

  app$get("/orgs/:org/repos", function(req, res) {
    org <- req$params$org
    if (identical(org, "gh-org-testing-no-repos")) {
      return(res$send_json(list(), auto_unbox = TRUE))
    }
    if (identical(org, "gh-org-testing-404")) {
      return(fake_gh_error(res, 404L, "Not Found"))
    }
    repos <- fake_repos_for(org)
    send_paginated(req, res, repos, base_path = req$path)
  })

  # Raw body echo: read the entire raw body and echo it back. Used to
  # exercise gh's raw-body POST path.
  app$post("/echo-raw", function(req, res) {
    body <- if (length(req$raw)) req$raw else raw()
    res$set_type("application/octet-stream")$send(body)
  })

  # Cursor-style pagination: emits only a `next` link (no `last`), so the
  # client cannot learn the total page count up front.
  app$get("/cursor-list", function(req, res) {
    per_page <- clamp_int(req$query$per_page, default = 10L, max = 100L)
    page <- clamp_int(req$query$page, default = 1L, max = .Machine$integer.max)
    total <- 25L
    last_page <- ceiling(total / per_page)
    start <- (page - 1L) * per_page + 1L
    stop <- min(page * per_page, total)
    slice <- lapply(start:stop, function(i) list(id = i))
    if (page < last_page) {
      host <- req$get_header("Host") %||% "127.0.0.1"
      url <- paste0(
        req$protocol,
        "://",
        host,
        req$path,
        "?per_page=",
        per_page,
        "&page=",
        page + 1L
      )
      res$set_header("Link", paste0("<", url, '>; rel="next"'))
    }
    res$send_json(slice, auto_unbox = TRUE)
  })

  # GitHub returns Link header URLs against /organizations/{id}/repos for
  # subsequent pages. The fake app mirrors that to exercise the
  # cross-path pagination logic.
  app$get(
    "/organizations/:id/repos",
    function(req, res) {
      repos <- fake_repos_for("tidyverse")
      send_paginated(req, res, repos, base_path = req$path)
    }
  )

  app$get("/users/:user/repos", function(req, res) {
    etag <- paste0("\"users-", req$params$user, "-v1\"")
    res$set_header("ETag", etag)
    if (identical(req$get_header("If-None-Match"), etag)) {
      return(res$send_status(304L))
    }
    repos <- fake_repos_for(req$params$user)
    send_paginated(req, res, repos, base_path = req$path)
  })

  app$get("/repositories", function(req, res) {
    repos <- fake_repos_for("public")
    send_paginated(req, res, repos, base_path = req$path)
  })

  app$get("/users", function(req, res) {
    users <- fake_user_list(60L)
    send_paginated(req, res, users, base_path = req$path)
  })

  app$get(
    "/repos/:owner/:repo/contents/:path",
    function(req, res) {
      raw_requested <- grepl(
        "application/vnd.github(\\.v3)?\\.raw",
        req$get_header("Accept") %||% "",
        ignore.case = TRUE
      )
      if (raw_requested) {
        body <- charToRaw(paste0(
          "Package: ",
          req$params$repo,
          "\n",
          "Title: Fake content of ",
          req$params$path,
          "\n"
        ))
        res$set_header("x-github-media-type", "github.v3; param=raw")$set_type(
          "application/octet-stream"
        )$send(body)
      } else {
        res$set_header(
          "x-github-media-type",
          "github.v3; param=json"
        )$send_json(
          list(
            name = req$params$path,
            path = req$params$path,
            type = "file",
            content = "ZmFrZQ==",
            encoding = "base64"
          ),
          auto_unbox = TRUE
        )
      }
    }
  )

  app$post("/graphql", function(req, res) {
    query <- req$json$query %||% ""
    res$send_json(
      list(
        data = list(
          viewer = list(login = "fakeuser"),
          echo = query
        )
      ),
      auto_unbox = TRUE
    )
  })

  app$post("/markdown", function(req, res) {
    text <- req$json$text %||% ""
    if (!nzchar(text)) {
      return(res$send_status(200L))
    }
    res$set_type("text/html")$send(paste0("<p>", text, "</p>\n"))
  })

  app$get("/search/repositories", function(req, res) {
    items <- fake_repos_for("search")
    send_search(req, res, items)
  })

  app$get("/search/issues", function(req, res) {
    items <- fake_issues(60L)
    # Filter by `label:"..."` so a mangled q (e.g. space re-encoded as
    # literal `+` on a later page) returns no matches. This is what
    # surfaces the bug from #210.
    q <- req$query$q %||% ""
    m <- regmatches(q, regexpr('label:"([^"]+)"', q))
    if (length(m) == 1L) {
      label <- sub('label:"([^"]+)"', "\\1", m)
      items <- Filter(
        function(x) {
          any(vapply(
            x$labels,
            function(l) identical(l$name, label),
            logical(1)
          ))
        },
        items
      )
    }
    send_search(req, res, items)
  })

  app$post("/gists", function(req, res) {
    id <- as.character(as.integer(Sys.time()) + sample.int(1e6, 1L))
    gist <- list(
      id = id,
      description = req$json$description %||% "",
      public = req$json$public %||% FALSE,
      files = req$json$files %||% list()
    )
    fake_state_init(req)
    req$app$locals$gists[[id]] <- gist
    res$set_status(201L)$send_json(gist, auto_unbox = TRUE)
  })

  app$patch("/gists/:id", function(req, res) {
    id <- req$params$id
    fake_state_init(req)
    gist <- req$app$locals$gists[[id]]
    if (is.null(gist)) {
      return(fake_gh_error(res, 404L, "Not Found"))
    }
    for (nm in names(req$json)) {
      gist[[nm]] <- req$json[[nm]]
    }
    req$app$locals$gists[[id]] <- gist
    res$send_json(gist, auto_unbox = TRUE)
  })

  app$delete("/gists/:id", function(req, res) {
    id <- req$params$id
    fake_state_init(req)
    if (is.null(req$app$locals$gists[[id]])) {
      return(fake_gh_error(res, 404L, "Not Found"))
    }
    req$app$locals$gists[[id]] <- NULL
    res$send_status(204L)
  })

  app
}

fake_state_init <- function(req) {
  if (is.null(req$app$locals$gists)) {
    req$app$locals$gists <- list()
  }
}

fake_gh_token_valid <- function(token) {
  grepl(
    "^(gh[pousr]_[A-Za-z0-9_]{36,251}|github_pat_[A-Za-z0-9_]{36,244}|ghs_.+)$",
    token
  ) ||
    grepl("^[[:xdigit:]]{40}$", token)
}

fake_gh_error <- function(res, status, message) {
  res$set_status(status)$send_json(
    list(
      message = message,
      documentation_url = "https://docs.github.com/rest"
    ),
    auto_unbox = TRUE
  )
}

fake_repos_for <- function(slug) {
  names <- switch(
    slug,
    "r-lib" = c(
      "actions",
      "cli",
      "covr",
      "devtools",
      "fs",
      "gh",
      "httr2",
      "pkgbuild",
      "pkgload",
      "rcmdcheck",
      "remotes",
      "rlang",
      "roxygen2",
      "testthat",
      "usethis",
      "withr"
    ),
    "tidyverse" = sprintf("repo-%02d", 1:60),
    "search" = sprintf("tidy-%02d", 1:60),
    sprintf("%s-repo-%02d", slug, 1:30)
  )
  owner <- if (slug %in% c("search", "public")) "someone" else slug
  lapply(seq_along(names), function(i) {
    list(
      id = 1000L + i,
      name = names[[i]],
      full_name = paste0(owner, "/", names[[i]]),
      owner = list(login = owner)
    )
  })
}

fake_user_list <- function(n) {
  lapply(seq_len(n), function(i) {
    list(
      id = i,
      login = sprintf("user%02d", i),
      type = "User"
    )
  })
}

fake_issues <- function(n) {
  lapply(seq_len(n), function(i) {
    list(
      id = i,
      number = i,
      title = sprintf("Issue %d", i),
      labels = list(list(name = "tidy-dev-day :nerd_face:"))
    )
  })
}

send_paginated <- function(req, res, items, base_path) {
  per_page <- clamp_int(req$query$per_page, default = 30L, max = 100L)
  page <- clamp_int(req$query$page, default = 1L, max = .Machine$integer.max)
  total <- length(items)
  last_page <- max(1L, ceiling(total / per_page))
  page <- min(page, last_page)
  start <- (page - 1L) * per_page + 1L
  stop <- min(page * per_page, total)
  slice <- if (total == 0L) list() else items[start:stop]

  set_link_header(req, res, base_path, page, last_page, per_page)
  res$send_json(slice, auto_unbox = TRUE)
}

send_search <- function(req, res, items) {
  per_page <- clamp_int(req$query$per_page, default = 30L, max = 100L)
  page <- clamp_int(req$query$page, default = 1L, max = .Machine$integer.max)
  total <- length(items)
  last_page <- max(1L, ceiling(total / per_page))
  page <- min(page, last_page)
  start <- (page - 1L) * per_page + 1L
  stop <- min(page * per_page, total)
  slice <- if (total == 0L) list() else items[start:stop]

  set_link_header(req, res, req$path, page, last_page, per_page)
  res$send_json(
    list(
      total_count = total,
      incomplete_results = FALSE,
      items = slice
    ),
    auto_unbox = TRUE
  )
}

set_link_header <- function(req, res, base_path, page, last_page, per_page) {
  if (last_page <= 1L) {
    return(invisible())
  }
  host <- req$get_header("Host") %||% "127.0.0.1"
  base <- paste0(req$protocol, "://", host)
  query <- req$query
  query$per_page <- per_page
  build <- function(p) {
    q <- query
    q$page <- p
    parts <- mapply(
      function(k, v) {
        # Match real GitHub: percent-encode reserved chars, but spaces
        # in query values come back as `+` rather than `%20`.
        enc <- utils::URLencode(as.character(v), reserved = TRUE)
        enc <- gsub("%20", "+", enc, fixed = TRUE)
        paste0(k, "=", enc)
      },
      names(q),
      q,
      USE.NAMES = FALSE
    )
    paste0(base, base_path, "?", paste(parts, collapse = "&"))
  }
  parts <- character()
  if (page < last_page) {
    parts <- c(parts, paste0("<", build(page + 1L), '>; rel="next"'))
    parts <- c(parts, paste0("<", build(last_page), '>; rel="last"'))
  }
  if (page > 1L) {
    parts <- c(parts, paste0("<", build(1L), '>; rel="first"'))
    parts <- c(parts, paste0("<", build(page - 1L), '>; rel="prev"'))
  }
  res$set_header("Link", paste(parts, collapse = ", "))
}

clamp_int <- function(x, default, max) {
  if (is.null(x) || !nzchar(x)) {
    return(default)
  }
  x <- suppressWarnings(as.integer(x))
  if (is.na(x) || x < 1L) {
    return(default)
  }
  min(x, max)
}

withr::local_options(
  gh_cache = FALSE,
  .local_envir = testthat::teardown_env()
)

# Make the test suite hermetic: prevent any real PAT being picked up from
# the user's environment or git credential store. Individual tests can
# still override these via withr.
withr::local_envvar(
  c(
    GITHUB_API_URL = NA,
    GITHUB_PAT_GITHUB_COM = "",
    GITHUB_PAT = "",
    GITHUB_TOKEN = ""
  ),
  .local_envir = testthat::teardown_env()
)

gh_has <- function(gh_response, link) {
  resp <- attr(gh_response, "httr2_response")
  !is.null(httr2::resp_link_url(resp, link))
}

gh_has_next <- function(gh_response) {
  gh_has(gh_response, "next")
}

gh_link_request <- function(gh_response, link, .token, .send_headers) {
  if (!inherits(gh_response, "gh_response")) {
    stop_input_type(gh_response, "a <gh_response> object")
  }

  resp <- attr(gh_response, "httr2_response")
  url <- httr2::resp_link_url(resp, link)
  if (is.null(url)) {
    cli::cli_abort("No {link} page")
  }

  req <- attr(gh_response, "request")
  req$url <- url
  req$token <- .token
  req$send_headers <- .send_headers
  req <- gh_set_headers(req)
  req
}

gh_link <- function(gh_response, link, .token, .send_headers) {
  req <- gh_link_request(gh_response, link, .token, .send_headers)
  raw <- gh_make_request(req)
  gh_process_response(raw, req)
}

#' Get the next, previous, first or last page of results
#'
#' @details
#' Note that these are not always defined. E.g. if the first
#' page was queried (the default), then there are no first and previous
#' pages defined. If there is no next page, then there is no
#' next page defined, etc.
#'
#' If the requested page does not exist, an error is thrown.
#'
#' @param gh_response An object returned by a [gh()] call.
#' @inheritParams gh
#' @return Answer from the API.
#'
#' @seealso The `.limit` argument to [gh()] supports fetching more than
#'   one page.
#'
#' @name gh_next
#' @export
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' x <- gh("/users")
#' vapply(x, "[[", character(1), "login")
#' x2 <- gh_next(x)
#' vapply(x2, "[[", character(1), "login")
gh_next <- function(gh_response, .token = NULL, .send_headers = NULL) {
  gh_link(gh_response, "next", .token = .token, .send_headers = .send_headers)
}

#' @name gh_next
#' @export

gh_prev <- function(gh_response, .token = NULL, .send_headers = NULL) {
  gh_link(gh_response, "prev", .token = .token, .send_headers = .send_headers)
}

#' @name gh_next
#' @export

gh_first <- function(gh_response, .token = NULL, .send_headers = NULL) {
  gh_link(gh_response, "first", .token = .token, .send_headers = .send_headers)
}

#' @name gh_next
#' @export

gh_last <- function(gh_response, .token = NULL, .send_headers = NULL) {
  gh_link(gh_response, "last", .token = .token, .send_headers = .send_headers)
}

test_that("can extract relative pages", {
  proc <- local_fake_github()
  page1 <- gh("/orgs/tidyverse/repos", per_page = 1)
  expect_true(gh_has(page1, "next"))
  expect_false(gh_has(page1, "prev"))

  page2 <- gh_next(page1)
  expect_equal(
    attr(page2, "request")$url,
    paste0(sub("/$", "", proc$url()), "/orgs/tidyverse/repos?per_page=1&page=2")
  )
  expect_true(gh_has(page2, "prev"))

  expect_snapshot(gh_prev(page1), error = TRUE)
})

test_that("can paginate even when space re-encoded to +", {
  local_fake_github()
  json <- gh::gh(
    "GET /search/issues",
    q = 'label:"tidy-dev-day :nerd_face:"',
    per_page = 10,
    .limit = 20
  )
  expect_length(json$items, 20)
})

test_that("gh_has_next, gh_first, gh_last work", {
  local_fake_github()
  page1 <- gh("/orgs/tidyverse/repos", per_page = 1)
  expect_true(gh_has_next(page1))

  page2 <- gh_next(page1)
  first <- gh_first(page2)
  expect_false(gh_has_next(gh_last(page1)))
  expect_equal(length(first), 1L)
})

test_that("gh_link_request errors on non-gh_response input", {
  expect_snapshot(
    error = TRUE,
    gh_link_request(list(), "next", .token = NULL, .send_headers = NULL)
  )
})

test_that("interrupt during pagination signals gh_interrupt with partial data", {
  local_fake_github()

  original_process <- gh_process_response
  call_count <- 0L
  fake_process <- function(resp, gh_req) {
    call_count <<- call_count + 1L
    if (call_count >= 2L) {
      rlang::interrupt()
    }
    original_process(resp, gh_req)
  }
  local_mocked_bindings(gh_process_response = fake_process)

  cond <- tryCatch(
    gh("/orgs/tidyverse/repos", per_page = 1, .limit = 5, .progress = FALSE),
    gh_interrupt = function(e) e
  )
  expect_s3_class(cond, "gh_interrupt")
  expect_s3_class(cond, "interrupt")
  expect_s3_class(cond$gh_result, "gh_response")
  expect_length(cond$gh_result, 1L)
})

test_that("generic interrupt handler also receives gh_result", {
  local_fake_github()

  original_process <- gh_process_response
  call_count <- 0L
  fake_process <- function(resp, gh_req) {
    call_count <<- call_count + 1L
    if (call_count >= 2L) {
      rlang::interrupt()
    }
    original_process(resp, gh_req)
  }
  local_mocked_bindings(gh_process_response = fake_process)

  cond <- tryCatch(
    gh("/orgs/tidyverse/repos", per_page = 1, .limit = 5, .progress = FALSE),
    interrupt = function(e) e
  )
  expect_s3_class(cond, "gh_interrupt")
  expect_s3_class(cond$gh_result, "gh_response")
})

test_that("interrupt before first page signals gh_interrupt with NULL gh_result", {
  local_fake_github()

  fake_process <- function(resp, gh_req) {
    rlang::interrupt()
  }
  local_mocked_bindings(gh_process_response = fake_process)

  cond <- tryCatch(
    gh("/orgs/tidyverse/repos", per_page = 1, .limit = 5, .progress = FALSE),
    gh_interrupt = function(e) e
  )
  expect_s3_class(cond, "gh_interrupt")
  expect_null(cond$gh_result)
})

test_that("cap_display_totals caps the limit-derived estimate (#239)", {
  # A huge finite `.limit` used to mean "everything": the displayed total
  # should drop to what the endpoint actually has, not the limit.
  capped <- cap_display_totals(
    display_total = 1e6,
    display_pages = 10000,
    last_page = 24,
    per_page = 100
  )
  expect_equal(capped$pages, 24)
  expect_equal(capped$total, 2400)

  # A small `.limit` (fewer than what's available) is kept as-is.
  capped <- cap_display_totals(
    display_total = 50,
    display_pages = 1,
    last_page = 24,
    per_page = 50
  )
  expect_equal(capped$pages, 1)
  expect_equal(capped$total, 50)

  # Infinite `.limit` (NA estimate): the header values win outright.
  capped <- cap_display_totals(
    display_total = NA_integer_,
    display_pages = NA_integer_,
    last_page = 24,
    per_page = 100
  )
  expect_equal(capped$pages, 24)
  expect_equal(capped$total, 2400)
})

test_that("paginated request gets max_wait and max_rate", {
  local_fake_github()
  gh <- gh(
    "/orgs/tidyverse/repos",
    per_page = 5,
    .max_wait = 1,
    .max_rate = 10
  )

  req <- gh_link_request(gh, "next", .token = NULL, .send_headers = NULL)
  expect_equal(req$max_wait, 1)
  expect_equal(req$max_rate, 10)

  url <- httr2::url_parse(req$url)
  expect_equal(url$query$page, "2")
})

test_package_root <- function() {
  x <- tryCatch(
    rprojroot::find_package_root_file(),
    error = function(e) NULL
  )

  if (!is.null(x)) {
    return(x)
  }

  pkg <- testthat::testing_package()
  x <- tryCatch(
    rprojroot::find_package_root_file(
      path = file.path("..", "..", "00_pkg_src", pkg)
    ),
    error = function(e) NULL
  )

  if (!is.null(x)) {
    return(x)
  }

  stop("Cannot find package root")
}

is_false_check_env_var <- function(x, default = "") {
  # like utils:::str2logical
  val <- Sys.getenv(x, default)
  if (isFALSE(as.logical(val))) {
    return(TRUE)
  }
  tolower(val) %in% c("0", "no")
}

skip_if_not_installed <- function(pkg) {
  if (!is_false_check_env_var("_R_CHECK_FORCE_SUGGESTS_")) {
    return()
  }
  testthat::skip_if_not_installed(pkg)
}

make_git_repo <- function(remotes = NULL, .local_envir = parent.frame()) {
  dir <- withr::local_tempdir(.local_envir = .local_envir)
  dir.create(file.path(dir, ".git"))
  cfg <- character()
  for (nm in names(remotes)) {
    cfg <- c(
      cfg,
      sprintf('[remote "%s"]', nm),
      sprintf("  url = %s", remotes[[nm]])
    )
  }
  writeLines(cfg, file.path(dir, ".git", "config"))
  dir
}

transform_tempdir <- function(x) {
  x <- sub(tempdir(), "<tempdir>", x, fixed = TRUE)
  x <- sub(normalizePath(tempdir()), "<tempdir>", x, fixed = TRUE)
  x <- sub(
    normalizePath(tempdir(), winslash = "/"),
    "<tempdir>",
    x,
    fixed = TRUE
  )
  x <- sub("\\R\\", "/R/", x, fixed = TRUE)
  x <- sub("[\\\\/]file[a-zA-Z0-9]+", "/<tempfile>", x)
  x <- sub("[A-Z]:.*Rtmp[a-zA-Z0-9]+[\\\\/]", "<tempdir>/", x)
  x <- gsub("'[A-Z]:/'", "'/'", x)
  x
}

test_that("spelling", {
  skip_on_cran()
  skip_on_covr()
  pkgroot <- test_package_root()
  err <- spelling::spell_check_package(pkgroot)
  num_spelling_errors <- nrow(err)
  expect_true(
    num_spelling_errors == 0,
    info = paste(
      c("\nSpelling errors:", capture.output(err)),
      collapse = "\n"
    )
  )
})

#' Return the local user's GitHub Personal Access Token (PAT)
#'
#' @description
#' If gh can find a personal access token (PAT) via `gh_token()`, it includes
#' the PAT in its requests. Some requests succeed without a PAT, but many
#' require a PAT to prove the request is authorized by a specific GitHub user. A
#' PAT also helps with rate limiting. If your gh use is more than casual, you
#' want a PAT.
#'
#' gh calls [gitcreds::gitcreds_get()] with the `api_url`, which checks session
#' environment variables (`GITHUB_PAT`, `GITHUB_TOKEN`)
#' and then the local Git credential store for a PAT
#' appropriate to the `api_url`. Therefore, if you have previously used a PAT
#' with, e.g., command line Git, gh may retrieve and re-use it. You can call
#' [gitcreds::gitcreds_get()] directly, yourself, if you want to see what is
#' found for a specific URL. If no matching PAT is found,
#' [gitcreds::gitcreds_get()] errors, whereas `gh_token()` does not and,
#' instead, returns `""`.
#'
#' See GitHub's documentation on [Creating a personal access
#' token](https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token),
#' or use `usethis::create_github_token()` for a guided experience, including
#' pre-selection of recommended scopes. Once you have a PAT, you can use
#' [gitcreds::gitcreds_set()] to add it to the Git credential store. From that
#' point on, gh (via [gitcreds::gitcreds_get()]) should be able to find it
#' without further effort on your part.
#'
#' # Token format validation
#'
#' gh warns if the PAT it retrieves does not match a known format.
#' Set `options(gh_validate_tokens = "off")` or the
#' `GH_VALIDATE_TOKENS=off` environment variable to avoid this warning.
#' The option takes precedence over the environment variable.
#'
#' Set `options(gh_validate_tokens = "error")` or the
#' `GH_VALIDATE_TOKENS=error` environment variable to make gh throw an
#' error for an unrecognized PAT format.
#'
#' @param api_url GitHub API URL. Defaults to the `GITHUB_API_URL` environment
#'   variable, if set, and otherwise to <https://api.github.com>.
#'
#' @return A string of characters, if a PAT is found, or the empty
#'   string, otherwise. For convenience, the return value has an S3 class in
#'   order to ensure that simple printing strategies don't reveal the entire
#'   PAT.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' gh_token()
#'
#' format(gh_token())
#'
#' str(gh_token())
#' }
gh_token <- function(api_url = NULL) {
  api_url <- api_url %||% default_api_url()
  check_string(api_url)
  host_url <- get_hosturl(api_url)
  # Check for credentials supplied by Posit Connect.
  if (is_installed("connectcreds")) {
    if (connectcreds::has_viewer_token(host_url)) {
      token <- connectcreds::connect_viewer_token(host_url)
      return(gh_pat(token$access_token))
    }
  }
  token <- tryCatch(
    gitcreds::gitcreds_get(host_url),
    error = function(e) NULL
  )
  gh_pat(token$password %||% "")
}

#' @export
#' @rdname gh_token
gh_token_exists <- function(api_url = NULL) {
  tryCatch(nzchar(gh_token(api_url)), error = function(e) FALSE)
}

gh_auth <- function(token) {
  if (isTRUE(token != "")) {
    if (any(grepl("\\s", token))) {
      warning("Token contains whitespace characters")
    }
    c("Authorization" = paste("token", trim_ws(token)))
  } else {
    character()
  }
}

# gh_pat class: exists in order have a print method that hides info ----
new_gh_pat <- function(x) {
  if (is.character(x) && length(x) == 1) {
    structure(x, class = "gh_pat")
  } else {
    cli::cli_abort("A GitHub PAT must be a string")
  }
}

# validates PAT only in a very narrow, technical, and local sense
validate_gh_pat <- function(x) {
  if (!inherits(x, "gh_pat")) {
    stop_input_type(x, "a <gh_pat> object")
  }
  mode <- get_validate_tokens_mode()
  if (mode == "off") {
    return(x)
  }
  if (
    x == "" ||
      # https://github.blog/changelog/2021-03-04-authentication-token-format-updates/
      # Fine grained tokens start with "github_pat_".
      # https://github.blog/changelog/2022-10-18-introducing-fine-grained-personal-access-tokens/
      # GitHub App installation tokens start with "ghs_".
      # https://github.blog/changelog/2026-04-24-notice-about-upcoming-new-format-for-github-app-installation-tokens/
      grepl(
        "^(gh[pousr]_[A-Za-z0-9_]{36,251}|github_pat_[A-Za-z0-9_]{36,244}|ghs_.+)$",
        x
      ) ||
      grepl("^[[:xdigit:]]{40}$", x)
  ) {
    return(x)
  }
  url <- "https://gh.r-lib.org/articles/managing-personal-access-tokens.html"
  msg <- c(
    "Invalid GitHub PAT format",
    "i" = "A GitHub PAT must have one of four forms:",
    "*" = "40 hexadecimal digits (older PATs)",
    "*" = "A 'ghp_' prefix followed by 36 to 251 more characters (newer
           PATs)",
    "*" = "A `ghs_` prefix followed by about 500 characters (GitHub App
           installation tokens)",
    "*" = "A 'github_pat_' prefix followed by 36 to 244 more characters
           (fine-grained PATs)",
    "i" = "Read more at {.url {url}}."
  )
  sil <- c(
    "i" = "Set {.code options(gh_validate_tokens = \"off\")} or env var
           {.envvar GH_VALIDATE_TOKENS=off} to silence this."
  )
  if (mode == "warn") {
    cli::cli_warn(
      c(msg, sil),
      .frequency = "regularly",
      .frequency_id = "gh_invalid_pat"
    )
    return(x)
  }
  cli::cli_abort(msg)
}

# Resolve token-validation mode: option > env var > "warn".
get_validate_tokens_mode <- function() {
  opt <- getOption("gh_validate_tokens")
  if (is.null(opt)) {
    env <- Sys.getenv("GH_VALIDATE_TOKENS", unset = "")
    mode <- if (nzchar(env)) env else "warn"
  } else {
    mode <- opt
  }
  if (!is.character(mode) || length(mode) != 1L || is.na(mode)) {
    cli::cli_abort(c(
      "Invalid token validation setting: must be a single string.",
      "i" = "Got {.obj_type_friendly {mode}}."
    ))
  }
  mode <- tolower(mode)
  if (!mode %in% c("off", "warn", "error")) {
    cli::cli_abort(c(
      "Invalid token validation mode: {.val {mode}}.",
      "i" = "Must be one of {.val off}, {.val warn}, or {.val error}.",
      "i" = "Configure via {.code options(gh_validate_tokens = ...)} or
             env var {.envvar GH_VALIDATE_TOKENS}."
    ))
  }
  mode
}

gh_pat <- function(x) {
  validate_gh_pat(new_gh_pat(x))
}

#' @export
format.gh_pat <- function(x, ...) {
  if (x == "") {
    "<no PAT>"
  } else {
    obfuscate(x)
  }
}

#' @export
print.gh_pat <- function(x, ...) {
  cat(format(x), sep = "\n")
  invisible(x)
}

#' @export
str.gh_pat <- function(object, ...) {
  cat(paste0("<gh_pat> ", format(object), "\n", collapse = ""))
  invisible()
}

obfuscate <- function(x, first = 4, last = 4) {
  paste0(
    substr(x, start = 1, stop = first),
    "...",
    substr(x, start = nchar(x) - last + 1, stop = nchar(x))
  )
}

test_that("URL specific token is used", {
  good <- gh_pat(strrep("a", 40))
  good2 <- gh_pat(strrep("b", 40))
  bad <- gh_pat(strrep("0", 40))
  bad2 <- gh_pat(strrep("1", 40))

  env <- c(
    GITHUB_API_URL = "https://github.acme.com",
    GITHUB_PAT_GITHUB_ACME_COM = good,
    GITHUB_PAT_GITHUB_ACME2_COM = good2,
    GITHUB_PAT = bad,
    GITHUB_TOKEN = bad2
  )
  withr::with_envvar(env, {
    expect_equal(gh_token(), good)
    expect_equal(gh_token("https://github.acme2.com"), good2)
  })

  env <- c(
    GITHUB_API_URL = NA,
    GITHUB_PAT_GITHUB_COM = good,
    GITHUB_PAT = bad,
    GITHUB_TOKEN = bad2
  )
  withr::with_envvar(env, {
    expect_equal(gh_token(), good)
    expect_equal(gh_token("https://api.github.com"), good)
  })
})

test_that("fall back to GITHUB_PAT, then GITHUB_TOKEN", {
  pat <- gh_pat(strrep("a", 40))
  token <- gh_pat(strrep("0", 40))

  env <- c(
    GITHUB_API_URL = NA,
    GITHUB_PAT_GITHUB_COM = NA,
    GITHUB_PAT = pat,
    GITHUB_TOKEN = token
  )
  withr::with_envvar(env, {
    expect_equal(gh_token(), pat)
    expect_equal(gh_token("https://api.github.com"), pat)
  })

  env <- c(
    GITHUB_API_URL = NA,
    GITHUB_PAT_GITHUB_COM = NA,
    GITHUB_PAT = NA,
    GITHUB_TOKEN = token
  )
  withr::with_envvar(env, {
    expect_equal(gh_token(), token)
    expect_equal(gh_token("https://api.github.com"), token)
  })
})

test_that("gh_token_exists works as expected", {
  withr::local_options(gh_validate_tokens = "error")
  withr::local_envvar(GITHUB_API_URL = "https://test.com")

  withr::local_envvar(GITHUB_PAT_TEST_COM = NA)
  expect_false(gh_token_exists())

  withr::local_envvar(GITHUB_PAT_TEST_COM = gh_pat(strrep("0", 40)))
  expect_true(gh_token_exists())

  withr::local_envvar(GITHUB_PAT_TEST_COM = "invalid")
  expect_false(gh_token_exists())
})

# gh_pat class ----
test_that("validate_gh_pat() rejects bad characters, wrong # of characters", {
  withr::local_options(gh_validate_tokens = "error")

  # older PATs
  expect_error(gh_pat(strrep("a", 40)), NA)
  expect_error(
    gh_pat(strrep("g", 40)),
    "40 hexadecimal digits",
    class = "error"
  )
  expect_error(gh_pat("aa"), "40 hexadecimal digits", class = "error")

  # newer PATs
  expect_error(gh_pat(paste0("ghp_", strrep("B", 36))), NA)
  expect_error(gh_pat(paste0("ghp_", strrep("3", 251))), NA)
  expect_error(gh_pat(paste0("github_pat_", strrep("A", 36))), NA)
  expect_error(gh_pat(paste0("github_pat_", strrep("3", 244))), NA)
  expect_error(
    gh_pat(paste0("ghJ_", strrep("a", 36))),
    "prefix",
    class = "error"
  )
  expect_error(
    gh_pat(paste0("github_pa_", strrep("B", 244))),
    "github_pat_",
    class = "error"
  )
})

test_that("validate_gh_pat() honors gh_validate_tokens option and env var", {
  bad <- "definitely-not-a-pat"

  # Default is "warn": warns but still returns the value. Reset the
  # session-wide warning throttle so the expected warning isn't suppressed
  # by an earlier firing.
  withr::local_options(gh_validate_tokens = NULL)
  withr::local_envvar(GH_VALIDATE_TOKENS = NA)
  rlang::reset_warning_verbosity("gh_invalid_pat")
  expect_warning(out <- gh_pat(bad), "Invalid GitHub PAT format")
  expect_s3_class(out, "gh_pat")
  expect_equal(unclass(out), bad)

  # "off" skips validation, no message of any kind.
  withr::local_options(gh_validate_tokens = "off")
  expect_silent(out <- gh_pat(bad))
  expect_equal(unclass(out), bad)

  # "error" aborts (current pre-default behavior).
  withr::local_options(gh_validate_tokens = "error")
  expect_error(gh_pat(bad), "Invalid GitHub PAT format")

  # Env var is honored when option is unset.
  withr::local_options(gh_validate_tokens = NULL)
  withr::local_envvar(GH_VALIDATE_TOKENS = "error")
  expect_error(gh_pat(bad), "Invalid GitHub PAT format")

  # Option takes precedence over env var (env var still "error" from above).
  withr::local_options(gh_validate_tokens = "off")
  expect_silent(gh_pat(bad))

  # Unknown mode errors loudly.
  withr::local_options(gh_validate_tokens = "bogus")
  expect_error(gh_pat(bad), "Invalid token validation mode")
})

test_that("get_validate_tokens_mode() rejects a setting that isn't a single string", {
  bad <- "definitely-not-a-pat"

  withr::local_options(gh_validate_tokens = c("warn", "error"))
  expect_snapshot(error = TRUE, gh_pat(bad))

  withr::local_options(gh_validate_tokens = 42L)
  expect_snapshot(error = TRUE, gh_pat(bad))

  withr::local_options(gh_validate_tokens = NA_character_)
  expect_snapshot(error = TRUE, gh_pat(bad))
})

test_that("format.gh_pat() and str.gh_pat() hide the middle stuff", {
  pat <- paste0(strrep("a", 10), strrep("4", 20), strrep("F", 10))
  expect_match(format(gh_pat(pat)), "[a-zA-Z]+")
  expect_output(str(gh_pat(pat)), "[a-zA-Z]+")
})

test_that("str.gh_pat() indicates it's a `gh_pat`", {
  pat <- paste0(strrep("a", 10), strrep("4", 20), strrep("F", 10))
  expect_output(str(gh_pat(pat)), "gh_pat")
})

test_that("format.gh_pat() handles empty string", {
  expect_match(format(gh_pat("")), "<no PAT>")
})

test_that("print.gh_pat prints the obfuscated format", {
  pat <- gh_pat(paste0("ghp_", strrep("A", 36)))
  expect_output(print(pat), "ghp_.*\\.\\.\\.")
})

test_that("new_gh_pat rejects non-string input", {
  expect_snapshot(error = TRUE, new_gh_pat(1L))
  expect_snapshot(error = TRUE, new_gh_pat(c("a", "b")))
})

test_that("validate_gh_pat rejects non-gh_pat input", {
  expect_snapshot(error = TRUE, validate_gh_pat("not-a-pat-object"))
})

test_that("gh_auth warns on token containing whitespace", {
  expect_warning(
    out <- gh_auth("token with space"),
    "whitespace"
  )
  expect_equal(unname(out), "token token with space")
})

# URL processing helpers ----
test_that("get_baseurl() insists on http(s)", {
  expect_snapshot(error = TRUE, {
    get_baseurl("github.com")
    get_baseurl("github.acme.com")
  })
})

test_that("get_baseurl() works", {
  x <- "https://github.com"
  expect_equal(get_baseurl("https://github.com"), x)
  expect_equal(get_baseurl("https://github.com/"), x)
  expect_equal(get_baseurl("https://github.com/stuff"), x)
  expect_equal(get_baseurl("https://github.com/stuff/"), x)
  expect_equal(get_baseurl("https://github.com/more/stuff"), x)

  x <- "https://api.github.com"
  expect_equal(get_baseurl("https://api.github.com"), x)
  expect_equal(get_baseurl("https://api.github.com/rate_limit"), x)

  x <- "https://github.acme.com"
  expect_equal(get_baseurl("https://github.acme.com"), x)
  expect_equal(get_baseurl("https://github.acme.com/"), x)
  expect_equal(get_baseurl("https://github.acme.com/api/v3"), x)

  # so (what little) support we have for user@host doesn't regress
  expect_equal(
    get_baseurl("https://jane@github.acme.com/api/v3"),
    "https://jane@github.acme.com"
  )
})

test_that("is_github_dot_com() works", {
  expect_true(is_github_dot_com("https://github.com"))
  expect_true(is_github_dot_com("https://api.github.com"))
  expect_true(is_github_dot_com("https://api.github.com/rate_limit"))
  expect_true(is_github_dot_com("https://api.github.com/graphql"))

  expect_false(is_github_dot_com("https://github.acme.com"))
  expect_false(is_github_dot_com("https://github.acme.com/api/v3"))
  expect_false(is_github_dot_com("https://github.acme.com/api/v3/user"))
})

test_that("get_hosturl() works", {
  x <- "https://github.com"
  expect_equal(get_hosturl("https://github.com"), x)
  expect_equal(get_hosturl("https://api.github.com"), x)

  x <- "https://github.acme.com"
  expect_equal(get_hosturl("https://github.acme.com"), x)
  expect_equal(get_hosturl("https://github.acme.com/api/v3"), x)
})

test_that("get_apiurl() works", {
  x <- "https://api.github.com"
  expect_equal(get_apiurl("https://github.com"), x)
  expect_equal(get_apiurl("https://github.com/"), x)
  expect_equal(get_apiurl("https://github.com/r-lib/gh/issues"), x)
  expect_equal(get_apiurl("https://api.github.com"), x)
  expect_equal(get_apiurl("https://api.github.com/rate_limit"), x)

  x <- "https://github.acme.com/api/v3"
  expect_equal(get_apiurl("https://github.acme.com"), x)
  expect_equal(get_apiurl("https://github.acme.com/OWNER/REPO"), x)
  expect_equal(get_apiurl("https://github.acme.com/api/v3"), x)
})

test_that("tokens can be requested from a Connect server", {
  skip_if_not_installed("connectcreds")

  token <- strrep("a", 40)
  connectcreds::local_mocked_connect_responses(token = token)
  expect_equal(gh_token(), gh_pat(token))
})

#' Return GitHub user's current rate limits
#'
#' @description
#' `gh_rate_limits()` reports on all rate limits for the authenticated user.
#' `gh_rate_limit()` reports on rate limits for previous successful request.
#'
#' Further details on GitHub's API rate limit policies are available at
#' <https://docs.github.com/v3/#rate-limiting>.
#'
#' @param response `gh_response` object from a previous `gh` call, rate
#' limit values are determined from values in the response header.
#' Optional argument, if missing a call to "GET /rate_limit" will be made.
#'
#' @inheritParams gh
#'
#' @return A `list` object containing the overall `limit`, `remaining` limit, and the
#' limit `reset` time.
#'
#' @export

gh_rate_limit <- function(
  response = NULL,
  .token = NULL,
  .api_url = NULL,
  .send_headers = NULL
) {
  if (is.null(response)) {
    # This end point does not count against limit
    .token <- .token %||% gh_token(.api_url)
    response <- gh(
      "GET /rate_limit",
      .token = .token,
      .api_url = .api_url,
      .send_headers = .send_headers
    )
  }

  if (!inherits(response, "gh_response")) {
    stop_input_type(response, "a <gh_response> object")
  }

  http_res <- attr(response, "response")

  reset <- as.integer(c(http_res[["x-ratelimit-reset"]], NA)[1])
  reset <- as.POSIXct(reset, origin = "1970-01-01")

  list(
    limit = as.integer(c(http_res[["x-ratelimit-limit"]], NA)[1]),
    remaining = as.integer(c(http_res[["x-ratelimit-remaining"]], NA)[1]),
    reset = reset
  )
}

#' @export
#' @rdname gh_rate_limit
gh_rate_limits <- function(
  .token = NULL,
  .api_url = NULL,
  .send_headers = NULL
) {
  .token <- .token %||% gh_token(.api_url)
  response <- gh(
    "GET /rate_limit",
    .token = .token,
    .api_url = .api_url,
    .send_headers = .send_headers
  )

  resources <- response$resources

  reset <- .POSIXct(sapply(resources, "[[", "reset"))

  data.frame(
    type = names(resources),
    limit = sapply(resources, "[[", "limit"),
    used = sapply(resources, "[[", "used"),
    remaining = sapply(resources, "[[", "remaining"),
    reset = reset,
    mins_left = round((unclass(reset) - unclass(Sys.time())) / 60, 1),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

test_that("good input", {
  mock_res <- structure(
    list(),
    class = "gh_response",
    response = list(
      "x-ratelimit-limit" = "5000",
      "x-ratelimit-remaining" = "4999",
      "x-ratelimit-reset" = "1580507619"
    )
  )

  limit <- gh_rate_limit(mock_res)

  expect_equal(limit$limit, 5000L)
  expect_equal(limit$remaining, 4999L)
  expect_s3_class(limit$reset, "POSIXct") # Avoiding tz issues
})

test_that("errors", {
  local_fake_github()
  expect_snapshot(error = TRUE, {
    gh_rate_limit(list())
    gh_rate_limits(.token = "bad")
  })
})

test_that("gh_rate_limit fetches from /rate_limit when no response is passed", {
  local_fake_github()
  limit <- gh_rate_limit()
  expect_equal(limit$limit, 5000L)
  expect_equal(limit$remaining, 5000L)
  expect_s3_class(limit$reset, "POSIXct")
})

test_that("gh_rate_limits returns a data frame of all resource limits", {
  local_fake_github()
  limits <- gh_rate_limits()
  expect_s3_class(limits, "data.frame")
  expect_setequal(
    limits$type,
    c(
      "core",
      "search",
      "graphql",
      "integration_manifest",
      "code_scanning_upload"
    )
  )
  expect_true(all(limits$limit == 5000L))
  expect_true(all(limits$used == 0L))
  expect_true(all(limits$remaining == 5000L))
  expect_s3_class(limits$reset, "POSIXct")
  expect_type(limits$mins_left, "double")
})

test_that("missing rate limit", {
  mock_res <- structure(
    list(),
    class = "gh_response",
    response = list()
  )

  limit <- gh_rate_limit(mock_res)

  expect_equal(limit$limit, NA_integer_)
  expect_equal(limit$remaining, NA_integer_)
  expect_equal(as.double(limit$reset), NA_real_)
})

test_that("errors return a github_error object", {

  skip_if_offline("github.com")
  skip_on_cran()
  skip_if_no_token()

  e <- tryCatch(gh("/missing", .token = tt()), error = identity)

  expect_s3_class(e, "github_error")
  expect_s3_class(e, "http_error_404")
})

test_that("can catch a given status directly", {

  skip_if_offline("github.com")
  skip_on_cran()
  skip_if_no_token()

  e <- tryCatch(
    gh("/missing", .token = tt()),
    "http_error_404" = identity
  )

  expect_s3_class(e, "github_error")
  expect_s3_class(e, "http_error_404")
})

:: Turn work folder into GOPATH
set GOPATH=%SRC_DR%
set PATH=%GOPATH%\bin:%PATH%

:: Change to directory with main.go
cd cmd\gh
if errorlevel 1 exit 1

:: Build
go build -v -o %PKG_NAME%.exe .
if errorlevel 1 exit 1

:: Install Binary into %SCRIPTS%
if not exist %SCRIPTS% mkdir %SCRIPTS%
if errorlevel 1 exit 1

mv %PKG_NAME% %SCRIPTS%\%PKG_NAME%
if errorlevel 1 exit 1

#!/bin/bash

set -ex

if [[ $target_platform == "osx-arm64" ]]; then
    # When cross-compiling, build gen-docs and manpages separately as
    # they need to be executed on the build platform.
    export GOARCH=amd64
    make manpages
    export GOARCH=arm64
else
    make manpages
fi

# Build bin/gh
make bin/gh

# Install manually, since `make install` will recreate the manpages
mkdir -p $PREFIX/bin
install -m755 bin/gh ${PREFIX}/bin
mkdir -p $PREFIX/share/man/man1
install -m644 ./share/man/man1/* $PREFIX/share/man/man1/

@echo off

if "%CERT_FILE%" == "" (
    echo skipping Windows code-signing; CERT_FILE not set
    exit /b
)

.\script\signtool sign /d "GitHub CLI" /f "%CERT_FILE%" /p "%CERT_PASSWORD%" /fd sha256 /tr http://timestamp.digicert.com /v "%1"

param (
    [string]$Certificate = $(throw "-Certificate is required."),
    [string]$Executable = $(throw "-Executable is required.")
)

Set-StrictMode -Version Latest
$ErrorActionPreference = "Stop"

$thumbprint = "fb713a60a7fa79dfc03cb301ca05d4e8c1bdd431"
$passwd = $env:GITHUB_CERT_PASSWORD
$ProgramName = "GitHub CLI"

$scriptPath = split-path -parent $MyInvocation.MyCommand.Definition

& $scriptPath\signtool.exe sign /d $ProgramName /f $Certificate /p $passwd `
    /sha1 $thumbprint /fd sha256 /tr http://timestamp.digicert.com /td sha256 /v `
    $Executable

$scriptPath = split-path -parent $MyInvocation.MyCommand.Definition
$certFile = "$scriptPath\windows-certificate.pfx"

$headers = New-Object "System.Collections.Generic.Dictionary[[String],[String]]"
$headers.Add("Authorization", "token $env:DESKTOP_CERT_TOKEN")
$headers.Add("Accept", 'application/vnd.github.v3.raw')

Invoke-WebRequest 'https://api.github.com/repos/desktop/desktop-secrets/contents/windows-certificate.pfx' `
              -Headers $headers `
              -OutFile "$certFile"

Write-Output "::set-output name=cert-file::$certFile"

param (
    [string]$Certificate = $(throw "-Certificate is required."),
    [string]$Executable = $(throw "-Executable is required.")
)

Set-StrictMode -Version Latest
$ErrorActionPreference = "Stop"

$ProgramName = "GitHub CLI"
$scriptPath = split-path -parent $MyInvocation.MyCommand.Definition

& $scriptPath\signtool.exe sign /d $ProgramName /f $Certificate /p $env:CERT_PASSWORD /fd sha256 /tr http://timestamp.digicert.com /v $Executable

#!/bin/bash
set -e

EXE="$1"

if [ -z "$CERT_FILE" ]; then
  echo "skipping Windows code-signing; CERT_FILE not set" >&2
  exit 0
fi

if [ ! -f "$CERT_FILE" ]; then
  echo "error Windows code-signing; file '$CERT_FILE' not found" >&2
  exit 1
fi

if [ -z "$CERT_PASSWORD" ]; then
  echo "error Windows code-signing; no value for CERT_PASSWORD" >&2
  exit 1
fi

osslsigncode sign -n "GitHub CLI" -t http://timestamp.digicert.com \
  -pkcs12 "$CERT_FILE" -readpass <(printf "%s" "$CERT_PASSWORD") -h sha256 \
  -in "$EXE" -out "$EXE"~

mv "$EXE"~ "$EXE"

---
name: "\U0001F41B Bug report"
about: Report a bug or unexpected behavior while using GitHub CLI
title: ''
labels: bug
assignees: ''

---

### Describe the bug

A clear and concise description of what the bug is. Include version by typing `gh --version`.

### Steps to reproduce the behavior

1. Type this '...'
2. View the output '....'
3. See error

### Expected vs actual behavior

A clear and concise description of what you expected to happen and what actually happened.

### Logs

Paste the activity from your command line. Redact if needed.

---
name: "⭐ Submit a request"
about: Surface a feature or problem that you think should be solved
title: ''
labels: enhancement
assignees: ''

---

### Describe the feature or problem you’d like to solve

A clear and concise description of what the feature or problem is.

### Proposed solution

How will it benefit CLI and its users?

### Additional context

Add any other context like screenshots or mockups are helpful, if applicable.

# Releasing

To initiate a new production deployment:
```sh
script/release vX.Y.Z
```
See `script/release --help` for more information.

> **Note:**  
> Every production release will request an approval by the select few people before it can proceed.

What this does is:
- Builds Linux binaries on Ubuntu;
- Builds and signs Windows binaries on Windows;
- Builds, signs, and notarizes macOS binaries on macOS;
- Uploads all release artifacts to a new GitHub Release;
- A new git tag `vX.Y.Z` is created in the remote repository;
- The changelog is [generated from the list of merged pull requests](https://docs.github.com/en/repositories/releasing-projects-on-github/automatically-generated-release-notes);
- Updates cli.github.com with the contents of the new release;
- Updates our Homebrew formula in the homebrew-core repo.

To test out the build system while avoiding creating an actual release:
```sh
script/release --staging vX.Y.Z --branch patch-1 -p macos
```
The build artifacts will be available via `gh run download <RUN> -n macos`.

## General guidelines

* Features to be released should be reviewed and approved at least one day prior to the release.
* Feature releases should bump up the minor version number.
* Breaking releases should bump up the major version number. These should generally be rare.

## Test the build system locally

A local release can be created for testing without creating anything official on the release page.

1. Make sure GoReleaser is installed: `brew install goreleaser`
2. `script/release --local`
3. Find the built products under `dist/`.

# Installing gh on Linux and BSD

Packages downloaded from https://cli.github.com or from https://github.com/cli/cli/releases
are considered official binaries. We focus on popular Linux distros and
the following CPU architectures: `i386`, `amd64`, `arm64`, `armhf`.

Other sources for installation are community-maintained and thus might lag behind
our release schedule.

## Official sources

### Debian, Ubuntu Linux, Raspberry Pi OS (apt)

Install:

```bash
sudo mkdir -p -m 755 /etc/apt/keyrings && wget -qO- https://cli.github.com/packages/githubcli-archive-keyring.gpg | sudo tee /etc/apt/keyrings/githubcli-archive-keyring.gpg > /dev/null \
&& sudo chmod go+r /etc/apt/keyrings/githubcli-archive-keyring.gpg \
&& echo "deb [arch=$(dpkg --print-architecture) signed-by=/etc/apt/keyrings/githubcli-archive-keyring.gpg] https://cli.github.com/packages stable main" | sudo tee /etc/apt/sources.list.d/github-cli.list > /dev/null \
&& sudo apt update \
&& sudo apt install gh -y
```

> **Note**
> We were recently forced to change our GPG signing key. If you've previously downloaded the `githubcli-archive-keyring.gpg` file, you should re-download it again per above instructions. If you are using a keyserver to download the key, the ID of the new key is `23F3D4EA75716059`.

Upgrade:

```bash
sudo apt update
sudo apt install gh
```

### Fedora, CentOS, Red Hat Enterprise Linux (dnf)

Install from our package repository for immediate access to latest releases:

```bash
sudo dnf install 'dnf-command(config-manager)'
sudo dnf config-manager --add-repo https://cli.github.com/packages/rpm/gh-cli.repo
sudo dnf install gh
```

Alternatively, install from the [community repository](https://packages.fedoraproject.org/pkgs/gh/gh/):

```bash
sudo dnf install gh
```

Upgrade:

```bash
sudo dnf update gh
```

### Amazon Linux 2 (yum)

Install using our package repository for immediate access to latest releases:

```bash
type -p yum-config-manager >/dev/null || sudo yum install yum-utils
sudo yum-config-manager --add-repo https://cli.github.com/packages/rpm/gh-cli.repo
sudo yum install gh
```

> **Note**
> We were recently forced to change our GPG signing key. If you've added the repository previously and now you're getting a GPG signing key error, disable the repository first with `sudo yum-config-manager --disable gh-cli` and add it again with `sudo yum-config-manager --add-repo https://cli.github.com/packages/rpm/gh-cli.repo`.

Upgrade:

```bash
sudo yum update gh
```

### openSUSE/SUSE Linux (zypper)

Install:

```bash
sudo zypper addrepo https://cli.github.com/packages/rpm/gh-cli.repo
sudo zypper ref
sudo zypper install gh
```

Upgrade:

```bash
sudo zypper ref
sudo zypper update gh
```

## Manual installation

* [Download release binaries][releases page] that match your platform; or
* [Build from source](./source.md).

## Unofficial, community-supported methods

The GitHub CLI team does not maintain the following packages or repositories and thus we are unable to provide support for those installation methods.

### Snap (do not use)

There are [so many issues with Snap](https://github.com/casperdcl/cli/issues/7) as a runtime mechanism for apps like GitHub CLI that our team suggests _never installing gh as a snap_.

### Arch Linux

Arch Linux users can install from the [extra repo][arch linux repo]:

```bash
sudo pacman -S github-cli
```

Alternatively, use the [unofficial AUR package][arch linux aur] to build GitHub CLI from source.

### Android

Android 7+ users can install via [Termux](https://wiki.termux.com/wiki/Main_Page):

```bash
pkg install gh
```

### FreeBSD

FreeBSD users can install from the [ports collection](https://www.freshports.org/devel/gh/):

```bash
cd /usr/ports/devel/gh/ && make install clean
```

Or via [pkg(8)](https://www.freebsd.org/cgi/man.cgi?pkg(8)):

```bash
pkg install gh
```

### NetBSD/pkgsrc

NetBSD users and those on [platforms supported by pkgsrc](https://pkgsrc.org/#index4h1) can install the [gh package](https://pkgsrc.se/net/gh):

```bash
pkgin install gh
```

To install from source:

```bash
cd /usr/pkgsrc/net/gh && make package-install
```

### OpenBSD

In -current, or in releases starting from 7.0, OpenBSD users can install from packages:

```
pkg_add github-cli
```

### Funtoo

Funtoo Linux has an autogenerated github-cli package, located in [dev-kit](https://github.com/funtoo/dev-kit/tree/1.4-release/dev-util/github-cli), which can be installed in the following way:

``` bash
emerge -av github-cli
```

Upgrading can be done by syncing the repos and then requesting an upgrade:

``` bash
ego sync
emerge -u github-cli
```

### Gentoo

Gentoo Linux users can install from the [main portage tree](https://packages.gentoo.org/packages/dev-util/github-cli):

``` bash
emerge -av github-cli
```

Upgrading can be done by updating the portage tree and then requesting an upgrade:

``` bash
emerge --sync
emerge -u github-cli
```

### Kiss Linux

Kiss Linux users can install from the [community repos](https://github.com/kisslinux/community):

```bash
kiss b github-cli && kiss i github-cli
```

### Nix/NixOS

Nix/NixOS users can install from [nixpkgs](https://search.nixos.org/packages?query=gh&sort=relevance&show=gh):

```bash
nix-env -iA nixos.gh
```

### openSUSE Tumbleweed

openSUSE Tumbleweed users can install from the [official distribution repo](https://software.opensuse.org/package/gh):
```bash
sudo zypper in gh
```

### Alpine Linux

Alpine Linux users can install from the [stable releases' community package repository](https://pkgs.alpinelinux.org/packages?name=github-cli&branch=v3.15).

```bash
apk add github-cli
```

Users wanting the latest version of the CLI without waiting to be backported into the stable release they're using should use the edge release's
community repo through this method below, without mixing packages from stable and unstable repos.[^1]

```bash
echo "@community http://dl-cdn.alpinelinux.org/alpine/edge/community" >> /etc/apk/repositories
apk add github-cli@community
```

### Void Linux
Void Linux users can install from the [official distribution repo](https://voidlinux.org/packages/?arch=x86_64&q=github-cli):

```bash
sudo xbps-install github-cli
```

[releases page]: https://github.com/cli/cli/releases/latest
[arch linux repo]: https://www.archlinux.org/packages/extra/x86_64/github-cli
[arch linux aur]: https://aur.archlinux.org/packages/github-cli-git
[^1]: https://wiki.alpinelinux.org/wiki/Package_management#Repository_pinning

---
title: "Managing Personal Access Tokens"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Managing Personal Access Tokens}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(gh)
```

<!-- This vignette uses a convention of "one sentence per line" in prose. -->

gh generally sends a Personal Access Token (PAT) with its requests.
Some endpoints of the GitHub API can be accessed without authenticating yourself.
But once your API use becomes more frequent, you will want a PAT to prevent problems with rate limits and to access all possible endpoints.

This article describes how to store your PAT, so that gh can find it (automatically, in most cases). The function gh uses for this is `gh_token()`.

More resources on PAT management:

  * GitHub documentation on [Creating a personal access token](https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token)
  * In the [usethis package](https://usethis.r-lib.org):
    - Vignette: [Managing Git(Hub) Credentials](https://usethis.r-lib.org/articles/articles/git-credentials.html) 
    - `usethis::gh_token_help()` and `usethis::git_sitrep()` help you check if
      a PAT is discoverable and has suitable scopes
    - `usethis::create_github_token()` guides you through the process of getting
      a new PAT
  * In the [gitcreds package](https://r-lib.github.io/gitcreds/):
    - `gitcreds::gitcreds_set()` helps you explicitly put your PAT into the Git
      credential store
  
## PAT and host

`gh::gh()` allows the user to provide a PAT via the `.token` argument and to specify a host other than "github.com" via the `.api_url` argument.
(Some companies and universities run their own instance of GitHub Enterprise.)

```{r, eval = FALSE}
gh(endpoint, ..., .token = NULL, ..., .api_url = NULL, ...)
```

However, it's annoying to always provide your PAT or host and it's unsafe for your PAT to appear explicitly in your R code.
It's important to make it *possible* for the user to provide the PAT and/or API URL directly, but it should rarely be necessary.
`gh::gh()` is designed to play well with more secure, less fiddly methods for expressing what you want.

How are `.api_url` and `.token` determined when the user does not provide them?

  1. `.api_url` defaults to the value of the `GITHUB_API_URL` environment
    variable and, if that is unset, falls back to `"https://api.github.com"`.
    This is always done before worrying about the PAT.
  1. The PAT is obtained via a call to `gh_token(.api_url)`. That is, the token
    is looked up based on the host.

## The gitcreds package

gh now uses the gitcreds package to interact with the Git credential store.

gh calls `gitcreds::gitcreds_get()` with a URL to try to find a matching PAT.
`gitcreds::gitcreds_get()` checks session environment variables and then the local Git credential store.
Therefore, if you have previously used a PAT with, e.g., command line Git, gh may retrieve and re-use it.
You can call `gitcreds::gitcreds_get()` directly, yourself, if you want to see what is found for a specific URL.

``` r
gitcreds::gitcreds_get()
```

If you see something like this:
``` r
#> <gitcreds>
#>   protocol: https
#>   host    : github.com
#>   username: PersonalAccessToken
#>   password: <-- hidden -->
```
that means that gitcreds could get the PAT from the Git credential store.
You can call `gitcreds_get()$password` to see the actual PAT.

If no matching PAT is found, `gitcreds::gitcreds_get()` errors.

## PAT in an environment variable

If you don't have a Git installation, or your Git installation does not have a working credential store, then you can specify the PAT in an environment variable.
For `github.com` you can set the `GITHUB_PAT_GITHUB_COM` or `GITHUB_PAT` variable.
For a different GitHub host, call `gitcreds::gitcreds_cache_envvar()` with the API URL to see the environment variable you need to set.
For example:

```{r}
gitcreds::gitcreds_cache_envvar("https://github.acme.com")
```

## Recommendations

On a machine used for interactive development, we recommend:

  * Store your PAT(s) in an official credential store.
  * Do **not** store your PAT(s) in plain text in, e.g., `.Renviron`. In the
    past, this has been a common and recommended practice for pragmatic reasons.
    However, gitcreds/gh have now evolved to the point where it's
    possible for all of us to follow better security practices.
  * If you use a general-purpose password manager, like 1Password or LastPass,
    you may *also* want to store your PAT(s) there. Why? If your PAT is
    "forgotten" from the OS-level credential store, intentionally or not, you'll
    need to provide it again when prompted.
    
    If you don't have any other record of your PAT, you'll have to get a new
    PAT whenever this happens. This is not the end of the world. But if you
    aren't disciplined about deleting lost PATs from
    <https://github.com/settings/tokens>, you will eventually find yourself in a
    confusing situation where you can't be sure which PAT(s) are in use.

On a headless system, such as on a CI/CD platform, provide the necessary PAT(s) via secure environment variables.
Regular environment variables can be used to configure less sensitive settings, such as the API host.
Don't expose your PAT by doing something silly like dumping all environment variables to a log file.

Note that on GitHub Actions, specifically, a personal access token is [automatically available to the workflow](https://docs.github.com/en/actions/configuring-and-managing-workflows/authenticating-with-the-github_token) as the `GITHUB_TOKEN` secret.
That is why many workflows in the R community contain this snippet:

``` yaml
env:
  GITHUB_PAT: ${{ secrets.GITHUB_TOKEN }}
```

This makes the automatic PAT available as the `GITHUB_PAT` environment variable.
If that PAT doesn't have the right permissions, then you'll need to explicitly provide one that does (see link above for more).

## Failure

If there is no PAT to be had, `gh::gh()` sends a request with no token.
(Internally, the `Authorization` header is omitted if the PAT is found to be the empty string, `""`.)

What do PAT-related failures look like?

If no PAT is sent and the endpoint requires no auth, the request probably succeeds!
At least until you run up against rate limits.
If the endpoint requires auth, you'll get an HTTP error, possibly this one:

```
GitHub API error (401): 401 Unauthorized
Message: Requires authentication
```

If a PAT is first discovered in an environment variable, it is taken at face value.
The two most common ways to arrive here are PAT specification via `.Renviron` or as a secret in a CI/CD platform, such as GitHub Actions.
If the PAT is invalid, the first affected request will fail, probably like so:

```
GitHub API error (401): 401 Unauthorized
Message: Bad credentials
```

This will also be the experience if an invalid PAT is provided directly via `.token`.

Even a valid PAT can lead to a downstream error, if it has insufficient scopes with respect to a specific request.

# Releasing

Our build system automatically compiles and attaches cross-platform binaries to any git tag named `vX.Y.Z`. The changelog is [generated from git commit log](https://docs.github.com/en/repositories/releasing-projects-on-github/automatically-generated-release-notes).

Users who run official builds of `gh` on their machines will get notified about the new version within a 24 hour period.

To test out the build system, publish a prerelease tag with a name such as `vX.Y.Z-pre.0` or `vX.Y.Z-rc.1`. Note that such a release will still be public, but it will be marked as a "prerelease", meaning that it will never show up as a "latest" release nor trigger upgrade notifications for users.

## General guidelines

* Features to be released should be reviewed and approved at least one day prior to the release.
* Feature releases should bump up the minor version number.

## Tagging a new release

1. `git tag v1.2.3 && git push origin v1.2.3`
2. Wait several minutes for builds to run: <https://github.com/cli/cli/actions>
3. Verify release is displayed and has correct assets: <https://github.com/cli/cli/releases>
4. Scan generated release notes and optionally add a human touch by grouping items under topic sections
5. Verify the marketing site was updated: <https://cli.github.com>
6. (Optional) Delete any pre-releases related to this release

A successful build will result in changes across several repositories:
* <https://github.com/github/cli.github.com>
* <https://github.com/Homebrew/homebrew-core/pulls>
* <https://github.com/cli/scoop-gh>

If the build fails, there is not a clean way to re-run it. The easiest way would be to start over by deleting the partial release on GitHub and re-publishing the tag. Note that this might be disruptive to users or tooling that were already notified about an upgrade. If a functional release and its binaries are already out there, it might be better to try to manually fix up only the specific workflow tasks that failed. Use your best judgement depending on the failure type.

## Release locally for debugging

A local release can be created for testing without creating anything official on the release page.

1. Make sure GoReleaser is installed: `brew install goreleaser`
2. `goreleaser --skip-validate --skip-publish --rm-dist`
3. Find the built products under `dist/`.

# Installing gh on Linux and BSD

Packages downloaded from https://cli.github.com or from https://github.com/cli/cli/releases
are considered official binaries. We focus on popular Linux distros and
the following CPU architectures: `i386`, `amd64`, `arm64`, `armhf`.

Other sources for installation are community-maintained and thus might lag behind
our release schedule.

## Official sources

### Debian, Ubuntu Linux, Raspberry Pi OS (apt)

Install:

```bash
curl -fsSL https://cli.github.com/packages/githubcli-archive-keyring.gpg | sudo dd of=/usr/share/keyrings/githubcli-archive-keyring.gpg
sudo chmod go+r /usr/share/keyrings/githubcli-archive-keyring.gpg
echo "deb [arch=$(dpkg --print-architecture) signed-by=/usr/share/keyrings/githubcli-archive-keyring.gpg] https://cli.github.com/packages stable main" | sudo tee /etc/apt/sources.list.d/github-cli.list > /dev/null
sudo apt update
sudo apt install gh
```

Upgrade:

```bash
sudo apt update
sudo apt install gh
```

### Fedora, CentOS, Red Hat Enterprise Linux (dnf)

Install from our package repository for immediate access to latest releases:

```bash
sudo dnf install 'dnf-command(config-manager)'
sudo dnf config-manager --add-repo https://cli.github.com/packages/rpm/gh-cli.repo
sudo dnf install gh
```

Alternatively, install from the [community repository](https://packages.fedoraproject.org/pkgs/gh/gh/):

```bash
sudo dnf install gh
```

Upgrade:

```bash
sudo dnf update gh
```

### openSUSE/SUSE Linux (zypper)

Install:

```bash
sudo zypper addrepo https://cli.github.com/packages/rpm/gh-cli.repo
sudo zypper ref
sudo zypper install gh
```

Upgrade:

```bash
sudo zypper ref
sudo zypper update gh
```

## Manual installation

* [Download release binaries][releases page] that match your platform; or
* [Build from source](./source.md).

## Unofficial, community-supported methods

The GitHub CLI team does not maintain the following packages or repositories and thus we are unable to provide support for those installation methods.

### Snap (do not use)

There are [so many issues with Snap](https://github.com/casperdcl/cli/issues/7) as a runtime mechanism for apps like GitHub CLI that our team suggests _never installing gh as a snap_.

### Arch Linux

Arch Linux users can install from the [community repo][arch linux repo]:

```bash
sudo pacman -S github-cli
```

Alternatively, use the [unofficial AUR package][arch linux aur] to build GitHub CLI from source.

### Android

Android 7+ users can install via [Termux](https://wiki.termux.com/wiki/Main_Page):

```bash
pkg install gh
```

### FreeBSD

FreeBSD users can install from the [ports collection](https://www.freshports.org/devel/gh/):

```bash
cd /usr/ports/devel/gh/ && make install clean
```

Or via [pkg(8)](https://www.freebsd.org/cgi/man.cgi?pkg(8)):

```bash
pkg install gh
```

### NetBSD/pkgsrc

NetBSD users and those on [platforms supported by pkgsrc](https://pkgsrc.org/#index4h1) can install the [gh package](https://pkgsrc.se/net/gh):

```bash
pkgin install gh
```

To install from source:

```bash
cd /usr/pkgsrc/net/gh && make package-install
```

### OpenBSD

In -current, or in releases starting from 7.0, OpenBSD users can install from packages:

```
pkg_add github-cli
```

### Funtoo

Funtoo Linux has an autogenerated github-cli package, located in [dev-kit](https://github.com/funtoo/dev-kit/tree/1.4-release/dev-util/github-cli), which can be installed in the following way:

``` bash
emerge -av github-cli
```

Upgrading can be done by syncing the repos and then requesting an upgrade:

``` bash
ego sync
emerge -u github-cli
```

### Gentoo

Gentoo Linux users can install from the [main portage tree](https://packages.gentoo.org/packages/dev-util/github-cli):

``` bash
emerge -av github-cli
```

Upgrading can be done by updating the portage tree and then requesting an upgrade:

``` bash
emerge --sync
emerge -u github-cli
```

### Kiss Linux

Kiss Linux users can install from the [community repos](https://github.com/kisslinux/community):

```bash
kiss b github-cli && kiss i github-cli
```

### Nix/NixOS

Nix/NixOS users can install from [nixpkgs](https://search.nixos.org/packages?show=gitAndTools.gh&query=gh&from=0&size=30&sort=relevance&channel=20.03#disabled):

```bash
nix-env -iA nixos.gitAndTools.gh
```

### openSUSE Tumbleweed

openSUSE Tumbleweed users can install from the [official distribution repo](https://software.opensuse.org/package/gh):
```bash
sudo zypper in gh
```

### Alpine Linux

Alpine Linux users can install from the [stable releases' community package repository](https://pkgs.alpinelinux.org/packages?name=github-cli&branch=v3.15).

```bash
apk add github-cli
```

Users wanting the latest version of the CLI without waiting to be backported into the stable release they're using should use the edge release's
community repo through this method below, without mixing packages from stable and unstable repos.[^1]

```bash
echo "@community http://dl-cdn.alpinelinux.org/alpine/edge/community" >> /etc/apk/repositories
apk add github-cli@community
```

### Void Linux
Void Linux users can install from the [official distribution repo](https://voidlinux.org/packages/?arch=x86_64&q=github-cli):

```bash
sudo xbps-install github-cli
```

[releases page]: https://github.com/cli/cli/releases/latest
[arch linux repo]: https://www.archlinux.org/packages/community/x86_64/github-cli
[arch linux aur]: https://aur.archlinux.org/packages/github-cli-git
[^1]: https://wiki.alpinelinux.org/wiki/Package_management#Repository_pinning

# Triage role

As we get more issues and pull requests opened on the GitHub CLI, we've decided on a weekly rotation
triage role. The initial expectation is that the person in the role for the week spends no more than
2 hours a day on this work; we can refine that as needed.

## Expectations for incoming issues

Review and label [open issues missing either the `enhancement`, `bug`, or `docs` label](https://github.com/cli/cli/issues?q=is%3Aopen+is%3Aissue+-label%3Abug%2Cenhancement%2Cdocs+).

To be considered triaged, `enhancement` issues require at least one of the following additional labels:

- `core`: reserved for the core CLI team
- `help wanted`: signal that we are accepting contributions for this
- `discuss`: add to our team's queue to discuss during a sync
- `needs-investigation`: work that requires a mystery be solved by the core team before it can move forward
- `needs-user-input`: we need more information from our users before the task can move forward

To be considered triaged, `bug` issues require a severity label: one of `p1`, `p2`, or `p3`, which are defined as follows:
 - `p1`: Affects a large population and inhibits work
 - `p2`: Affects more than a few users but doesn't prevent core functions
 - `p3`: Affects a small number of users or is largely cosmetic

## Expectations for community pull requests

All incoming pull requests are assigned to one of the engineers for review on a round-robin basis.
The person in a triage role for a week could take a glance at these pull requests, mostly to see whether
the changeset is feasible and to allow the associated CI run for new contributors.

## Issue triage flowchart

- can this be closed outright?
  - e.g. spam/junk
  - close without comment
- do we not want to do it?
  - e.g. have already discussed not wanting to do or duplicate issue
  - comment and close
- are we ok with outside contribution for this?
  - e.g. the task is relatively straightforward, but no people on our team have the bandwidth to take it on at the moment
  - ensure that the thread contains all the context necessary for someone new to pick this up
  - add `help wanted` label
  - consider adding `good first issue` label
- do we want to do it?
  - comment acknowledging that
  - add `core` label
  - add to the project “TODO” column if this is something that should ship soon
- is it intriguing, but requires discussion?
  - label `discuss`
  - label `needs-investigation` if engineering research is required before action can be taken
- does it need more info from the issue author?
  - ask the user for details
  - add `needs-user-input` label
- is it a usage/support question?
  - consider converting the Issue to a Discussion

## Weekly PR audit

In the interest of not letting our open PR list get out of hand (20+ total PRs _or_ multiple PRs
over a few months old), try to audit open PRs each week with the goal of getting them merged and/or
closed. It's likely too much work to deal with every PR, but even getting a few closer to done is
helpful.

For each PR, ask:

- is this too stale (more than two months old or too many conflicts)? close with comment
- is this really close but author is absent? push commits to finish, request review
- is this waiting on triage? go through the PR triage flow

## Useful aliases

This gist has some useful aliases for first responders:

https://gist.github.com/vilmibm/ee6ed8a783e4fef5b69b2ed42d743b1a

## Examples

We want our project to be a safe and encouraging open-source environment. Below are some examples
of how to empathetically respond to or close an issue/PR:

- [Closing a quality PR its scope is too large](https://github.com/cli/cli/pull/1161)
- [Closing a stale PR](https://github.com/cli/cli/pull/557#issuecomment-639077269)
- [Closing a PR that doesn't follow our CONTRIBUTING policy](https://github.com/cli/cli/pull/864)
- [Responding to a bug report](https://github.com/desktop/desktop/issues/9195#issuecomment-592243129)
- [Closing an issue that out of scope](https://github.com/cli/cli/issues/777#issuecomment-612926229)
- [Closing an issue with a feature request](https://github.com/desktop/desktop/issues/9722#issuecomment-625461766)

# Triage role

As we get more issues and pull requests opened on the GitHub CLI, we've decided on a weekly rotation
triage role. The initial expectation is that the person in the role for the week spends no more than
2 hours a day on this work; we can refine that as needed.

## Expectations for incoming issues

All incoming issues need either an `enhancement`, `bug`, or `docs` label.

To be considered triaged, `enhancement` issues require at least one of the following additional labels:

- `core`: reserved for the core CLI team
- `help wanted`: signal that we are accepting contributions for this
- `discuss`: add to our team's queue to discuss during a sync
- `needs-investigation`: work that requires a mystery be solved by the core team before it can move forward
- `needs-user-input`: we need more information from our users before the task can move forward

To be considered triaged, `bug` issues require a severity label: one of `p1`, `p2`, or `p3`

## Expectations for community pull requests

All incoming pull requests are assigned to one of the engineers for review on a round-robin basis.
The person in a triage role for a week could take a glance at these pull requests, mostly to see whether
the changeset is feasible and to allow the associated CI run for new contributors.

## Issue triage flowchart

- can this be closed outright?
  - e.g. spam/junk
  - close without comment
- do we not want to do it?
  - e.g. have already discussed not wanting to do or duplicate issue
  - comment and close
- are we ok with outside contribution for this?
  - e.g. the task is relatively straightforward, but no people on our team have the bandwidth to take it on at the moment
  - ensure that the thread contains all the context necessary for someone new to pick this up
  - add `help wanted` label
  - consider adding `good first issue` label
- do we want to do it?
  - comment acknowledging that
  - add `core` label
  - add to the project “TODO” column if this is something that should ship soon
- is it intriguing, but requires discussion?
  - label `discuss`
  - label `needs-investigation` if engineering research is required before action can be taken
- does it need more info from the issue author?
  - ask the user for details
  - add `needs-user-input` label
- is it a usage/support question?
  - consider converting the Issue to a Discussion

## Weekly PR audit

In the interest of not letting our open PR list get out of hand (20+ total PRs _or_ multiple PRs
over a few months old), try to audit open PRs each week with the goal of getting them merged and/or
closed. It's likely too much work to deal with every PR, but even getting a few closer to done is
helpful.

For each PR, ask:

- is this too stale (more than two months old or too many conflicts)? close with comment
- is this really close but author is absent? push commits to finish, request review
- is this waiting on triage? go through the PR triage flow

## Useful aliases

This gist has some useful aliases for first responders:

https://gist.github.com/vilmibm/ee6ed8a783e4fef5b69b2ed42d743b1a

## Examples

We want our project to be a safe and encouraging open-source environment. Below are some examples
of how to empathetically respond to or close an issue/PR:

- [Closing a quality PR its scope is too large](https://github.com/cli/cli/pull/1161)
- [Closing a stale PR](https://github.com/cli/cli/pull/557#issuecomment-639077269)
- [Closing a PR that doesn't follow our CONTRIBUTING policy](https://github.com/cli/cli/pull/864)
- [Responding to a bug report](https://github.com/desktop/desktop/issues/9195#issuecomment-592243129)
- [Closing an issue that out of scope](https://github.com/cli/cli/issues/777#issuecomment-612926229)
- [Closing an issue with a feature request](https://github.com/desktop/desktop/issues/9722#issuecomment-625461766)

# Installation from source

1. Verify that you have Go 1.21+ installed

   ```sh
   $ go version
   ```

   If `go` is not installed, follow instructions on [the Go website](https://golang.org/doc/install).

2. Clone this repository

   ```sh
   $ git clone https://github.com/cli/cli.git gh-cli
   $ cd gh-cli
   ```

3. Build and install

   #### Unix-like systems
   ```sh
   # installs to '/usr/local' by default; sudo may be required, or sudo -E for configured go environments
   $ make install

   # or, install to a different location
   $ make install prefix=/path/to/gh
   ```

   #### Windows
   ```pwsh
   # build the `bin\gh.exe` binary
   > go run script\build.go
   ```
   There is no install step available on Windows.

4. Run `gh version` to check if it worked.

   #### Windows
   Run `bin\gh version` to check if it worked.

## Cross-compiling binaries for different platforms

You can use any platform with Go installed to build a binary that is intended for another platform
or CPU architecture. This is achieved by setting environment variables such as GOOS and GOARCH.

For example, to compile the `gh` binary for the 32-bit Raspberry Pi OS:
```sh
# on a Unix-like system:
$ GOOS=linux GOARCH=arm GOARM=7 CGO_ENABLED=0 make clean bin/gh
```
```pwsh
# on Windows, pass environment variables as arguments to the build script:
> go run script\build.go clean bin\gh GOOS=linux GOARCH=arm GOARM=7 CGO_ENABLED=0
```

Run `go tool dist list` to list all supported values of GOOS/GOARCH.

Tip: to reduce the size of the resulting binary, you can use `GO_LDFLAGS="-s -w"`. This omits
symbol tables used for debugging. See the list of [supported linker flags](https://golang.org/cmd/link/).

---
name: "\U0001F4E3 Feedback"
about: Give us general feedback about the GitHub CLI
title: ''
labels: feedback
assignees: ''

---

# CLI Feedback

You can use this template to give us structured feedback or just wipe it and leave us a note. Thank you!

## What have you loved?

_eg "the nice colors"_

## What was confusing or gave you pause?

_eg "it did something unexpected"_

## Are there features you'd like to see added?

_eg "gh cli needs mini-games"_

## Anything else?

_eg "have a nice day"_

## Contributing

Hi! Thanks for your interest in contributing to the GitHub CLI!

We accept pull requests for bug fixes and features where we've discussed the approach in an issue and given the go-ahead for a community member to work on it. We'd also love to hear about ideas for new features as issues.

Please do:

* Check existing issues to verify that the [bug][bug issues] or [feature request][feature request issues] has not already been submitted.
* Open an issue if things aren't working as expected.
* Open an issue to propose a significant change.
* Open a pull request to fix a bug.
* Open a pull request to fix documentation about a command.
* Open a pull request for any issue labelled [`help wanted`][hw] or [`good first issue`][gfi].

Please avoid:

* Opening pull requests for issues marked `needs-design`, `needs-investigation`, or `blocked`.
* Adding installation instructions specifically for your OS/package manager.
* Opening pull requests for any issue marked `core`. These issues require additional context from
  the core CLI team at GitHub and any external pull requests will not be accepted.

## Building the project

Prerequisites:
- Go 1.16+

Build with:
* Unix-like systems: `make`
* Windows: `go run script/build.go`

Run the new binary as:
* Unix-like systems: `bin/gh`
* Windows: `bin\gh`

Run tests with: `go test ./...`

See [project layout documentation](../docs/project-layout.md) for information on where to find specific source files.

## Submitting a pull request

1. Create a new branch: `git checkout -b my-branch-name`
1. Make your change, add tests, and ensure tests pass
1. Submit a pull request: `gh pr create --web`

Contributions to this project are [released][legal] to the public under the [project's open source license][license].

Please note that this project adheres to a [Contributor Code of Conduct][code-of-conduct]. By participating in this project you agree to abide by its terms.

We generate manual pages from source on every release. You do not need to submit pull requests for documentation specifically; manual pages for commands will automatically get updated after your pull requests gets accepted.

## Design guidelines

You may reference the [CLI Design System][] when suggesting features, and are welcome to use our [Google Docs Template][] to suggest designs.

## Resources

- [How to Contribute to Open Source][]
- [Using Pull Requests][]
- [GitHub Help][]


[bug issues]: https://github.com/cli/cli/issues?q=is%3Aopen+is%3Aissue+label%3Abug
[feature request issues]: https://github.com/cli/cli/issues?q=is%3Aopen+is%3Aissue+label%3Aenhancement
[hw]: https://github.com/cli/cli/labels/help%20wanted
[gfi]: https://github.com/cli/cli/labels/good%20first%20issue
[legal]: https://docs.github.com/en/free-pro-team@latest/github/site-policy/github-terms-of-service#6-contributions-under-repository-license
[license]: ../LICENSE
[code-of-conduct]: ./CODE-OF-CONDUCT.md
[How to Contribute to Open Source]: https://opensource.guide/how-to-contribute/
[Using Pull Requests]: https://docs.github.com/en/free-pro-team@latest/github/collaborating-with-issues-and-pull-requests/about-pull-requests
[GitHub Help]: https://docs.github.com/
[CLI Design System]: https://primer.style/cli/
[Google Docs Template]: https://docs.google.com/document/d/1JIRErIUuJ6fTgabiFYfCH3x91pyHuytbfa0QLnTfXKM/edit#heading=h.or54sa47ylpg

## Contributing

Hi! Thanks for your interest in contributing to the GitHub CLI!

We accept pull requests for bug fixes and features where we've discussed the approach in an issue and given the go-ahead for a community member to work on it. We'd also love to hear about ideas for new features as issues.

Please do:

* Check existing issues to verify that the [bug][bug issues] or [feature request][feature request issues] has not already been submitted.
* Open an issue if things aren't working as expected.
* Open an issue to propose a significant change.
* Open a pull request to fix a bug.
* Open a pull request to fix documentation about a command.
* Open a pull request for any issue labelled [`help wanted`][hw] or [`good first issue`][gfi].

Please avoid:

* Opening pull requests for issues marked `needs-design`, `needs-investigation`, or `blocked`.
* Opening pull requests that haven't been approved for work in an issue
* Adding installation instructions specifically for your OS/package manager.
* Opening pull requests for any issue marked `core`. These issues require additional context from
  the core CLI team at GitHub and any external pull requests will not be accepted.

## Building the project

Prerequisites:
- Go 1.21+

Build with:
* Unix-like systems: `make`
* Windows: `go run script/build.go`

Run the new binary as:
* Unix-like systems: `bin/gh`
* Windows: `bin\gh`

Run tests with: `go test ./...`

See [project layout documentation](../docs/project-layout.md) for information on where to find specific source files.

## Submitting a pull request

1. Create a new branch: `git checkout -b my-branch-name`
1. Make your change, add tests, and ensure tests pass
1. Submit a pull request: `gh pr create --web`

Contributions to this project are [released][legal] to the public under the [project's open source license][license].

Please note that this project adheres to a [Contributor Code of Conduct][code-of-conduct]. By participating in this project you agree to abide by its terms.

We generate manual pages from source on every release. You do not need to submit pull requests for documentation specifically; manual pages for commands will automatically get updated after your pull requests gets accepted.

## Design guidelines

You may reference the [CLI Design System][] when suggesting features, and are welcome to use our [Google Docs Template][] to suggest designs.

## Resources

- [How to Contribute to Open Source][]
- [Using Pull Requests][]
- [GitHub Help][]


[bug issues]: https://github.com/cli/cli/issues?q=is%3Aopen+is%3Aissue+label%3Abug
[feature request issues]: https://github.com/cli/cli/issues?q=is%3Aopen+is%3Aissue+label%3Aenhancement
[hw]: https://github.com/cli/cli/labels/help%20wanted
[gfi]: https://github.com/cli/cli/labels/good%20first%20issue
[legal]: https://docs.github.com/en/free-pro-team@latest/github/site-policy/github-terms-of-service#6-contributions-under-repository-license
[license]: ../LICENSE
[code-of-conduct]: ./CODE-OF-CONDUCT.md
[How to Contribute to Open Source]: https://opensource.guide/how-to-contribute/
[Using Pull Requests]: https://docs.github.com/en/free-pro-team@latest/github/collaborating-with-issues-and-pull-requests/about-pull-requests
[GitHub Help]: https://docs.github.com/
[CLI Design System]: https://primer.style/cli/
[Google Docs Template]: https://docs.google.com/document/d/1JIRErIUuJ6fTgabiFYfCH3x91pyHuytbfa0QLnTfXKM/edit#heading=h.or54sa47ylpg

# GitHub CLI project layout

At a high level, these areas make up the `github.com/cli/cli` project:
- [`cmd/`](../cmd) - `main` packages for building binaries such as the `gh` executable
- [`pkg/`](../pkg) - most other packages, including the implementation for individual gh commands
- [`docs/`](../docs) - documentation for maintainers and contributors
- [`script/`](../script) - build and release scripts
- [`internal/`](../internal) - Go packages highly specific to our needs and thus internal
- [`go.mod`](../go.mod) - external Go dependencies for this project, automatically fetched by Go at build time

Some auxiliary Go packages are at the top level of the project for historical reasons:
- [`api/`](../api) - main utilities for making requests to the GitHub API
- [`context/`](../context) - DEPRECATED: use only for referencing git remotes
- [`git/`](../git) - utilities to gather information from a local git repository
- [`test/`](../test) - DEPRECATED: do not use
- [`utils/`](../utils) - DEPRECATED: use only for printing table output

## Command-line help text

Running `gh help issue list` displays help text for a topic. In this case, the topic is a specific command,
and help text for every command is embedded in that command's source code. The naming convention for gh
commands is:
```
pkg/cmd/<command>/<subcommand>/<subcommand>.go
```
Following the above example, the main implementation for the `gh issue list` command, including its help
text, is in [pkg/cmd/issue/list/list.go](../pkg/cmd/issue/list/list.go)

Other help topics not specific to any command, for example `gh help environment`, are found in
[pkg/cmd/root/help_topic.go](../pkg/cmd/root/help_topic.go).

During our release process, these help topics are [automatically converted](../cmd/gen-docs/main.go) to
manual pages and published under https://cli.github.com/manual/.

## How GitHub CLI works

To illustrate how GitHub CLI works in its typical mode of operation, let's build the project, run a command,
and talk through which code gets run in order.

1. `go run script/build.go` - Makes sure all external Go dependencies are fetched, then compiles the
   `cmd/gh/main.go` file into a `bin/gh` binary.
2. `bin/gh issue list --limit 5` - Runs the newly built `bin/gh` binary (note: on Windows you must use
   backslashes like `bin\gh`) and passes the following arguments to the process: `["issue", "list", "--limit", "5"]`.
3. `func main()` inside `cmd/gh/main.go` is the first Go function that runs. The arguments passed to the
   process are available through `os.Args`.
4. The `main` package initializes the "root" command with `root.NewCmdRoot()` and dispatches execution to it
   with `rootCmd.ExecuteC()`.
5. The [root command](../pkg/cmd/root/root.go) represents the top-level `gh` command and knows how to
   dispatch execution to any other gh command nested under it.
6. Based on `["issue", "list"]` arguments, the execution reaches the `RunE` block of the `cobra.Command`
   within [pkg/cmd/issue/list/list.go](../pkg/cmd/issue/list/list.go).
7. The `--limit 5` flag originally passed as arguments be automatically parsed and its value stored as
   `opts.LimitResults`.
8. `func listRun()` is called, which is responsible for implementing the logic of the `gh issue list` command.
9. The command collects information from sources like the GitHub API then writes the final output to
   standard output and standard error [streams](../pkg/iostreams/iostreams.go) available at `opts.IO`.
10. The program execution is now back at `func main()` of `cmd/gh/main.go`. If there were any Go errors as a
    result of processing the command, the function will abort the process with a non-zero exit status.
    Otherwise, the process ends with status 0 indicating success.

## How to add a new command

1. First, check on our issue tracker to verify that our team had approved the plans for a new command.
2. Create a package for the new command, e.g. for a new command `gh boom` create the following directory
   structure: `pkg/cmd/boom/`
3. The new package should expose a method, e.g. `NewCmdBoom()`, that accepts a `*cmdutil.Factory` type and
   returns a `*cobra.Command`.
   * Any logic specific to this command should be kept within the command's package and not added to any
     "global" packages like `api` or `utils`.
4. Use the method from the previous step to generate the command and add it to the command tree, typically
   somewhere in the `NewCmdRoot()` method.

## How to write tests

This task might be tricky. Typically, gh commands do things like look up information from the git repository
in the current directory, query the GitHub API, scan the user's `~/.ssh/config` file, clone or fetch git
repositories, etc. Naturally, none of these things should ever happen for real when running tests, unless
you are sure that any filesystem operations are strictly scoped to a location made for and maintained by the
test itself. To avoid actually running things like making real API requests or shelling out to `git`
commands, we stub them. You should look at how that's done within some existing tests.

To make your code testable, write small, isolated pieces of functionality that are designed to be composed
together. Prefer table-driven tests for maintaining variations of different test inputs and expectations
when exercising a single piece of functionality.

# Working with the GitHub CLI Team: Hubber Edition

POV: your team at GitHub is interested in shipping a new command in `gh`.

This document outlines the process the CLI team prefers for helping ensure success both for your new feature and the CLI project as a whole.

## Step 0: Create an extension

Even if you want to see your code merged into `gh`, you should start with [an extension](https://docs.github.com/en/github-cli/github-cli/creating-github-cli-extensions) written in Go and leveraging [go-gh](https://github.com/cli/go-gh). Though `gh` extensions can be written in any language, we treat Go as a first class experience and ship a library of helpers for extensions written in Go.

Creating an extension enables you to start prototyping immediately, without waiting for us, and gives us something tangible to review if you decide you'd like the work incorporated into `gh`. It also means that you can decide to simply release your work without waiting for us to merge it, which leaves you in charge of release scheduling moving forward.

If you know from this point that you're comfortable with your new feature being an extension, don't worry about the rest of this document. We don't dictate how people create and release `gh` extensions.

If you do want your feature merged into `gh`, read on.

## Step 1: UX review

No matter what state your code is in, open up an issue either in [the open source cli/cli repository](https://github.com/cli/cli) or, if you'd rather not make the new feature public yet, [the closed github/cli repository](https://github.com/github/cli).

Describe how your new command would be used. Include mock-up examples, including a mock-up of what usage information would be printed if a user ran your command with `--help`.

We take this step seriously because we believe in keeping `gh`'s interface consistent and intuitive.

## Step 2: Beta

Once we've signed off on the proposed UX on the issue opened in step 1, develop your extension to at least beta quality. It's up to you if you actually want to go through a beta release phase with real users or not.

## Step 3: Merge or no merge

With a beta in hand it's time to decide whether or not to mainline your extension into the `trunk` of `gh`. Some questions to consider:

- How complex is the support burden for your feature?

If this feature requires extensive or specialized support, you will either need to release it as an extension or work with the CLI team to get maintainer access to `cli/cli`. The CLI team is very small and cannot promise any kind of SLA for supporting your work. For example, the `gh cs` command is sufficiently specialized and complex that we have given the `codespaces` team write access to the repository to maintain their own pull request review process. We have not put it in an extension as Codespaces are a core GitHub product with widespread use among our users.

- What kind of release cadence do you want?

We do a `gh` release roughly every other week, but if the changeset for a given week is light we may skip one. We make no official promise as to our cadence, and while we do have an on-call rotation there is no guarantee that you'll be able to get emergency fixes out within hours. If this is troubling, consider keeping your work in an extension.

- What kind of audience are you trying to reach?

Is this new feature intended for all GitHub users or just a few? If it's as applicable to your average GitHub user or customer as something like Codespaces or Pull Requests, that's a strong indication it should be merged into `trunk`. If not, consider keeping it an extension.

If after all of this consideration you think your feature should be merged, please open an issue in [cli/cli](https://github.com/cli/cli) with a link to your extension's code. It will go into our triage queue and we'll confirm that merging into `trunk` is feasible and appropriate.

## Step 4

Once we've signed off, open up a pull request in [cli/cli](https://github.com/cli/cli) adding your command. Since we make use of `go-gh` within our code already, it shouldn't be too onerous to make your extension merge-able. Link to the issue you opened in step 3 so we have some context on the pull request.

Keep in mind that our expectation of non-trivial commands that end up merged into `cli/cli` is that your team will continue to maintain what they merged over time. We can help redirect issues your way as part of our first responder rotation, but are unable to take on the full support burden for your new command.

## Other considerations

- If you have a high need for secrecy until the point of release, let us know in [#cli on slack](https://github.slack.com/archives/CLLG3RMAR). We'll come up with a solution to work on merging your command in private.
- We are a highly asynchronous team due to wide timezone differences. The best way to get in touch with us is via issue and pull request comments to which we'll respond within 24 hours. You can ping us on Slack but that's generally not our preference.
- We are happy to pair with you on extension authoring! Just let us know if we can provide guidance and we can schedule synchronous time to work together with you.

# get_validate_tokens_mode() rejects a setting that isn't a single string

    Code
      gh_pat(bad)
    Condition
      Error in `get_validate_tokens_mode()`:
      ! Invalid token validation setting: must be a single string.
      i Got a character vector.

---

    Code
      gh_pat(bad)
    Condition
      Error in `get_validate_tokens_mode()`:
      ! Invalid token validation setting: must be a single string.
      i Got an integer.

---

    Code
      gh_pat(bad)
    Condition
      Error in `get_validate_tokens_mode()`:
      ! Invalid token validation setting: must be a single string.
      i Got a character `NA`.

# new_gh_pat rejects non-string input

    Code
      new_gh_pat(1L)
    Condition
      Error in `new_gh_pat()`:
      ! A GitHub PAT must be a string

---

    Code
      new_gh_pat(c("a", "b"))
    Condition
      Error in `new_gh_pat()`:
      ! A GitHub PAT must be a string

# validate_gh_pat rejects non-gh_pat input

    Code
      validate_gh_pat("not-a-pat-object")
    Condition
      Error in `validate_gh_pat()`:
      ! `x` must be a <gh_pat> object, not the string "not-a-pat-object".

# get_baseurl() insists on http(s)

    Code
      get_baseurl("github.com")
    Condition
      Error in `get_baseurl()`:
      ! Only works with HTTP(S) protocols
    Code
      get_baseurl("github.acme.com")
    Condition
      Error in `get_baseurl()`:
      ! Only works with HTTP(S) protocols

# generates a useful message

    Code
      gh("/missing")
    Condition
      Error in `gh()`:
      ! GitHub API error (404): Not Found
      x URL not found: <https://api.github.com/missing>
      i Read more at <https://docs.github.com/rest>

# handles 422 responses with `errors` as a plain string

    Code
      gh("POST /repos/{owner}/{repo}/statuses/{sha}", owner = "r-lib", repo = "gh",
        sha = "deadbeef", state = "success")
    Condition
      Error in `gh()`:
      ! GitHub API error (422): Validation Failed
      i Read more at <https://docs.github.com/rest/commits/statuses#create-a-commit-status>
      message
      Validation failed: This SHA and context has reached the maximum number of statuses.

# can use per_page or .per_page but not both

    Code
      gh("/orgs/tidyverse/repos", per_page = 1, .per_page = 2)
    Condition
      Error in `gh()`:
      ! Exactly one of `per_page` or `.per_page` must be supplied.

# paginated request surfaces a 4xx error

    Code
      gh("/orgs/{org}/repos", org = "gh-org-testing-404", .limit = 100)
    Condition
      Error in `gh()`:
      ! GitHub API error (404): Not Found
      x URL not found: <https://api.github.com/orgs/gh-org-testing-404/repos?per_page=100>
      i Read more at <https://docs.github.com/rest>

GitHub takes the security of our software products and services seriously, including the open source code repositories managed through our GitHub organizations, such as [cli](https://github.com/cli).

If you believe you have found a security vulnerability in GitHub CLI, you can report it to us in one of two ways:

* Report it to this repository directly using [private vulnerability reporting][]. Such reports are not eligible for a bounty reward.

* Submit the report through [HackerOne][] to be eligible for a bounty reward.

**Please do not report security vulnerabilities through public GitHub issues, discussions, or pull requests.**

Thanks for helping make GitHub safe for everyone.

  [private vulnerability reporting]: https://github.com/cli/cli/security/advisories
  [HackerOne]: https://hackerone.com/github

# named NA is error

    Code
      check_named_nas(tc)
    Condition
      Error in `check_named_nas()`:
      ! Named NA parameters are not allowed: `a`

---

    Code
      check_named_nas(tc)
    Condition
      Error in `check_named_nas()`:
      ! Named NA parameters are not allowed: `a`

---

    Code
      check_named_nas(tc)
    Condition
      Error in `check_named_nas()`:
      ! Named NA parameters are not allowed: `c`

# discard errors on logical selector of wrong length

    Code
      discard(list("a", "b"), c(TRUE, FALSE, TRUE))
    Condition
      Error in `probe()`:
      ! `.p` must have the same length as `.x`.
      i This is an internal error that was detected in the gh package.
        Please report it at <https://github.com/r-lib/gh/issues> with a reprex (<https://tidyverse.org/help/>) and the full backtrace.

# gh_process_response errors on non-httr2_response input

    Code
      gh_process_response(list(), list())
    Condition
      Error in `gh_process_response()`:
      ! `resp` must be an <httr2_response> object, not an empty list.

# warns if output is HTML

    Code
      res <- gh("POST /markdown", text = "foo")
    Condition
      Warning:
      Response came back as html :(

# errors

    Code
      gh_rate_limit(list())
    Condition
      Error in `gh_rate_limit()`:
      ! `response` must be a <gh_response> object, not an empty list.
    Code
      gh_rate_limits(.token = "bad")
    Condition
      Error in `gh()`:
      ! GitHub API error (401): Bad credentials
      i Read more at <https://docs.github.com/rest>

# can extract relative pages

    Code
      gh_prev(page1)
    Condition
      Error in `gh_link_request()`:
      ! No prev page

# gh_link_request errors on non-gh_response input

    Code
      gh_link_request(list(), "next", .token = NULL, .send_headers = NULL)
    Condition
      Error in `gh_link_request()`:
      ! `gh_response` must be a <gh_response> object, not an empty list.

# gh_set_endpoint() refuses to substitute an NA

    Code
      gh_set_endpoint(input)
    Condition
      Error in `gh_set_endpoint()`:
      ! Named NA parameters are not allowed: org

# gh_make_request() errors if unknown verb

    Unknown HTTP verb: "GEEET"

# gh_set_query errors when GET params are not all named

    Code
      gh_set_query(input)
    Condition
      Error in `gh_set_query()`:
      ! All elements of `params` must be named for `GET` requests.

# gh_tree_remote errors when no github remotes are configured

    Code
      gh_tree_remote(repo)
    Condition
      Error in `github_remote()`:
      ! No GitHub remotes found at '<tempdir>/<tempfile>'

# git_config errors when .git/config does not exist

    Code
      git_config(dir)
    Condition
      Error in `git_config()`:
      ! git config does not exist at '<tempdir>/<tempfile>'

# repo_root errors on a path that does not exist

    Code
      repo_root(file.path(tempdir(), "does-not-exist-xyz"))
    Condition
      Error in `repo_root()`:
      ! Can't find repo at '<tempdir>/does-not-exist-xyz'

# repo_root errors when no git root is found

    Code
      repo_root(dir)
    Condition
      Error in `repo_root()`:
      ! Could not find git root from '/'.

# can print all types of object

    Code
      json
    Output
      {
        "name": "LICENSE",
        "path": "LICENSE",
        "type": "file",
        "content": "ZmFrZQ==",
        "encoding": "base64"
      } 
    Code
      file
    Output
      [1] "LICENSE"
      attr(,"class")
      [1] "gh_response" "path"       
    Code
      raw
    Output
       [1] 50 61 63 6b 61 67 65 3a 20 67 68 0a 54 69 74 6c 65 3a 20 46 61 6b 65 20 63
      [26] 6f 6e 74 65 6e 74 20 6f 66 20 4c 49 43 45 4e 53 45 0a
      attr(,"class")
      [1] "gh_response" "raw"        

# whoami errors with bad/absent PAT

    Code
      gh_whoami(.token = "")
    Message
      No personal access token (PAT) available.
      Obtain a PAT from here:
      https://github.com/settings/tokens
      For more on what to do with the PAT, see ?gh_whoami.
    Code
      gh_whoami(.token = NA)
    Condition
      Error in `gh()`:
      ! GitHub API error (401): Requires authentication
      i Read more at <https://docs.github.com/rest>
    Code
      gh_whoami(.token = "blah")
    Condition
      Error in `gh()`:
      ! GitHub API error (401): Bad credentials
      i Read more at <https://docs.github.com/rest>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.gh_response}
\alias{print.gh_response}
\title{Print the result of a GitHub API call}
\usage{
\method{print}{gh_response}(x, ...)
}
\arguments{
\item{x}{The result object.}

\item{...}{Ignored.}
}
\value{
The JSON result.
}
\description{
Print the result of a GitHub API call
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh.R
\name{gh}
\alias{gh}
\title{Query the GitHub API}
\usage{
gh(
  endpoint,
  ...,
  per_page = NULL,
  .per_page = NULL,
  .token = NULL,
  .destfile = NULL,
  .overwrite = FALSE,
  .api_url = NULL,
  .method = "GET",
  .limit = NULL,
  .accept = "application/vnd.github.v3+json",
  .send_headers = NULL,
  .progress = TRUE,
  .params = list(),
  .max_wait = 600,
  .max_rate = NULL
)
}
\arguments{
\item{endpoint}{GitHub API endpoint. Must be one of the following forms:
\itemize{
\item \verb{METHOD path}, e.g. \code{GET /rate_limit},
\item \code{path}, e.g. \verb{/rate_limit},
\item \verb{METHOD url}, e.g. \verb{GET https://api.github.com/rate_limit},
\item \code{url}, e.g. \verb{https://api.github.com/rate_limit}.
}

If the method is not supplied, will use \code{.method}, which defaults
to \code{"GET"}.}

\item{...}{Name-value pairs giving API parameters. Will be matched into
\code{endpoint} placeholders, sent as query parameters in GET requests, and as a
JSON body of POST requests. If there is only one unnamed parameter, and it
is a raw vector, then it will not be JSON encoded, but sent as raw data, as
is. This can be used for example to add assets to releases. Named \code{NULL}
values are silently dropped. For GET requests, named \code{NA} values trigger an
error. For other methods, named \code{NA} values are included in the body of the
request, as JSON \code{null}.}

\item{per_page, .per_page}{Number of items to return per page. If omitted,
will be substituted by \code{max(.limit, 100)} if \code{.limit} is set,
otherwise determined by the API (never greater than 100).}

\item{.token}{Authentication token. Defaults to \code{GITHUB_PAT} or
\code{GITHUB_TOKEN} environment variables, in this order if any is set.
See \code{\link[=gh_token]{gh_token()}} if you need more flexibility, e.g. different tokens
for different GitHub Enterprise deployments.}

\item{.destfile}{Path to write response to disk. If \code{NULL} (default),
response will be processed and returned as an object. If path is given,
response will be written to disk in the form sent. gh writes the
response to a temporary file, and renames that file to \code{.destfile}
after the request was successful. The name of the temporary file is
created by adding a \verb{-<random>.gh-tmp} suffix to it, where \verb{<random>}
is an ASCII string with random characters. gh removes the temporary
file on error.}

\item{.overwrite}{If \code{.destfile} is provided, whether to overwrite an
existing file.  Defaults to \code{FALSE}. If an error happens the original
file is kept.}

\item{.api_url}{Github API url (default: \url{https://api.github.com}). Used
if \code{endpoint} just contains a path. Defaults to \code{GITHUB_API_URL}
environment variable if set.}

\item{.method}{HTTP method to use if not explicitly supplied in the
\code{endpoint}.}

\item{.limit}{Number of records to return. This can be used
instead of manual pagination. By default it is \code{NULL},
which means that the defaults of the GitHub API are used.
You can set it to a number to request more (or less)
records, and also to \code{Inf} to request all records.
Note, that if you request many records, then multiple GitHub
API calls are used to get them, and this can take a potentially
long time.}

\item{.accept}{The value of the \code{Accept} HTTP header. Defaults to
\code{"application/vnd.github.v3+json"} . If \code{Accept} is given in
\code{.send_headers}, then that will be used. This parameter can be used to
provide a custom media type, in order to access a preview feature of
the API.}

\item{.send_headers}{Named character vector of header field values
(except \code{Authorization}, which is handled via \code{.token}). This can be
used to override or augment the default \code{User-Agent} header:
\code{"https://github.com/r-lib/gh"}.}

\item{.progress}{Whether to show a progress indicator for calls that
need more than one HTTP request.}

\item{.params}{Additional list of parameters to append to \code{...}.
It is easier to use this than \code{...} if you have your parameters in
a list already.}

\item{.max_wait}{Maximum number of seconds to wait if rate limited.
Defaults to 10 minutes.}

\item{.max_rate}{Maximum request rate in requests per second. Set
this to automatically throttle requests.}
}
\value{
Answer from the API as a \code{gh_response} object, which is also a
\code{list}. Failed requests will generate an R error. Requests that
generate a raw response will return a raw vector.
}
\description{
This is an extremely minimal client. You need to know the API
to be able to use this client. All this function does is:
\itemize{
\item Try to substitute each listed parameter into \code{endpoint}, using the
\code{{parameter}} notation.
\item If a GET request (the default), then add all other listed parameters
as query parameters.
\item If not a GET request, then send the other parameters in the request
body, as JSON.
\item Convert the response to an R list using \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}}.
}
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Repositories of a user, these are equivalent
gh("/users/hadley/repos", .limit = 2)
gh("/users/{username}/repos", username = "hadley", .limit = 2)

## Starred repositories of a user
gh("/users/hadley/starred", .limit = 2)
gh("/users/{username}/starred", username = "hadley", .limit = 2)
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Create a repository, needs a token in GITHUB_PAT (or GITHUB_TOKEN)
## environment variable
gh("POST /user/repos", name = "foobar")
\dontshow{\}) # examplesIf}
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Issues of a repository
gh("/repos/hadley/dplyr/issues")
gh("/repos/{owner}/{repo}/issues", owner = "hadley", repo = "dplyr")

## Automatic pagination
users <- gh("/users", .limit = 50)
length(users)
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Access developer preview of Licenses API (in preview as of 2015-09-24)
gh("/licenses") # used to error code 415
gh("/licenses", .accept = "application/vnd.github.drax-preview+json")
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Access Github Enterprise API
## Use GITHUB_API_URL environment variable to change the default.
gh("/user/repos", type = "public", .api_url = "https://github.foobar.edu/api/v3")
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Use I() to force body part to be sent as an array, even if length 1
## This works whether assignees has length 1 or > 1
assignees <- "gh_user"
assignees <- c("gh_user1", "gh_user2")
gh("PATCH /repos/OWNER/REPO/issues/1", assignees = I(assignees))
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## There are two ways to send JSON data. One is that you supply one or
## more objects that will be converted to JSON automatically via
## jsonlite::toJSON(). In this case sometimes you need to use
## jsonlite::unbox() because fromJSON() creates lists from scalar vectors
## by default. The Content-Type header is automatically added in this
## case. For example this request turns on GitHub Pages, using this
## API: https://docs.github.com/v3/repos/pages/#enable-a-pages-site

gh::gh(
  "POST /repos/{owner}/{repo}/pages",
  owner = "r-lib",
  repo = "gh",
  source = list(
    branch = jsonlite::unbox("gh-pages"),
    path = jsonlite::unbox("/")
  ),
  .send_headers = c(Accept = "application/vnd.github.switcheroo-preview+json")
)

## The second way is to handle the JSON encoding manually, and supply it
## as a raw vector in an unnamed argument, and also a Content-Type header:

body <- '{ "source": { "branch": "gh-pages", "path": "/" } }'
gh::gh(
  "POST /repos/{owner}/{repo}/pages",
  owner = "r-lib",
  repo = "gh",
  charToRaw(body),
  .send_headers = c(
    Accept = "application/vnd.github.switcheroo-preview+json",
    "Content-Type" = "application/json"
  )
)
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
## Pass along a query to the search/code endpoint via the ... argument
x <- gh::gh(
            "/search/code",
            q = "installation repo:r-lib/gh",
            .send_headers = c("X-GitHub-Api-Version" = "2022-11-28")
            )
 str(x, list.len = 3, give.attr = FALSE)

\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=gh_gql]{gh_gql()}} if you want to use the GitHub GraphQL API,
\code{\link[=gh_whoami]{gh_whoami()}} for details on GitHub API token management.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh_token.R
\name{gh_token}
\alias{gh_token}
\title{Return the local user's GitHub Personal Access Token (PAT)}
\usage{
gh_token(api_url = NULL)
}
\arguments{
\item{api_url}{GitHub API URL. Defaults to the \code{GITHUB_API_URL} environment
variable, if set, and otherwise to \url{https://api.github.com}.}
}
\value{
A string of characters, if a PAT is found, or the empty
string, otherwise. For convenience, the return value has an S3 class in
order to ensure that simple printing strategies don't reveal the entire
PAT.
}
\description{
If gh can find a personal access token (PAT) via \code{gh_token()}, it includes
the PAT in its requests. Some requests succeed without a PAT, but many
require a PAT to prove the request is authorized by a specific GitHub user. A
PAT also helps with rate limiting. If your gh use is more than casual, you
want a PAT.

gh calls \code{\link[gitcreds:gitcreds_get]{gitcreds::gitcreds_get()}} with the \code{api_url}, which checks session
environment variables and then the local Git credential store for a PAT
appropriate to the \code{api_url}. Therefore, if you have previously used a PAT
with, e.g., command line Git, gh may retrieve and re-use it. You can call
\code{\link[gitcreds:gitcreds_get]{gitcreds::gitcreds_get()}} directly, yourself, if you want to see what is
found for a specific URL. If no matching PAT is found,
\code{\link[gitcreds:gitcreds_get]{gitcreds::gitcreds_get()}} errors, whereas \code{gh_token()} does not and,
instead, returns \code{""}.

See GitHub's documentation on \href{https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token}{Creating a personal access token},
or use \code{usethis::create_github_token()} for a guided experience, including
pre-selection of recommended scopes. Once you have a PAT, you can use
\code{\link[gitcreds:gitcreds_get]{gitcreds::gitcreds_set()}} to add it to the Git credential store. From that
point on, gh (via \code{\link[gitcreds:gitcreds_get]{gitcreds::gitcreds_get()}}) should be able to find it
without further effort on your part.
}
\examples{
\dontrun{
gh_token()

format(gh_token())

str(gh_token())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh_whoami.R
\name{gh_whoami}
\alias{gh_whoami}
\title{Info on current GitHub user and token}
\usage{
gh_whoami(.token = NULL, .api_url = NULL, .send_headers = NULL)
}
\arguments{
\item{.token}{Authentication token. Defaults to \code{\link[=gh_token]{gh_token()}}.}

\item{.api_url}{Github API url (default: \url{https://api.github.com}). Used
if \code{endpoint} just contains a path. Defaults to \code{GITHUB_API_URL}
environment variable if set.}

\item{.send_headers}{Named character vector of header field values
(except \code{Authorization}, which is handled via \code{.token}). This can be
used to override or augment the default \code{User-Agent} header:
\code{"https://github.com/r-lib/gh"}.}
}
\value{
A \code{gh_response} object, which is also a \code{list}.
}
\description{
Reports wallet name, GitHub login, and GitHub URL for the current
authenticated user, the first bit of the token, and the associated scopes.
}
\details{
Get a personal access token for the GitHub API from
\url{https://github.com/settings/tokens} and select the scopes necessary for your
planned tasks. The \code{repo} scope, for example, is one many are likely to need.

On macOS and Windows it is best to store the token in the git credential
store, where most GitHub clients, including gh, can access it. You can
use the gitcreds package to add your token to the credential store:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{gitcreds::gitcreds_set()
}\if{html}{\out{</div>}}

See \url{https://gh.r-lib.org/articles/managing-personal-access-tokens.html}
and \url{https://usethis.r-lib.org/articles/articles/git-credentials.html}
for more about managing GitHub (and generic git) credentials.

On other systems, including Linux, the git credential store is
typically not as convenient, and you might want to store your token in
the \code{GITHUB_PAT} environment variable, which you can set in your
\code{.Renviron} file.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) withAutoprint(\{ # examplesIf}
gh_whoami()
\dontshow{\}) # examplesIf}
\dontshow{if (FALSE) withAutoprint(\{ # examplesIf}
## explicit token + use with GitHub Enterprise
gh_whoami(
  .token = "8c70fd8419398999c9ac5bacf3192882193cadf2",
  .api_url = "https://github.foobar.edu/api/v3"
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pagination.R
\name{gh_next}
\alias{gh_next}
\alias{gh_prev}
\alias{gh_first}
\alias{gh_last}
\title{Get the next, previous, first or last page of results}
\usage{
gh_next(gh_response)

gh_prev(gh_response)

gh_first(gh_response)

gh_last(gh_response)
}
\arguments{
\item{gh_response}{An object returned by a \code{\link[=gh]{gh()}} call.}
}
\value{
Answer from the API.
}
\description{
Get the next, previous, first or last page of results
}
\details{
Note that these are not always defined. E.g. if the first
page was queried (the default), then there are no first and previous
pages defined. If there is no next page, then there is no
next page defined, etc.

If the requested page does not exist, an error is thrown.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
x <- gh("/users")
vapply(x, "[[", character(1), "login")
x2 <- gh_next(x)
vapply(x2, "[[", character(1), "login")
\dontshow{\}) # examplesIf}
}
\seealso{
The \code{.limit} argument to \code{\link[=gh]{gh()}} supports fetching more than
one page.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh_gql.R
\name{gh_gql}
\alias{gh_gql}
\title{A simple interface for the GitHub GraphQL API v4.}
\usage{
gh_gql(query, ...)
}
\arguments{
\item{query}{The GraphQL query, as a string.}

\item{...}{Name-value pairs giving API parameters. Will be matched into
\code{endpoint} placeholders, sent as query parameters in GET requests, and as a
JSON body of POST requests. If there is only one unnamed parameter, and it
is a raw vector, then it will not be JSON encoded, but sent as raw data, as
is. This can be used for example to add assets to releases. Named \code{NULL}
values are silently dropped. For GET requests, named \code{NA} values trigger an
error. For other methods, named \code{NA} values are included in the body of the
request, as JSON \code{null}.}
}
\description{
See more about the GraphQL API here:
\url{https://docs.github.com/graphql}
}
\details{
Note: pagination and the \code{.limit} argument does not work currently,
as pagination in the GraphQL API is different from the v3 API.
If you need pagination with GraphQL, you'll need to do that manually.
}
\examples{
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
gh_gql("query { viewer { login }}")

# Get rate limit
ratelimit_query <- "query {
  viewer {
    login
  }
  rateLimit {
    limit
    cost
    remaining
    resetAt
  }
}"

gh_gql(ratelimit_query)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=gh]{gh()}} for the GitHub v3 API.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fake-github-app.R
\name{fake_github_app}
\alias{fake_github_app}
\title{A fake GitHub web app}
\usage{
fake_github_app()
}
\value{
A \code{webfakes_app} object.
}
\description{
A \code{\link[webfakes:new_app]{webfakes::new_app()}} application that implements the subset of the
GitHub REST API needed by the gh test suite. It is exported so that
downstream packages that depend on gh can use it to test their own
GitHub-backed code without hitting the real API.
}
\details{
The app accepts any token whose shape is a valid GitHub PAT (40 hex
characters or a \code{ghp_} / \code{github_pat_} / \code{ghs_} / \code{ghr_} / \code{gho_} /
\code{ghu_} prefix). Any other non-empty token is rejected with a 401
"Bad credentials". A missing \code{Authorization} header yields a 401
"Requires authentication" for endpoints that need a user.
}
\examples{
\dontshow{if (rlang::is_installed("webfakes")) withAutoprint(\{ # examplesIf}
app <- fake_github_app()
proc <- webfakes::new_app_process(app)
proc$url()
proc$stop()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git.R
\name{gh_tree_remote}
\alias{gh_tree_remote}
\title{Find the GitHub remote associated with a path}
\usage{
gh_tree_remote(path = ".")
}
\arguments{
\item{path}{Path that is contained within a git repo.}
}
\value{
If the repo has a github remote, a list containing \code{username}
and \code{repo}. Otherwise, an error.
}
\description{
This is handy helper if you want to make gh requests related to the
current project.
}
\examples{
\dontshow{if (interactive()) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
gh_tree_remote()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh-package.R
\docType{package}
\name{gh-package}
\alias{gh-package}
\title{gh: 'GitHub' 'API'}
\description{
Minimal client to access the 'GitHub' 'API'.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://gh.r-lib.org/}
  \item \url{https://github.com/r-lib/gh#readme}
  \item Report bugs at \url{https://github.com/r-lib/gh/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com} [contributor]

Authors:
\itemize{
  \item Jennifer Bryan
  \item Hadley Wickham
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gh_rate_limit.R
\name{gh_rate_limit}
\alias{gh_rate_limit}
\alias{gh_rate_limits}
\title{Return GitHub user's current rate limits}
\usage{
gh_rate_limit(
  response = NULL,
  .token = NULL,
  .api_url = NULL,
  .send_headers = NULL
)

gh_rate_limits(.token = NULL, .api_url = NULL, .send_headers = NULL)
}
\arguments{
\item{response}{\code{gh_response} object from a previous \code{gh} call, rate
limit values are determined from values in the response header.
Optional argument, if missing a call to "GET /rate_limit" will be made.}

\item{.token}{Authentication token. Defaults to \code{GITHUB_PAT} or
\code{GITHUB_TOKEN} environment variables, in this order if any is set.
See \code{\link[=gh_token]{gh_token()}} if you need more flexibility, e.g. different tokens
for different GitHub Enterprise deployments.}

\item{.api_url}{Github API url (default: \url{https://api.github.com}). Used
if \code{endpoint} just contains a path. Defaults to \code{GITHUB_API_URL}
environment variable if set.}

\item{.send_headers}{Named character vector of header field values
(except \code{Authorization}, which is handled via \code{.token}). This can be
used to override or augment the default \code{User-Agent} header:
\code{"https://github.com/r-lib/gh"}.}
}
\value{
A \code{list} object containing the overall \code{limit}, \code{remaining} limit, and the
limit \code{reset} time.
}
\description{
\code{gh_rate_limits()} reports on all rate limits for the authenticated user.
\code{gh_rate_limit()} reports on rate limits for previous successful request.

Further details on GitHub's API rate limit policies are available at
\url{https://docs.github.com/v3/#rate-limiting}.
}
